/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 1019 $
 * $Date: 2020-05-05 20:54:39 +0200 (Tue, 05 May 2020) $
 */
package at.lcm.bu21.general.dtypes;

import java.util.ArrayList;
import java.util.Arrays;

import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLStreamWriter;

import at.lcm.bu21.general.utils.Clone;
import at.lcm.bu21.general.utils.CloneException;

/**
 * Representation of a array-type
 * 
 * @author Werner Koppelsttter
 */
public class TArray extends TData {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** Data variable */
	protected ArrayList<ArrayList<TData>> data = new ArrayList<ArrayList<TData>>();

	/**
	 * Create empty object
	 */
	public TArray() {

	}

	/**
	 * Create object with given size
	 * 
	 * @param size
	 *            Size of Array
	 * @throws TDataException
	 *             Wrong dimension
	 */
	public TArray(int[] size) throws TDataException {
		if (size.length == 2) {
			this.expandArray(size[0], size[1]);
		} else {
			throw new TDataException("Only Array of dimension 2 allowed!");
		}
	}

	/**
	 * Create object with given size
	 * 
	 * @param row
	 *            Number of rows
	 * @param col
	 *            Number of column
	 * @throws TDataException
	 *             Unable to create data-object
	 */
	public TArray(int row, int col) throws TDataException {
		this(new int[] { row, col });
	}

	/**
	 * Create object with given data (ArrayList)
	 * 
	 * @param row
	 *            Row of data to insert (ArrayList)
	 */
	public TArray(ArrayList<TData> row) {
		this.data = new ArrayList<ArrayList<TData>>();
		this.data.add(row);
	}

	@Override
	protected TData getData(String[] ident) {
		TData data, tmp;
		String[] items;
		int row, col;

		data = super.getData(ident);
		if (data == null) {
			items = ident[0].split("[,\\]]");
			if (items.length == 2) {
				try {
					row = Integer.parseInt(items[0]);
					col = Integer.parseInt(items[1]);
					tmp = this.get(row, col);
					if (tmp != null) {
						data = tmp.getData(Arrays.copyOfRange(ident, 1, ident.length));
					}
				} catch (Exception ex) {
				}
			}
		}

		return data;
	}

	@Override
	public TArray deepClone() throws CloneException {
		return (TArray) Clone.deepClone(this);
	}

	/**
	 * Expand size of array
	 * 
	 * @param row
	 *            Number of rows
	 * @param col
	 *            Number of columns
	 */
	protected void expandArray(int row, int col) {
		ArrayList<TData> r;

		// add absent rows
		while (row > this.data.size()) {
			this.data.add(new ArrayList<TData>());
		}

		// add absent columns
		for (int i = 0; i < this.data.size(); i++) {
			r = this.data.get(i);
			while (col > r.size()) {
				r.add(new TNull());
			}
		}
	}

	/**
	 * Set value of array
	 * 
	 * @param row
	 *            Index of row of array
	 * @param col
	 *            Index of column of array
	 * @param value
	 *            Value to insert
	 */
	public void set(int row, int col, TData value) {

		if (value != null) {
			// expand array
			if ((row >= this.size()[0]) || (col >= this.size()[1])) {
				this.expandArray(Math.max(row + 1, this.size()[0]), Math.max(col + 1, this.size()[1]));
			}

			// replace value on give position
			this.data.get(row).set(col, value);
		} else {
			throw new NullPointerException("Null is not allowed to add to TArray");
		}
	}

	/**
	 * Get value of array
	 * 
	 * @param row
	 *            Index of row of array
	 * @param col
	 *            INdex of column of array
	 * @return Value of array
	 * @throws TDataException
	 *             Invalid index
	 */
	public TData get(int row, int col) throws TDataException {
		if ((row < 0) || (row >= this.data.size()) || (col < 0) || (col >= this.data.get(0).size())) {
			throw new TDataException("Index out of bounds!");
		}

		return this.data.get(row).get(col);
	}

	/**
	 * Get a row of a array
	 * 
	 * @param row
	 *            Index of row of array
	 * @return Row of array
	 * @throws TDataException
	 *             Invalid index
	 */
	public TArray getRow(int row) throws TDataException {

		if ((row < 0) || (this.data.size() <= row)) {
			throw new TDataException("Index out of bounds!");
		}

		return new TArray(this.data.get(row));
	}

	/**
	 * Get a column of a array
	 * 
	 * @param col
	 *            Index of column of array
	 * @return Column of array
	 * @throws TDataException
	 *             Invalid index
	 */
	public TArray getColumn(int col) throws TDataException {
		TArray array;

		if ((col < 0) || (this.size()[1] <= col)) {
			throw new TDataException("Index out of bounds!");
		}

		array = new TArray();
		for (int row = 0; row < this.size()[0]; row++) {
			array.set(row, 0, this.get(row, col));
		}

		return array;
	}

	/**
	 * Get a array of the size (two dimension)
	 * 
	 * @return Size of array
	 */
	public int[] size() {
		int[] size = { 0, 0 };
		size[0] = this.data.size();
		if (size[0] != 0) {
			size[1] = this.data.get(0).size();
		}
		return size;
	}

	@Override
	public void xml2data(XMLStreamReader xmlReader) throws XMLStreamException, TDataException {
		ArrayList<TData> row;
		Integer nrows;
		Integer ncolumns;

		nrows = Integer.parseInt(xmlReader.getAttributeValue(null, "nrows"));
		ncolumns = Integer.parseInt(xmlReader.getAttributeValue(null, "ncolumns"));
		this.data = new ArrayList<ArrayList<TData>>(nrows);

		xmlReader.next();
		while (xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT) {
			xmlReader.next();
			row = new ArrayList<TData>(ncolumns);
			this.data.add(row);

			while (xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT) {
				row.add(TData.xml2obj(xmlReader));
			}
			if (row.size() != ncolumns) {
				throw new TDataException("Wrong number of elements!");
			}
			xmlReader.next();
		}
		if (this.data.size() != nrows) {
			throw new TDataException("Wrong number of elements!");
		}
		xmlReader.next();
	}

	@Override
	public void data2xml(XMLStreamWriter xmlWriter) throws XMLStreamException {
		int rows = 0, cols = 0;
		rows = this.data.size();
		if (rows > 0) {
			cols = this.data.get(0).size();
		}

		xmlWriter.writeStartElement("Array");
		xmlWriter.writeAttribute("nrows", String.valueOf(rows));
		xmlWriter.writeAttribute("ncolumns", String.valueOf(cols));
		for (int r = 0; r < this.data.size(); r++) {
			xmlWriter.writeStartElement("row");
			for (int c = 0; c < this.data.get(r).size(); c++) {
				this.data.get(r).get(c).data2xml(xmlWriter);
			}
			xmlWriter.writeEndElement();
		}
		xmlWriter.writeEndElement();

	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (obj == null) {
			return false;
		}
		if (this.getClass() != obj.getClass()) {
			return false;
		}
		TArray other = (TArray) obj;
		if (this.data == null) {
			if (other.data != null) {
				return false;
			}
		} else if (!this.data.equals(other.data)) {
			return false;
		}
		return true;
	}

	@Override
	public String toString() {
		int nrow = 0, ncol = 0;
		nrow = this.data.size();
		if (nrow != 0) {
			ncol = this.data.get(0).size();
		}
		return "[" + nrow + "x" + ncol + " Array]";
	}

}
