/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 997 $
 * $Date: 2020-03-19 13:49:34 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.general.dtypes;

import java.util.Arrays;
import java.util.Hashtable;
import java.util.Map;

import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLStreamWriter;

import at.lcm.bu21.general.utils.Clone;
import at.lcm.bu21.general.utils.CloneException;

/**
 * Representation of a structure-type
 * 
 * @author Werner Koppelsttter
 */
public class TStruct extends TData {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** Data variable */
	protected Hashtable<String, TData> data = new Hashtable<String, TData>();

	/**
	 * Create empty object
	 */
	public TStruct() {

	}

	/**
	 * Create TStruct using Map
	 * 
	 * @param map
	 *            Map with entries for struct
	 */
	public TStruct(Map<String, TData> map) {
		this.data = new Hashtable<String, TData>(map);
	}

	@Override
	protected TData getData(String[] ident) {
		TData data, tmp;

		data = super.getData(ident);
		if (data == null) {
			tmp = this.get(ident[0]);
			if (tmp != null) {
				data = tmp.getData(Arrays.copyOfRange(ident, 1, ident.length));
			}
		}

		return data;
	}

	@Override
	public TStruct deepClone() throws CloneException {
		return (TStruct) Clone.deepClone(this);
	}

	@Override
	public void xml2data(XMLStreamReader xmlReader)
			throws XMLStreamException, TDataException {
		String fieldName;

		xmlReader.next();
		while (xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT) {
			fieldName = xmlReader.getAttributeValue(null, "name");
			xmlReader.next();
			this.data.put(fieldName, TData.xml2obj(xmlReader));
			xmlReader.next();
		}
		xmlReader.next();
	}

	@Override
	public void data2xml(XMLStreamWriter xmlWriter) throws XMLStreamException {

		String[] keys = this.data.keySet().toArray(new String[0]);
		Arrays.sort(keys);

		xmlWriter.writeStartElement("Struct");
		for (String key : keys) {
			xmlWriter.writeStartElement("field");
			xmlWriter.writeAttribute("name", key);
			this.data.get(key).data2xml(xmlWriter);
			xmlWriter.writeEndElement();
		}
		xmlWriter.writeEndElement();
	}

	/**
	 * Set data of a field-name
	 * 
	 * @param key
	 *            Field-name
	 * @param value
	 *            Data to insert
	 */
	public void put(String key, TData value) {
		// if (key.matches("[a-zA-Z]\\w*")) {
		this.data.put(key, value);
		// } else {
		// throw new TDataException("Invalid Key \"+ key+ \"");
		// }
	}

	/**
	 * Get the data of a field-name
	 * 
	 * @param key
	 *            Name of field
	 * @return Field value
	 */
	public TData get(String key) {
		return this.data.get(key);
	}

	/**
	 * Removes the given field from struct
	 * 
	 * @param key
	 *            Name of field
	 */
	public void remove(String key) {
		this.data.remove(key);
	}

	/**
	 * Get field-names of struct
	 * 
	 * @return List of field-names
	 */
	public String[] getKeys() {
		return this.data.keySet().toArray(new String[0]);
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (obj == null) {
			return false;
		}
		if (this.getClass() != obj.getClass()) {
			return false;
		}
		TStruct other = (TStruct) obj;
		if (this.data == null) {
			if (other.data != null) {
				return false;
			}
		} else if (!this.data.equals(other.data)) {
			return false;
		}
		return true;
	}

	@Override
	public String toString() {
		return "[struct with " + this.data.size() + " fields]";
	}

}
