/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 1002 $
 * $Date: 2020-04-10 23:31:24 +0200 (Fri, 10 Apr 2020) $
 */
package at.lcm.bu21.general.dtypes.editor2.TDataModel;

import at.lcm.bu21.general.dtypes.TPeriodic;

/**
 * Representation of a TPeriodicModel - type
 * 
 * @author Johannes Kaisinger
 */
public class TPeriodicModel extends TDataModel {

	/** Data variable */
	private TPeriodic data;

	/**
	 * Create object with size 1
	 * 
	 */
	public TPeriodicModel() {
		this.data = new TPeriodic(1, 1.0, 0);
	}

	/**
	 * Create object with given data (TPeriodic)
	 * 
	 * @param data
	 *            Reference of data (TPeriodic)
	 */
	public TPeriodicModel(TPeriodic data) {
		this.data = data;
	}

	/**
	 * Get all data (TData)
	 * 
	 * @return Data Reference of data (TData)
	 */
	@Override
	public TPeriodic getData() {
		return this.data;
	}

	/**
	 * @param data
	 * 
	 */
	public void setData(TPeriodic data) {
		this.data = data;
	}

	/**
	 * Add a new signal (maxorder+1)
	 */
	public void addSignal() {
		int max = this.data.getMaxOrder();
		int next = max + 1;

		TPeriodic tp = new TPeriodic(next, 1.0, 0.0);

		this.data = this.data.add(tp);
		// this.data.compress();
	}

	/**
	 * Delete the signal with the order
	 * 
	 * @param order
	 *            Index of row of list (place of delete)
	 * 
	 */
	public void removeSignal(int order) {
		int[] orders = this.data.usedOrders();
		TPeriodic temp = new TPeriodic();

		for (int o : orders) {
			if (o != order) {
				int[] add = { o };
				temp = temp.add(this.data.select(add));
			}
		}
		this.data = temp;
	}

	/**
	 * Get all usedOrders
	 * 
	 * @return usedOrders
	 */
	public int[] getUsedOrders() {
		return this.data.usedOrders();
	}

	/**
	 * Get CoeffA
	 * 
	 * @param order
	 *            Order of CoeffA
	 * @return CoeffA
	 */
	public double getCoeffA(int order) {
		return this.data.getCoeffA(order);
	}

	/**
	 * Get the coeffB of the sub-signal
	 * 
	 * @param order
	 *            Order of the sub-signal
	 * @return CoeffB
	 */
	public double getCoeffB(int order) {
		return this.data.getCoeffB(order);
	}

	/**
	 * Get the magnitude of the sub-signal
	 * 
	 * @param order
	 *            Order of the sub-signal
	 * @return Magnitude Magnitude of the sub-signal
	 */
	public double getMag(int order) {
		return this.data.magnitude(order);
	}

	/**
	 * Get the phase of the sub-signal (TPeriodic)
	 * 
	 * @param order
	 *            Order of the sub-signal
	 * @return Phase Phase of the sub-signal
	 */
	public double getPha(int order) {
		return this.data.phase(order);
	}

	/**
	 * Get the dcOffset of the signal (TPeriodic)
	 * 
	 * @return dcOffset dcOffset
	 */
	public double getDcOffset() {
		return this.data.mean();
	}

	/**
	 * Get the number of orders used in the signal (TPeriodic)
	 * 
	 * @return length
	 */
	public int getLength() {
		return this.data.usedOrders().length;
	}

	/**
	 * Set the order of the sub-signal (TPeriodic)
	 * 
	 * @param order
	 *            Order of the sub-signal
	 * @param newOrder
	 *            New order of the sub-signal
	 */
	public void resetOrder(int order, int newOrder) {
		int[] o = { order };
		TPeriodic oldTp = this.data.select(o);
		this.removeSignal(order);

		double mag = oldTp.magnitude(order);
		double pha = oldTp.phase(order);

		TPeriodic newTp = new TPeriodic(newOrder, mag, pha);
		this.data = this.data.add(newTp);
		// this.data.compress();
	}

	/**
	 * Set the coefficient A of the sub-signal (TPeriodic)
	 * 
	 * @param order
	 *            Order of the sub-signal
	 * @param CoeffA
	 *            Value of the coeffA
	 */
	public void setCoeffA(int order, double CoeffA) {

		final double A0 = this.data.mean();
		final int maxOrder = this.data.getMaxOrder();

		double[] nA = new double[maxOrder];
		double[] nB = new double[maxOrder];

		for (int i = 0; i < maxOrder; i++) {
			nA[i] = this.data.getCoeffA(i + 1);
			nB[i] = this.data.getCoeffB(i + 1);
		}

		nA[order - 1] = CoeffA;

		this.data.setCoeffs(A0, nA, nB);
		this.data.compress();
	}

	/**
	 * Set the coefficient B of the sub-signal (TPeriodic)
	 * 
	 * @param order
	 *            Order of the sub-signal
	 * @param CoeffB
	 *            Value of the coeffB
	 */
	public void setCoeffB(int order, double CoeffB) {

		final double A0 = this.data.mean();
		final int maxOrder = this.data.getMaxOrder();

		double[] nA = new double[maxOrder];
		double[] nB = new double[maxOrder];

		for (int i = 0; i < maxOrder; i++) {
			nA[i] = this.data.getCoeffA(i + 1);
			nB[i] = this.data.getCoeffB(i + 1);
		}

		nB[order - 1] = CoeffB;

		this.data.setCoeffs(A0, nA, nB);
		this.data.compress();
	}

	/**
	 * Set the magnitude of the sub-signal (TPeriodic)
	 * 
	 * @param order
	 *            Order of magnitude
	 * @param mag
	 *            magnitude
	 */
	public void setMag(int order, double mag) {
		int[] o = { order };
		TPeriodic oldTp = this.data.select(o);
		this.removeSignal(order);

		double pha = oldTp.phase(order);

		TPeriodic newTp = new TPeriodic(order, mag, pha);
		this.data = this.data.add(newTp);
		// this.data.compress();
	}

	/**
	 * Set the phase of the sub-signal (TPeriodic)
	 * 
	 * @param order
	 *            Order
	 * @param pha
	 *            Phase
	 */
	public void setPha(int order, double pha) {
		int[] o = { order };
		TPeriodic oldTp = this.data.select(o);
		this.removeSignal(order);

		double mag = oldTp.magnitude(order);

		TPeriodic newTp = new TPeriodic(order, mag, pha);
		this.data = this.data.add(newTp);
		// this.data.compress();
	}

	/**
	 * Set the dc-offset of the signal (TPeriodic)
	 * 
	 * @param offset
	 *            New value for offset.
	 */
	public void setDcOffset(double offset) {
		this.data = this.data.offset(-this.data.mean() + offset);
		// this.data.compress();
	}

	/**
	 * Shows the signal (TPeriodic)
	 */
	public void showTime() {
		try {
			this.data.showPreview();
		} catch (Exception e) {
		}
	}

	/**
	 * Shows the spectrum of the signal (TPeriodic)
	 */
	public void showSpectrum() {
		try {
			this.data.showSpectrum(null);
		} catch (Exception e) {
		}
	}

}
