/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 1022 $
 * $Date: 2020-06-03 14:13:51 +0200 (Wed, 03 Jun 2020) $
 */
package at.lcm.bu21.general.dtypes.unit;

import java.io.Reader;
import java.io.StringReader;

/**
 * Parse syntax of unit (syntax analyze)
 *
 * @author Werner Koppelsttter
 */
public class Parser {

	/** Lexical scanner */
	protected Scanner scan;

	/** Parsed syntax tree */
	protected SyntaxNode syntax;

	/** Scanned token */
	protected Token t;

	/** Look-ahead token */
	protected Token la;

	/** Kind of look-ahead token */
	protected TokenKind sym;

	/**
	 * Initialize Parser with unit
	 * 
	 * @param unit
	 *            Unit to parse
	 * 
	 * @throws ScannerException
	 * @throws ParserException
	 */
	public Parser(String unit) throws ParserException, ScannerException {
		this(new StringReader(unit));
	}

	/**
	 * Initialize Parser with reader of unit
	 * 
	 * @param reader
	 *            Reader with u
	 * 
	 * @throws ScannerException
	 * @throws ParserException
	 */
	public Parser(Reader reader) throws ParserException, ScannerException {

		// lexical analyze
		this.scan = new Scanner(reader);
		// syntax analyze
		this.parse();
	}

	/**
	 * Get syntax-node of unit.
	 * 
	 * @return Syntax-node
	 * @throws EvalException
	 */
	public SyntaxNode getSyntax() throws EvalException {
		return this.syntax;
	}

	/**
	 * Parse unit to create syntax-tree and list variables
	 * 
	 * @throws ParserException
	 * @throws ScannerException
	 */
	private void parse() throws ParserException, ScannerException {
		// syntax analyzer
		this.scan();
		this.syntax = this.Unit();
		this.check(TokenKind.EOF);
	}

	/**
	 * Scan next token
	 * 
	 * @throws ScannerException
	 */
	private void scan() throws ScannerException {
		this.t = this.la;
		this.la = this.scan.next();
		this.sym = this.la.kind;
	}

	/**
	 * Check if next token is the expected
	 * 
	 * @param expected
	 *            Expected token
	 * @throws ScannerException
	 * @throws ParserException
	 */
	private void check(TokenKind expected) throws ScannerException, ParserException {
		if (this.sym == expected) {
			this.scan();
		} else {
			throw new ParserException(expected.getDescription() + " was expected", this.la);
		}
	}

	/**
	 * Unit = Statement | "?".
	 * 
	 * @return Syntax-tree
	 * 
	 * @throws ParserException
	 * @throws ScannerException
	 */
	private SyntaxNode Unit() throws ParserException, ScannerException {
		SyntaxNode sn;

		if (this.sym == TokenKind.QUESTION) {
			sn = new SyntaxNode(this.la);
			this.scan();
		} else {
			sn = this.Statement();
		}

		return sn;
	}

	/**
	 * Statement = Factor {["/"] Factor}.
	 * 
	 * @return Syntax-tree
	 * 
	 * @throws ParserException
	 * @throws ScannerException
	 */
	private SyntaxNode Statement() throws ParserException, ScannerException {
		SyntaxNode sn, tmp;
		Token tok;

		sn = this.Factor();
		while ((this.sym == TokenKind.SLASH) || (this.sym == TokenKind.LPAR) || (this.sym == TokenKind.ELEMENT)
				|| (this.sym == TokenKind.NUMBER)) {
			switch (this.sym) {
			case SLASH:
				tmp = new SyntaxNode(this.la);
				this.scan();
				tmp.addChild(sn);
				tmp.addChild(this.Factor());
				sn = tmp;
				break;
			case LPAR:
			case ELEMENT:
			case NUMBER:
			case QUESTION:
				tok = new Token();
				tok.kind = TokenKind.APPEND;
				tok.line = this.la.line;
				tok.col = tok.col;
				tmp = new SyntaxNode(tok);
				tmp.addChild(sn);
				tmp.addChild(this.Factor());
				sn = tmp;
				break;
			default:
				throw new ParserException("Missplaced token " + this.la.kind.getDescription(), this.la);
			}
		}

		return sn;
	}

	/**
	 * Factor = Part [number].
	 * 
	 * @return Syntax-tree
	 * 
	 * @throws ScannerException
	 * @throws ParserException
	 */
	private SyntaxNode Factor() throws ParserException, ScannerException {
		SyntaxNode sn, tmp;
		Token tok;

		sn = this.Part();
		if (this.sym == TokenKind.NUMBER) {
			tok = new Token();
			tok.kind = TokenKind.EXPONENT;
			tok.line = this.la.line;
			tok.col = tok.col;
			tmp = new SyntaxNode(tok);
			tmp.addChild(sn);
			tmp.addChild(new SyntaxNode(this.la));
			this.scan();
			sn = tmp;
		}

		return sn;
	}

	/**
	 * Value = "(" Statement ")" | element | number | "?".
	 * 
	 * @return Syntax-tree
	 * 
	 * @throws ParserException
	 * @throws ScannerException
	 */
	private SyntaxNode Part() throws ParserException, ScannerException {
		SyntaxNode sn;

		switch (this.sym) {
		case LPAR:
			this.scan();
			sn = this.Statement();
			this.check(TokenKind.RPAR);
			break;
		case NUMBER:
		case ELEMENT:
			sn = new SyntaxNode(this.la);
			this.scan();
			break;
		default:
			throw new ParserException("Missplaced token " + this.la.kind.getDescription(), this.la);
		}

		return sn;
	}

	@Override
	public String toString() {
		return this.syntax.toString();
	}
}
