from abc import ABCMeta, abstractmethod

import numpy as np


class AbstractKernel(metaclass=ABCMeta):
    """
    Abstract class for RBF kernel definitions
    """

    @classmethod
    def fromID(cls, kernel_id, kernel_params=None):
        if kernel_params is None:
            kernel_params = []

        if kernel_id == 0:
            new_cls = KernelLS
        elif kernel_id == 1:
            new_cls = KernelCS
        elif kernel_id == 2:
            new_cls = KernelTPS
        elif kernel_id == 3:
            new_cls = KernelMQ
        elif kernel_id == 4:
            new_cls = KernelGAUSS
        else:
            raise NotImplementedError
        instance = new_cls(*kernel_params)
        return instance

    def __init__(self, kernel_id=-1, kernel_params=None):
        """
        RBF Kernel

        Parameters
        ----------
        kernel_id : int
            unique number representing the algorithm
        kernel_params : array_like
            kernel parameters, (empty) array of floats
        """
        if kernel_params is None:
            kernel_params = []
        self._kernel_id = kernel_id
        self._kernel_params = np.array(kernel_params)

    @property
    def id(self):
        return self._kernel_id

    @property
    def params(self):
        return self._kernel_params

    @abstractmethod
    def eval(self, r):
        pass

    @abstractmethod
    def dkernel_dradius_dividedby_radius(self, r):
        pass

    def to_msgpack(self):
        """
        Create msgpack serializable dict representing an instance of this class
        ref: http://msgpack.org/

        Returns
        -------
        msgpack : dict
            msgpack representation

        """

        msgpack = {"id": self._kernel_id,
                   "params": self._kernel_params.astype('>f8', order='C').tobytes()}

        return msgpack


class KernelLS(AbstractKernel):
    """
    Linear spline kernel
    """

    def __init__(self):
        """
        object initializer
        """
        super().__init__(kernel_id=0)

    def eval(self, r):
        """

        Parameters
        ----------
        r : (n_centers, n_samples) ndarray
            Array of radii

        Returns
        -------
        phi : ndarray
            Array of kernel function values

        """
        return r

    def dkernel_dradius_dividedby_radius(self, r):
        """

        Evaluates the derivative of the kernel value with respect to radius r,
        divided by the radius r.

        Parameters
        ----------
        r : (n_centers, n_samples) ndarray
            Array of radii


        Returns
        -------
        res : (n_centers, n_samples) ndarray

        """
        # raise Exception("not implemented yet!")
        return 1 / r


class KernelCS(AbstractKernel):
    """
    Cubic spline kernel
    """

    def __init__(self):
        """
        object initializer
        """
        super().__init__(kernel_id=1)

    def eval(self, r):
        """

        Parameters
        ----------
        r : (n_centers, n_samples) ndarray
            Array of radii

        Returns
        -------
        phi : ndarray
            Array of kernel function values

        """
        return r ** 3.

    def dkernel_dradius_dividedby_radius(self, r):
        """

        Evaluates the derivative of the kernel value with respect to radius r,
        divided by the radius r.


        Parameters
        ----------
        r : (n_centers, n_samples) ndarray
            Array of radii


        Returns
        -------
        res : (n_centers, n_samples) ndarray

        """
        return r * 3.


class KernelTPS(AbstractKernel):
    """
    thin plate spline kernel r^2*log(r)
    """

    def __init__(self):
        """
        object initializer
        """
        super().__init__(kernel_id=2)

    def eval(self, r):
        """
        evaluate thin plate spline kernel r^2*log(r)
        equivalent formulation 'r*np.log(r**r)' avoids algebraic issues at r = [0...1]

        Parameters
        ----------
        r : (n_centers, n_samples) ndarray
            Array of radii

        Returns
        -------
        phi : ndarray
            Array of kernel function values

        """
        r = np.nan_to_num(r)

        ret = np.zeros_like(r)
        mask = r < 1.

        r1 = r[np.where(mask)]
        ret[np.where(mask)] += r1 * np.log(r1 ** r1)

        mask = np.invert(mask)
        r2 = r[np.where(mask)]
        ret[np.where(mask)] += r2 ** 2. * np.log(r2)

        return ret

    def dkernel_dradius_dividedby_radius(self, r):
        """

        Evaluates the derivative of the kernel value with respect to radius r,
        divided by the radius r.

        Parameters
        ----------
        r : (n_centers, n_samples) ndarray
            Array of radii


        Returns
        -------
        res : (n_centers, n_samples) ndarray

        """
        r = np.nan_to_num(r)
        resolution = np.finfo('d').resolution
        r[r < resolution] = resolution
        ret = 1. + 2. * np.log(r)
        # ret[ret == -np.inf] = 0.

        return ret


class KernelMQ(AbstractKernel):
    """
    Generalized Multiquadrics kernel phi = (1+(p1*r)^2)**p2
    """

    def __init__(self, p1, p2):
        """
        object initializer
        """
        super().__init__(kernel_id=3, kernel_params=[p1, p2])

    def eval(self, r):
        """

        Parameters
        ----------
        r : (n_centers, n_samples) ndarray
            Array of radii

        Returns
        -------
        phi : ndarray
            Array of kernel function values

        """

        p1 = self._kernel_params[0]
        p2 = self._kernel_params[1]

        return (1. + (p1 * r) ** 2.) ** p2

    def dkernel_dradius_dividedby_radius(self, r):
        """

        Evaluates the derivative of the kernel value with respect to radius r,
        divided by the radius r.

        Parameters
        ----------
        r : (n_centers, n_samples) ndarray
            Array of radii


        Returns
        -------
        res : (n_centers, n_samples) ndarray

        """
        # raise Exception("not implemented yet!")
        p1 = self._kernel_params[0]
        p2 = self._kernel_params[1]

        return 2 * p1 ** 2 * p2 * (p1 ** 2 * r ** 2 + 1) ** (p2 - 1)


class KernelGAUSS(AbstractKernel):
    """
    Generalized Multiquadrics kernel phi = exp(-(p1*r)**2)
    """

    def __init__(self, p1):
        """
        object initializer
        """
        super().__init__(kernel_id=4, kernel_params=[p1])

    def eval(self, r):
        """

        Parameters
        ----------
        r : (n_centers, n_samples) ndarray
            Array of radii

        Returns
        -------
        phi : ndarray
            Array of kernel function values

        """
        p1 = self._kernel_params[0]
        return np.exp(-(p1 * r) ** 2)

    def dkernel_dradius_dividedby_radius(self, r):
        """

        Evaluates the derivative of the kernel value with respect to radius r,
        divided by the radius r.

        Parameters
        ----------
        r : (n_centers, n_samples) ndarray
            Array of radii


        Returns
        -------
        res : (n_centers, n_samples) ndarray

        """
        # raise Exception("not implemented yet!")

        p1 = self._kernel_params[0]

        return -2 * p1 ** 2 * np.exp(-(p1 * r) ** 2)
