/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9715 $
 * $Date: 2020-03-19 10:01:15 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.symspace.center.views.indices;

import java.awt.AlphaComposite;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.GradientPaint;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.SystemColor;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.awt.dnd.DnDConstants;
import java.awt.dnd.DragGestureEvent;
import java.awt.dnd.DragGestureListener;
import java.awt.dnd.DragSource;
import java.awt.image.BufferedImage;

import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.TransferHandler;

/**
 * 
 * @author Markus Priglinger
 *
 */
public class DnDButton extends JButton
		implements Transferable, DragGestureListener {

	private static final long serialVersionUID = 19700717201309131L;

	/** The type of DnD object being dragged... */
	public static final DataFlavor DNDBUTTON_FLAVOR = new DataFlavor(
			Object.class, "DnDButton");

	/** use own DataFlavor */
	private DataFlavor[] _flavors = { DnDButton.DNDBUTTON_FLAVOR };

	private DragSource source;
	private TransferHandler t;

	private int _index = -1;

	public DnDButton() {

		// The TransferHandler returns a new DnDButton
		// to be transferred in the Drag
		this.t = new TransferHandler() {

			/**
			 * 
			 */
			private static final long serialVersionUID = 19700717201309132L;

			@Override
			public Transferable createTransferable(JComponent c) {
				return DnDButton.this;
				// return new DnDButton(getText());
			}
		};
		this.setTransferHandler(this.t);

		// The Drag will copy the DnDButton rather than moving it
		this.source = new DragSource();
		this.source.createDefaultDragGestureRecognizer(this,
				DnDConstants.ACTION_MOVE, this);
	}

	// The DataFlavor is a marker to let the DropTarget know how to
	// handle the Transferable
	@Override
	public DataFlavor[] getTransferDataFlavors() {

		return this._flavors;
	}

	@Override
	public boolean isDataFlavorSupported(DataFlavor flavor) {

		return java.util.Arrays.asList(this._flavors).contains(flavor);
	}

	@Override
	public Object getTransferData(DataFlavor flavor)
			throws UnsupportedFlavorException {

		if (flavor.isMimeTypeEqual(DnDButton.DNDBUTTON_FLAVOR.getMimeType())) {
			return this;
		} else {
			throw new UnsupportedFlavorException(flavor);
		}
	}

	// when a DragGesture is recognized, initiate the Drag
	@Override
	public void dragGestureRecognized(DragGestureEvent dge) {

		JPanel pnl = new JPanel();
		pnl.setLayout(new BoxLayout(pnl, BoxLayout.X_AXIS));

		JButton buttonToDraw = new JButton();
		buttonToDraw.setIcon(this.getIcon());

		buttonToDraw.setBorder(BorderFactory.createLineBorder(Color.GRAY, 1));
		pnl.add(buttonToDraw);
		pnl.add(Box.createRigidArea(new Dimension(24, 24)));
		pnl.add(Box.createVerticalGlue());
		pnl.add(new JLabel("Index #" + this.getIndex()));

		// we need a frame to render the component onto
		JFrame picFrame = new JFrame();
		picFrame.add(pnl);
		picFrame.pack(); // this calculates the bounding

		// // Get a buffered image of the selection for dragging a ghost image
		BufferedImage _imgGhost = new BufferedImage(pnl.getWidth(),
				pnl.getHeight(), BufferedImage.TYPE_INT_ARGB);

		Graphics2D g2 = _imgGhost.createGraphics();
		g2.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC, 0.8f));
		// Make the image ghostlike

		g2.setPaint(new GradientPaint(0, 0, SystemColor.controlShadow,
				pnl.getWidth(), pnl.getHeight(), new Color(0, 255, 255, 0)));

		pnl.paint(g2);
		g2.dispose();

		picFrame.dispose();

		this.source.startDrag(dge, DragSource.DefaultMoveDrop, _imgGhost,
				new Point(12, 12), this, null);
		// this.source.startDrag(dge, DragSource.DefaultMoveDrop, _imgGhost,
		// new Point(12, 12), this, new DragSourceListener() {
		//
		// @Override
		// public void dropActionChanged(DragSourceDragEvent dsde) {
		// }
		//
		// @Override
		// public void dragOver(DragSourceDragEvent dsde) {
		// }
		//
		// @Override
		// public void dragExit(DragSourceEvent dse) {
		// }
		//
		// @Override
		// public void dragEnter(DragSourceDragEvent dsde) {
		// }
		//
		// @Override
		// public void dragDropEnd(DragSourceDropEvent dsde) {
		// }
		// });
	}

	public void setIndex(int index) {

		this._index = index;
	}

	public int getIndex() {

		return this._index;
	}
}
