/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9714 $
 * $Date: 2020-03-19 09:37:13 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.magopt.docu;

import java.io.File;

import at.lcm.bu21.general.utils.FileUtils;
import at.lcm.bu21.general.utils.TexWriter;
import at.lcm.bu21.magopt.structure.DocFile;

/**
 * 
 * 
 * @author Werner Koppelsttter
 */
public class DocTexWriter extends TexWriter {
	/**
	 * Create DocTexWriter
	 */
	protected DocTexWriter() {
		super();
	}

	/**
	 * Create DocTexWriter for given file
	 * 
	 * @param file
	 *            directory for writer
	 */
	public DocTexWriter(String file) {
		super(file);
	}

	/**
	 * Create DocTexWriter for given file
	 * 
	 * @param file
	 *            directory for writer
	 */
	public DocTexWriter(File file) {
		super(file);
	}

	/**
	 * Add file to content
	 * 
	 * @param file
	 *            DocFile to add
	 * @return Status of insert graphic (false if file not found)
	 */
	public boolean addGraphic(DocFile file) {
		return this.addGraphic(file, null);
	}

	/**
	 * Add file to content
	 * 
	 * @param file
	 *            DocFile to add
	 * @param width
	 *            Width of figure (unit cm)
	 * @return Status of insert graphic (false if file not found)
	 */
	public boolean addGraphic(DocFile file, double width) {
		return this.addGraphic(file, "width=" + width + "cm");
	}

	/**
	 * Add file to content
	 * 
	 * @param file
	 *            DocFile to add
	 * @param property
	 *            Property of graphic
	 * @return Status of insert graphic (false if file not found)
	 */
	public boolean addGraphic(DocFile file, String property) {
		boolean status = false;
		if (file != null) {
			status = this.addGraphic(FileUtils.getRelativePath(this
					.getDirectoryFile(), FileUtils.getFile(this.getDirectory(),
					"..", file.getFile())), property);
		} else {
			this.add("\\missingfigure{Undefined DocFile} ");
		}

		return status;
	}

	/**
	 * Add tex-table to content
	 * 
	 * @param file
	 *            DocFile to add
	 * @param caption
	 *            Caption of table
	 * @param label
	 *            Label of Table
	 * @return Status of insert table (false if file not found)
	 */
	public boolean addTable(DocFile file, String caption, String label) {
		boolean status = false;
		if (file != null) {
			status = this.addTable(FileUtils.getRelativePath(this
					.getDirectoryFile(), FileUtils.getFile(this.getDirectory(),
					"..", file.getFile())), caption, label);
		} else {
			this.add("\\missingfigure{Undefined DocFile}");
		}

		return status;
	}

	/**
	 * Add tex-file to content
	 * 
	 * @param file
	 *            DocFile to add
	 * @return Status of insert tex (false if file not found)
	 */
	public boolean addTex(DocFile file) {
		boolean status = false;
		if (file != null) {
			status = this.addTex(FileUtils.getRelativePath(this
					.getDirectoryFile(), FileUtils.getFile(this.getDirectory(),
					"..", file.getFile())));
		} else {
			this.add("\\missingfigure{Undefined DocFile}");
		}

		return status;
	}

	/**
	 * Replaces each '\MOREPLACE' substring with parameter that matches the
	 * target string with replacement string of graphic
	 * 
	 * @param target
	 *            String to be replaced
	 * @param file
	 *            Replacement file
	 * @return Status of replacement
	 */
	public boolean replaceGraphic(String target, DocFile file) {
		return this.replaceGraphic(target, file, null);
	}

	/**
	 * Replaces each '\MOREPLACE' substring with parameter that matches the
	 * target string with replacement string of graphic
	 * 
	 * @param target
	 *            String to be replaced
	 * @param file
	 *            Replacement file
	 * @param width
	 *            Width of graphic (unit cm)
	 * @return Status of replacement
	 */
	public boolean replaceGraphic(String target, DocFile file, double width) {
		return this.replaceGraphic(target, file, "width=" + width + "cm");
	}

	/**
	 * Replaces each '\MOREPLACE' substring with parameter that matches the
	 * target string with replacement string of graphic
	 * 
	 * @param target
	 *            String to be replaced
	 * @param file
	 *            Replacement file
	 * @param property
	 *            Property of graphic
	 * @return Status of replacement
	 */
	public boolean replaceGraphic(String target, DocFile file, String property) {
		boolean status;
		DocTexWriter content;

		content = new DocTexWriter(this.getFile());
		status = content.addGraphic(file, property);
		this.replaceString(target, content.toString());

		return status;
	}

}
