/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 1012 $
 * $Date: 2020-04-16 10:07:11 +0200 (Do., 16 Apr 2020) $
 */
package at.lcm.bu21.magopt.structure;

import at.lcm.bu21.general.dtypes.TData;
import at.lcm.bu21.general.dtypes.TString;
import at.lcm.bu21.general.dtypes.TStruct;
import at.lcm.bu21.general.dtypes.unit.ConversionException;
import at.lcm.bu21.general.dtypes.unit.UnitConversion;

/**
 * Collector-structure for access to description, unit and value
 *
 * @author Werner Koppelsttter
 */
public class CollectorStructure {

	/**
	 * Exception of CollectorStructure
	 * 
	 *
	 * @author Werner Koppelsttter
	 */
	public class CollectorStructureException extends Exception {

		/** Serialization version number */
		private static final long serialVersionUID = 1L;

		/**
		 * Constructor of exception
		 * 
		 * @param message
		 *            Exception message
		 */
		public CollectorStructureException(String message) {
			super(message);
		}
	}

	/** Description keyword */
	protected static final String DESCRIPTION_KW = "description__";

	/** Unit keyword */
	protected static final String UNIT_KW = "unit__";

	/** Value keyword */
	protected static final String VALUE_KW = "value__";

	/** Collect structure */
	TStruct struct;

	/**
	 * Create CollectorStructure
	 */
	public CollectorStructure() {
		this(null);
	}

	/**
	 * Create CollectorStructure
	 * 
	 * @param struct
	 *            Structure of collection
	 */
	public CollectorStructure(TStruct struct) {
		if (struct != null) {
			this.struct = struct;
		} else {
			this.struct = new TStruct();
		}

	}

	/**
	 * Get selected data
	 * 
	 * @param identifier
	 *            Identifier of structure
	 * @return Collect-Structure
	 * @throws CollectorStructureException
	 */
	public CollectorStructure getData(String identifier) throws CollectorStructureException {
		TStruct struct;
		TData data;
		CollectorStructure collect = null;

		if ((identifier == null) || (identifier.isEmpty())) {
			collect = this;
		} else {
			data = this.struct.getData(identifier);
			if ((data != null) && (data instanceof TStruct)) {
				struct = (TStruct) data;
			} else {
				struct = null;
			}

			if (struct != null) {
				collect = new CollectorStructure(struct);
			}
		}

		if (collect == null) {
			throw new CollectorStructureException("Parameter of identifer doesn't exist");
		}

		return collect;
	}

	/**
	 * Add data-field to collection
	 * 
	 * @param identifier
	 *            Identifier of data to select
	 * @param name
	 *            Name of new field
	 * @return Created CollectorStructure
	 * @throws CollectorStructureException
	 */
	public CollectorStructure addData(String identifier, String name) throws CollectorStructureException {
		return this.addData(identifier, name, null);
	}

	/**
	 * 
	 * @param identifier
	 *            Identifier of data to select
	 * @param name
	 *            Name of new field
	 * @param collectStruct
	 *            Data for new field
	 * @return Created CollectorStructure
	 * @throws CollectorStructureException
	 */
	public CollectorStructure addData(String identifier, String name, CollectorStructure collectStruct)
			throws CollectorStructureException {
		CollectorStructure collect;
		TStruct s;

		collect = this.getData(identifier);
		if (collectStruct != null) {
			s = collectStruct.struct;
		} else {
			s = new TStruct();
		}
		collect.struct.put(name, s);

		return collect.getData(name);
	}

	/**
	 * Remove field from structure
	 * 
	 * @param identifier
	 *            Identifier of data to select
	 * @param name
	 *            Name of field
	 * @throws CollectorStructureException
	 */
	public void rmData(String identifier, String name) throws CollectorStructureException {
		CollectorStructure collect;

		collect = this.getData(identifier);
		collect.struct.remove(name);
	}

	/**
	 * Get description of selected data
	 * 
	 * @param identifier
	 *            Identifier of data to select
	 * @return Unit of data
	 * @throws CollectorStructureException
	 */
	public String getDescription(String identifier) throws CollectorStructureException {
		CollectorStructure collect;
		TData data;
		String desc = null;

		collect = this.getData(identifier);
		data = collect.struct.get(CollectorStructure.DESCRIPTION_KW);
		if ((data != null) && (data instanceof TString)) {
			desc = ((TString) data).getString();
		}

		return desc;
	}

	/**
	 * Set description of selected data
	 * 
	 * @param identifier
	 *            Identifier of data to select
	 * @param description
	 *            Description of data
	 * @throws CollectorStructureException
	 */
	public void setDescription(String identifier, String description) throws CollectorStructureException {
		CollectorStructure collect;

		collect = this.getData(identifier);
		collect.struct.put(CollectorStructure.DESCRIPTION_KW, new TString(description));
	}

	/**
	 * Get value of selected data
	 * 
	 * @param identifier
	 *            Identifier of data to select
	 * @return Value of data
	 * @throws CollectorStructureException
	 */
	public TData getValue(String identifier) throws CollectorStructureException {
		CollectorStructure collect;
		TData data = null;

		collect = this.getData(identifier);
		data = collect.struct.get(CollectorStructure.VALUE_KW);

		return data;
	}

	/**
	 * Get value of selected data
	 * 
	 * @param identifier
	 *            Identifier of data to select
	 * @param unit
	 *            Unit of the wanted value
	 * @return Value of data
	 * @throws ConversionException
	 * @throws CollectorStructureException
	 */
	public TData getValue(String identifier, String unit) throws ConversionException, CollectorStructureException {
		CollectorStructure collect;
		TData value = null;
		String sUnit = null;
		UnitConversion uc;

		collect = this.getData(identifier);
		value = collect.getValue(null);
		if (value != null) {
			if (unit == null) {
				unit = "";
			}
			sUnit = collect.getUnit(null);
			if (sUnit == null) {
				sUnit = "";
			}

			if (!(unit.isEmpty() && sUnit.isEmpty())) {
				if (!unit.equals(sUnit)) {
					uc = new UnitConversion(sUnit, unit);
					value = uc.convert(value);
				}
			}
		}

		return value;
	}

	/**
	 * Set value of selected data
	 * 
	 * @param identifier
	 *            Identifier of data to select
	 * @param value
	 *            Value to set
	 * @throws CollectorStructureException
	 */
	public void setValue(String identifier, TData value) throws CollectorStructureException {
		CollectorStructure collect;

		collect = this.getData(identifier);
		collect.struct.put(CollectorStructure.VALUE_KW, value);
	}

	/**
	 * Set value of selected data
	 * 
	 * @param identifier
	 *            Identifier of data to select
	 * @param value
	 *            Value to set
	 * @param unit
	 *            Unit of the wanted value
	 * @throws ConversionException
	 * @throws CollectorStructureException
	 */
	public void setValue(String identifier, TData value, String unit)
			throws ConversionException, CollectorStructureException {
		CollectorStructure collect;
		String dUnit = null;
		UnitConversion uc;

		collect = this.getData(identifier);
		if (value != null) {
			if (unit == null) {
				unit = "";
			}
			dUnit = collect.getUnit(null);
			if (dUnit == null) {
				dUnit = "";
			}

			if (!(unit.isEmpty() && dUnit.isEmpty())) {
				if (!unit.equals(dUnit)) {
					uc = new UnitConversion(unit, dUnit);
					value = uc.convert(value);
				}
			}
		}
		collect.setValue(null, value);
	}

	/**
	 * Get unit of selected data
	 * 
	 * @param identifier
	 *            Identifier of data to select
	 * @return Unit of data
	 * @throws CollectorStructureException
	 */
	public String getUnit(String identifier) throws CollectorStructureException {
		CollectorStructure collect;
		TData data;
		String unit = null;

		collect = this.getData(identifier);
		data = collect.struct.get(CollectorStructure.UNIT_KW);
		if ((data != null) && (data instanceof TString)) {
			unit = ((TString) data).getString();
		}

		return unit;
	}

	/**
	 * Set unit of selected data
	 * 
	 * @param identifier
	 *            Identifier of data to select
	 * @param unit
	 *            Unit to set
	 * @throws CollectorStructureException
	 */
	public void setUnit(String identifier, String unit) throws CollectorStructureException {
		CollectorStructure collect;
		String oldUnit;
		TData tmp;

		collect = this.getData(identifier);
		tmp = collect.struct.getUnit(null);
		if ((tmp != null) && (tmp instanceof TString)) {
			oldUnit = ((TString) tmp).getString();
		} else {
			oldUnit = "";
		}

		collect.struct.put(CollectorStructure.UNIT_KW, new TString(unit));
		try {
			collect.setValue(null, collect.getValue(null), oldUnit);
		} catch (ConversionException e) {
		}
	}

}
