/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9713 $
 * $Date: 2020-03-19 09:28:58 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.magopt.structure;

import java.io.File;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Hashtable;

import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLStreamWriter;

import at.lcm.bu21.general.dtypes.TData;
import at.lcm.bu21.general.utils.Clone;
import at.lcm.bu21.general.utils.CloneException;
import at.lcm.bu21.general.utils.FileUtils;
import at.lcm.bu21.general.utils.Pair;
import at.lcm.bu21.magopt.structure.event.EventLog;

/**
 * Dataset data
 * 
 * @author Werner Koppelsttter
 */
public class Dataset implements Serializable {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** Dataset-identifier */
	private String dsId;

	/** Comment of dataset */
	private String comment = "";

	/** Parameter entries of dataset */
	private Hashtable<String, ParamSet> parameter;

	/** EventLog of dataset */
	private EventLog log = new EventLog();

	/**
	 * Create new dataset with identifier
	 * 
	 * @param dsId
	 *            Dataset-identifier
	 */
	public Dataset(String dsId) {
		this.parameter = new Hashtable<String, ParamSet>();
		this.dsId = dsId;
	}

	/**
	 * Create a deep clone of given parameter
	 * 
	 * @return Deep clone of parameter
	 * @throws CloneException
	 */
	public Dataset deepClone() throws CloneException {
		return (Dataset) Clone.deepClone(this);
	}

	/**
	 * Check if selected datasets are equal
	 * 
	 * @param obj
	 *            Dataset to check if equal
	 * @param link
	 *            List of linked parameter-identifier
	 * @return If datasets are equal
	 */
	public boolean equals(Object obj, ArrayList<Pair<String, String>> link) {
		ParamSet psThis, psOther;

		if (this == obj) {
			return true;
		}
		if (obj == null) {
			return false;
		}
		if (this.getClass() != obj.getClass()) {
			return false;
		}
		Dataset other = (Dataset) obj;
		if (this.comment == null) {
			if (other.comment != null) {
				return false;
			}
		} else if (!this.comment.equals(other.comment)) {
			return false;
		}
		if (this.parameter == null) {
			if (other.parameter != null) {
				return false;
			}
		} else {
			for (Pair<String, String> pair : link) {
				psThis = this.parameter.get(pair.getFirst());
				psOther = other.parameter.get(pair.getSecond());

				if (psThis == null) {
					if (psOther != null) {
						return false;
					}
				} else if (!psThis.equals(psOther)) {
					return false;
				}
			}
		}

		return true;
	}

	/**
	 * Get dataset-identifier
	 * 
	 * @return Dataset-identifier
	 */
	String getDsId() {
		return this.dsId;
	}

	/**
	 * Set dataset-identifier
	 * 
	 * @param dsId
	 *            Dataset-identifier
	 */
	void setDsId(String dsId) {
		this.dsId = dsId;
	}

	/**
	 * Get dataset-comment
	 * 
	 * @return Comment
	 */
	String getComment() {
		return this.comment;
	}

	/**
	 * Set dataset-comment
	 * 
	 * @param comment
	 *            Comment
	 */
	void setComment(String comment) {
		this.comment = comment;
	}

	/**
	 * Get event-log of dataset
	 * 
	 * @return Event-log
	 */
	EventLog getEventLog() {
		return this.log;
	}

	/**
	 * Set event-log of dataset
	 * 
	 * @param log
	 *            Event-log
	 */
	void setEventLog(EventLog log) {
		this.log = log;
	}

	/**
	 * Get value of ParamSet
	 * 
	 * @param pId
	 *            Parameter-identifier
	 * @param key
	 *            Key of value
	 * @return Value of ParamSet
	 */
	TData getValue(String pId, ParamSetKey key) {
		ParamSet param;
		TData value = null;

		param = this.parameter.get(pId);
		if (param != null) {
			value = param.getData(key);
		}

		return value;
	}

	/**
	 * Set value of ParamSet
	 * 
	 * @param pId
	 *            Parameter-identifier
	 * @param key
	 *            Key of value
	 * @param value
	 *            Value to set
	 */
	void setValue(String pId, ParamSetKey key, TData value) {
		ParamSet param;

		param = this.parameter.get(pId);
		if (param == null) {
			this.addParameter(pId);
			param = this.parameter.get(pId);
		}
		param.setData(key, value);
	}

	/**
	 * Remove value from ParamSet
	 * 
	 * @param pId
	 *            Parameter-identifier
	 * @param key
	 *            Key of value
	 * @return Removed data
	 */
	TData rmValue(String pId, ParamSetKey key) {
		TData data = null;
		ParamSet param;

		param = this.parameter.get(pId);
		if (param != null) {
			data = param.rmData(key);
		}

		return data;
	}

	/**
	 * Add Parameter-set for identifier
	 * 
	 * @param pId
	 *            Parameter-identifier
	 */
	private void addParameter(String pId) {
		this.parameter.put(pId, new ParamSet());
	}

	/**
	 * Remove parameter-set of identifier
	 * 
	 * @param pId
	 *            Parameter-identifier
	 */
	void rmParameter(String pId) {
		this.parameter.remove(pId);
	}

	/**
	 * Get temporary-data
	 * 
	 * @param pId
	 *            Parameter-identifier
	 * @param key
	 *            Key of data
	 * @return Value of temporary-data
	 */
	Object getTmpData(String pId, String key) {
		ParamSet param;
		Object obj = null;

		param = this.parameter.get(pId);
		if (param != null) {
			obj = param.getTmpData(key);
		}

		return obj;
	}

	/**
	 * Set temporary-data
	 * 
	 * @param pId
	 *            Parameter-identifier
	 * @param key
	 *            Key of data
	 * @param obj
	 *            Value to set
	 */
	void setTmpData(String pId, String key, Object obj) {
		ParamSet param;

		param = this.parameter.get(pId);
		if (param == null) {
			this.addParameter(pId);
			param = this.parameter.get(pId);
		}
		param.setTmpData(key, obj);
	}

	/**
	 * Remove temporary-data
	 * 
	 * @param pId
	 *            Parameter-identifier
	 * @param key
	 *            Key of data
	 * @return Removed data
	 */
	Object rmTmpData(String pId, String key) {
		Object obj = null;
		ParamSet param;

		param = this.parameter.get(pId);
		if (param != null) {
			obj = param.rmTmpData(key);
		}

		return obj;
	}

	/**
	 * Clear temporary-data
	 * 
	 * @param pId
	 *            Parameter-identifier
	 */
	void clearTmpData(String pId) {
		ParamSet param;

		param = this.parameter.get(pId);
		if (param != null) {
			param.clearTmpData();
		}
	}

	/**
	 * Get DocFile
	 * 
	 * @param pId
	 *            Parameter-identifier
	 * @param name
	 *            Name of DocFile
	 * @return DocFile
	 */
	DocFile getDocFile(String pId, String name) {
		ParamSet param;
		DocFile doc = null;

		param = this.parameter.get(pId);
		if (param != null) {
			doc = param.getDocFile(name);
		}

		return doc;
	}

	/**
	 * Set DocFile
	 * 
	 * @param pId
	 *            Parameter-identifier
	 * @param doc
	 *            DocFile to set
	 */
	void setDocFile(String pId, DocFile doc) {
		ParamSet param;

		param = this.parameter.get(pId);
		if (param == null) {
			this.addParameter(pId);
			param = this.parameter.get(pId);
		}
		param.setDocFile(doc);
	}

	/**
	 * Get list of assigned DocFiles
	 * 
	 * @param pId
	 *            Parameter-identifier
	 * @return List of DocFiles
	 */
	DocFile[] getDocFiles(String pId) {
		ParamSet param;
		DocFile[] docs = new DocFile[0];

		param = this.parameter.get(pId);
		if (param != null) {
			docs = param.getDocFiles();
		}

		return docs;
	}

	/**
	 * Reset DocFiles (set checksum to 0)
	 * 
	 * @param pId
	 *            Parameter-identifier
	 */
	void resetDocFiles(String pId) {
		ParamSet param;

		param = this.parameter.get(pId);
		if (param != null) {
			param.resetDocFiles();
		}
	}

	/**
	 * Unregister assigned DocFiles
	 * 
	 * @param pId
	 *            Parameter-identifier
	 */
	void unregisterDocFiles(String pId) {
		ParamSet param;

		param = this.parameter.get(pId);
		if (param != null) {
			param.unregisterDocFiles();
		}
	}

	/**
	 * Unregister one DocFile
	 * 
	 * @param pId
	 *            Parameter-identifier
	 * @param name
	 *            Name of DocFile
	 */
	void unregisterDocFile(String pId, String name) {
		ParamSet param;

		param = this.parameter.get(pId);
		if (param != null) {
			param.unregisterDocFile(name);
		}
	}

	/**
	 * Read object from xml-file
	 * 
	 * @param xmlReader
	 *            XMLReader to read xml-file
	 * @throws Exception
	 */
	void xml2obj(XMLStreamReader xmlReader) throws Exception {
		String element, key;
		ParamSet param;

		// xmlReader.next();
		while (xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT) {
			element = xmlReader.getLocalName();
			if (element.equals("DSID")) {
				xmlReader.next();
				this.dsId = new String();
				while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
					this.dsId += xmlReader.getText();
					xmlReader.next();
				}
				xmlReader.next();
			} else if (element.equals("Comment")) {
				xmlReader.next();
				this.comment = new String();
				while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
					this.comment += xmlReader.getText();
					xmlReader.next();
				}
				xmlReader.next();
			} else if (element.equals("ParamData")) {
				xmlReader.next();
				while (xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT) {
					key = xmlReader.getAttributeValue(null, "pid");
					xmlReader.next();
					param = new ParamSet();
					param.xml2obj(xmlReader);
					this.parameter.put(key, param);
					xmlReader.next();
				}
				xmlReader.next();
			} else if (element.equals("EventLog")) {
				xmlReader.next();
				this.log.xml2obj(xmlReader);
				xmlReader.next();
			} else {
				throw new Exception("Unrecognized property: "
						+ xmlReader.getLocalName());
			}
		}
		// xmlReader.next();
	}

	/**
	 * Write object to xml-file
	 * 
	 * @param xmlWriter
	 *            XMLWriter to write xml-file
	 * @throws Exception
	 */
	void obj2xml(XMLStreamWriter xmlWriter) throws Exception {
		String[] keys;
		// xmlWriter.writeStartElement("Dataset");

		// Dataset-Identifier
		xmlWriter.writeStartElement("DSID");
		xmlWriter.writeCharacters(this.dsId);
		xmlWriter.writeEndElement();

		// Comment
		xmlWriter.writeStartElement("Comment");
		if (this.comment != null) {
			xmlWriter.writeCharacters(this.comment);
		}
		xmlWriter.writeEndElement();

		// Parameter
		xmlWriter.writeStartElement("ParamData");

		keys = this.parameter.keySet().toArray(new String[0]);
		Arrays.sort(keys);
		for (String pid : keys) {
			xmlWriter.writeStartElement("ParamSet");
			xmlWriter.writeAttribute("pid", pid);
			this.parameter.get(pid).obj2xml(xmlWriter);
			xmlWriter.writeEndElement();
		}
		xmlWriter.writeEndElement();

		// Event-log
		xmlWriter.writeStartElement("EventLog");
		this.log.obj2xml(xmlWriter);
		xmlWriter.writeEndElement();

		// xmlWriter.writeEndElement();
	}

	/**
	 * Cleanup dataset
	 * 
	 * @param proj
	 *            Project structure
	 * @param directory
	 *            Direcotry of dataset
	 */
	void cleanup(Project proj, File directory) {
		File dir;
		ArrayList<String> docs;
		ParamSet ps;
		String name;
		String[] keys;
		Parameter p;

		// paramset
		keys = this.parameter.keySet().toArray(new String[0]);
		for (String pid : keys) {
			p = proj.getParamByPid(pid);
			if (p == null) {
				this.parameter.remove(pid);
			} else {
				this.parameter.get(pid).cleanup(p);
			}
		}

		// Parameter-directory
		dir = new File(directory, "param");
		if (dir.exists()) {
			for (File f : dir.listFiles()) {
				if (!(f.isDirectory() && (this.parameter.get(f.getName()) != null))) {
					FileUtils.delete(f);
				}
			}
		}

		// DocFiles
		dir = new File(directory, "docFiles");
		if (dir.exists()) {
			// list all docfiles
			docs = new ArrayList<String>();
			for (String pid : this.parameter.keySet()) {
				ps = this.parameter.get(pid);
				for (DocFile doc : ps.getDocFiles()) {
					name = new File(dir, doc.getFile()).getName();
					docs.add(name);
					docs.add(FileUtils.splitName(name)[0]);
				}
			}

			// check file is docfile
			for (File f : dir.listFiles()) {
				if ((f.getName().compareTo("index.html") != 0)
						&& (f.getName().compareTo("LibContents_ACCM.css") != 0)
						&& !docs.contains(f.getName())
						&& !docs.contains(FileUtils.splitName(f.getName())[0])) {
					FileUtils.delete(f);
				}
			}
		}
	}
}
