/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9713 $
 * $Date: 2020-03-19 09:28:58 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.magopt.structure;

import java.io.File;
import java.io.IOException;
import java.io.Serializable;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.security.NoSuchAlgorithmException;
import java.util.List;

import at.lcm.bu21.general.dtypes.TData;

/**
 * Interface of parameter
 * 
 * @author Werner Koppelsttter
 */
public interface IParameter extends Remote, Serializable, Cloneable {

	/**
	 * Get ID of parameter
	 * 
	 * @return Parameter ID
	 * @throws RemoteException
	 */
	public String getPId() throws RemoteException;

	/**
	 * Get name of parameter
	 * 
	 * @return Name
	 * @throws RemoteException
	 */
	public String getName() throws RemoteException;

	/**
	 * Set name of parameter and change in Dependency and Interface
	 * 
	 * @param name
	 *            Name
	 * @throws Exception
	 * @throws RemoteException
	 */
	public void setName(String name) throws RemoteException, Exception;

	/**
	 * Get description of parameter
	 * 
	 * @return description
	 * @throws RemoteException
	 */
	public String getDescription() throws RemoteException;

	/**
	 * Set description of parameter
	 * 
	 * @param description
	 *            Description
	 * @throws RemoteException
	 */
	public void setDescription(String description) throws RemoteException;

	/**
	 * Get comment of parameter
	 * 
	 * @return comment
	 * @throws RemoteException
	 */
	public String getComment() throws RemoteException;

	/**
	 * Set comment of parameter
	 * 
	 * @param comment
	 *            Comment
	 * @throws RemoteException
	 */
	public void setComment(String comment) throws RemoteException;

	/**
	 * Get identifier of this parameter
	 * 
	 * @return Identifier
	 * @throws RemoteException
	 */
	public String getIdentifier() throws RemoteException;

	/**
	 * Get short identifier of this parameter that is unique
	 * 
	 * @return Identifier
	 * @throws RemoteException
	 */
	public String getShortIdentifier() throws RemoteException;

	/**
	 * Get current dataset index
	 * 
	 * @return Index
	 * @throws RemoteException
	 */
	public int getIndex() throws RemoteException;

	/**
	 * Set current dataset index
	 * 
	 * @param index
	 *            Index
	 * @throws RemoteException
	 */
	public void setIndex(int index) throws RemoteException;

	/**
	 * Get size of datasets
	 * 
	 * @return size of dataset
	 * @throws RemoteException
	 */
	public int getLength() throws RemoteException;

	/**
	 * Get data of Parameter
	 * 
	 * @return Data
	 * @throws RemoteException
	 */
	public TData getData() throws RemoteException;

	/**
	 * Get data of Parameter
	 * 
	 * @param index
	 *            Index of dataset
	 * @return Data
	 * @throws RemoteException
	 */
	public TData getData(int index) throws RemoteException;

	/**
	 * Set data of parameter
	 * 
	 * @param data
	 *            Data
	 * @throws Exception
	 * @throws RemoteException
	 */
	public void setData(TData data) throws RemoteException, Exception;

	/**
	 * Get print-flag
	 * 
	 * @return Print-flag
	 * @throws RemoteException
	 */
	public boolean getPrint() throws RemoteException;

	/**
	 * Set print-flag
	 * 
	 * @param print
	 *            Print-flag
	 * @throws RemoteException
	 */
	public void setPrint(boolean print) throws RemoteException;

	/**
	 * Get constant-flag
	 * 
	 * @return Constant-flag
	 * @throws RemoteException
	 */
	public boolean getConstant() throws RemoteException;

	/**
	 * Set Constant-flag
	 * 
	 * @param constant
	 *            Constant-flag
	 * @throws RemoteException
	 */
	public void setConstant(boolean constant) throws RemoteException;

	/**
	 * Get used-status
	 * 
	 * @return Used-status
	 * @throws RemoteException
	 */
	public boolean getUsed() throws RemoteException;

	/**
	 * Get read-only-status
	 * 
	 * @return Read-only-status
	 * @throws RemoteException
	 */
	public boolean getReadOnly() throws RemoteException;

	/**
	 * Get modified-status
	 * 
	 * @return modified-flag
	 * @throws RemoteException
	 */
	public boolean getModified() throws RemoteException;

	/**
	 * Set modified-status
	 * 
	 * @param modified
	 *            modified-flag
	 * @throws RemoteException
	 */
	public void setModified(boolean modified) throws RemoteException;

	/**
	 * Get list of parameter which use this parameter
	 * 
	 * @return List of parameter
	 * @throws RemoteException
	 */
	public IParameter[] getUsedBy() throws RemoteException;

	/**
	 * Get list of parameter which export to this parameter
	 * 
	 * @return List of parameter
	 * @throws RemoteException
	 */
	public IParameter[] getSetBy() throws RemoteException;

	/**
	 * Remove this parameter from parent parameter
	 * 
	 * @return removed parameter
	 * @throws RemoteException
	 */
	public IParameter removeParam() throws RemoteException;

	/**
	 * Move this parameter up in list
	 * 
	 * @throws RemoteException
	 */
	public void moveUp() throws RemoteException;

	/**
	 * Move this parameter down in list
	 * 
	 * @throws RemoteException
	 */
	public void moveDown() throws RemoteException;

	/**
	 * Get index of parameter in param-list of parent
	 * 
	 * @return Index of parameter
	 * @throws RemoteException
	 */
	public int getIndexOfParameter() throws RemoteException;

	/**
	 * Get uncertain-status
	 * 
	 * @return uncertain-flag
	 * @throws RemoteException
	 */
	public boolean getUncertain() throws RemoteException;

	/**
	 * Get error-status of current dataset
	 * 
	 * @return error-flag
	 * @throws RemoteException
	 */
	public boolean getError() throws RemoteException;

	/**
	 * Get a list of error-flags of all datasets
	 * 
	 * @return list of error-flags
	 * @throws RemoteException
	 */
	public boolean[] getAllError() throws RemoteException;

	/**
	 * Set error-status of current dataset
	 * 
	 * @param error
	 *            error-flag
	 * @throws RemoteException
	 */
	public void setError(boolean error) throws RemoteException;

	/**
	 * Refresh structure and flags
	 * 
	 * @throws RemoteException
	 */
	public void refresh() throws RemoteException;

	/**
	 * Get parent-parameter
	 * 
	 * @return Parent
	 * @throws RemoteException
	 */
	public IParamContainer getParentParameter() throws RemoteException;

	/**
	 * Get root-parameter
	 * 
	 * @return Root
	 * @throws RemoteException
	 */
	public IParameter getRoot() throws RemoteException;

	/**
	 * Get object which unique identifier
	 * 
	 * @param uniqueIdentifier
	 *            Unique identifier
	 * @return Found parameter
	 * @throws RemoteException
	 */
	public IParameter getUniqueObject(String uniqueIdentifier) throws RemoteException;

	/**
	 * Get first object which identifier correlate with given identifier
	 * 
	 * @param identifier
	 *            Identifier
	 * @return Found parameter
	 * @throws RemoteException
	 */
	public IParameter getObject(String identifier) throws RemoteException;

	/**
	 * Get parameter by parameter-identifier
	 * 
	 * @param pid
	 *            Parameter-identifier
	 * @return Selected parameter
	 * @throws RemoteException
	 */
	public IParameter getParamByPid(String pid) throws RemoteException;

	/**
	 * Get list of Parameter that are instances of class
	 * 
	 * @param c
	 *            Name of class
	 * @return List of parameter
	 * @throws RemoteException
	 */
	public IParameter[] getParameters(String c) throws RemoteException;

	/**
	 * Get list of analysis parameter. Analysis parameters are ParamField which are
	 * used
	 * 
	 * @return List of parameter
	 * @throws RemoteException
	 */
	public IParameter[] getAnalysisParams() throws RemoteException;

	/**
	 * Get list of dynamic parameters. Dynamic parameter are parameter which can
	 * diversify. ParamField which are used, not constant, writable and has one
	 * numeric value
	 * 
	 * @return List of parameter
	 * @throws RemoteException
	 */
	public IParameter[] getDynamicParams() throws RemoteException;

	/**
	 * Add information message to event-log
	 * 
	 * @param message
	 *            Message
	 * @throws RemoteException
	 */
	public void info(String message) throws RemoteException;

	/**
	 * Add information message to event-log
	 * 
	 * @param message
	 *            Message
	 * @param stack
	 *            Stack-trace
	 * @throws RemoteException
	 */
	public void info(String message, StackTraceElement[] stack) throws RemoteException;

	/**
	 * Add information message to event-log
	 * 
	 * @param identifier
	 *            Identifier
	 * @param message
	 *            Message
	 * @param stack
	 *            Stack-trace
	 * @throws RemoteException
	 */
	public void info(String identifier, String message, StackTraceElement[] stack) throws RemoteException;

	/**
	 * Add warning message to event-log
	 * 
	 * @param message
	 *            Message
	 * @throws RemoteException
	 */
	public void warning(String message) throws RemoteException;

	/**
	 * Add warning message to event-log
	 * 
	 * @param message
	 *            Message
	 * @param stack
	 *            Stack-trace
	 * @throws RemoteException
	 */
	public void warning(String message, StackTraceElement[] stack) throws RemoteException;

	/**
	 * Add warning message to event-log
	 * 
	 * @param identifier
	 *            Identifier
	 * @param message
	 *            Message
	 * @param stack
	 *            Stack-trace
	 * @throws RemoteException
	 */
	public void warning(String identifier, String message, StackTraceElement[] stack) throws RemoteException;

	/**
	 * Add error message to event-log
	 * 
	 * @param message
	 *            Message
	 * @throws RemoteException
	 */
	public void error(String message) throws RemoteException;

	/**
	 * Add error message to event-log
	 * 
	 * @param message
	 *            Message
	 * @param stack
	 *            Stack-trace
	 * @throws RemoteException
	 */
	public void error(String message, StackTraceElement[] stack) throws RemoteException;

	/**
	 * Add error message to event-log
	 * 
	 * @param identifier
	 *            Identifier
	 * @param message
	 *            Message
	 * @param stack
	 *            Stack-trace
	 * @throws RemoteException
	 */
	public void error(String identifier, String message, StackTraceElement[] stack) throws RemoteException;

	/**
	 * Get unique doc-file name in directory. Compare without extension
	 * 
	 * @param file
	 *            File to start
	 * @return Unique file
	 * @throws RemoteException
	 */
	public String getUniqueDocFileName(String file) throws RemoteException;

	/**
	 * Get unique doc-file name in directory. Compare without extension
	 * 
	 * @param file
	 *            File to start
	 * @return Unique file
	 * @throws RemoteException
	 */
	public File getUniqueDocFileName(File file) throws RemoteException;

	/**
	 * Get doc-files
	 * 
	 * @return DocFiles
	 * @throws RemoteException
	 */
	public DocFile[] getDocFiles() throws RemoteException;

	/**
	 * Get doc-files with the given name
	 * 
	 * @param name
	 *            Name of file
	 * @return DocFile
	 * @throws RemoteException
	 */
	public DocFile getDocFile(String name) throws RemoteException;

	/**
	 * Get file from doc-file
	 * 
	 * @param doc
	 *            DocFile
	 * @return File of DocFile
	 * @throws RemoteException
	 */
	public File getFileFromDocFile(DocFile doc) throws RemoteException;

	/**
	 * Register doc file
	 * 
	 * @param name
	 *            Name to register
	 * @param strFile
	 *            File to register
	 * @param description
	 *            Description
	 * @throws IOException
	 * @throws NoSuchAlgorithmException
	 * @throws RemoteException
	 */
	public void registerDocFile(String name, String strFile, String description)
			throws RemoteException, NoSuchAlgorithmException, IOException;

	/**
	 * Unregister all doc files
	 * 
	 * @throws RemoteException
	 */
	public void unregisterDocFiles() throws RemoteException;

	/**
	 * Unregister specified doc file
	 * 
	 * @param name
	 *            Name of docfile to remove
	 * @throws RemoteException
	 */
	public void unregisterDocFile(String name) throws RemoteException;

	/**
	 * Reset DocFiles (set files to invalid)
	 * 
	 * @throws RemoteException
	 */
	public void resetDocFiles() throws RemoteException;

	/**
	 * Reset DocFiles (set files to invalid)
	 * 
	 * @param index
	 *            Index to reset
	 * @throws RemoteException
	 */
	public void resetDocFiles(int index) throws RemoteException;

	/**
	 * Delete registered doc-files
	 * 
	 * @throws RemoteException
	 */
	public void deleteDocFiles() throws RemoteException;

	/**
	 * Delete registered doc-files
	 * 
	 * @param index
	 *            Index of dataset
	 * @throws RemoteException
	 */
	public void deleteDocFiles(int index) throws RemoteException;

	/**
	 * Check if all files are valid (exist, checksum)
	 * 
	 * @return Files valid
	 * @throws RemoteException
	 */
	public boolean checkDocFiles() throws RemoteException;

	/**
	 * Check if all files are valid (exist, checksum)
	 * 
	 * @param index
	 *            Index to reset
	 * @return Files valid
	 * @throws RemoteException
	 */
	public boolean checkDocFiles(int index) throws RemoteException;

	/**
	 * Check if all files are valid (exist, checksum)
	 * 
	 * @param docfile
	 *            Name of DocFile
	 * 
	 * @return Files valid
	 * @throws RemoteException
	 */
	public boolean checkDocFile(String docfile) throws RemoteException;

	/**
	 * Check if all files are valid (exist, checksum)
	 * 
	 * @param index
	 *            Index to reset
	 * @param docfile
	 *            Name of docfile
	 * @return Files valid
	 * @throws RemoteException
	 */
	public boolean checkDocFile(int index, String docfile) throws RemoteException;

	/**
	 * Get identifier of datasets
	 * 
	 * @return Dataset-identifier
	 * @throws RemoteException
	 */
	public String getDsId() throws RemoteException;

	/**
	 * Get dataset-directory of parameter
	 * 
	 * @return Parameter-directory
	 * @throws RemoteException
	 */
	public File getParamDirectory() throws RemoteException;

	/**
	 * To string for interface
	 * 
	 * @return String representation of string
	 * @throws RemoteException
	 */
	public String toStringI() throws RemoteException;

	/**
	 * List name of tabs
	 * 
	 * @return List of tab-names
	 * @throws RemoteException
	 */
	public List<String> listTabNames() throws RemoteException;

	// public IParameter cloneParam() throws RemoteException, Exception;
	//
	// public IParameter cloneParam(IParameter clone, int index)
	// throws RemoteException, Exception;
	//
	// public IParameter cloneParam(IParameter clone, int index, boolean
	// docFiles)
	// throws RemoteException;
	//
	// public IParameter deepClone() throws RemoteException, CloneException;

	/**
	 * Get client-implementation of interface
	 * 
	 * @return Client-implementation
	 * @throws RemoteException
	 */
	public IParameter getClientImpl() throws RemoteException;
}
