/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9713 $
 * $Date: 2020-03-19 09:28:58 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.magopt.structure;

import java.io.File;
import java.io.IOException;
import java.rmi.RemoteException;
import java.util.Map;
import java.util.concurrent.ExecutionException;

import at.lcm.bu21.general.utils.CloneException;
import at.lcm.bu21.magopt.structure.event.IEventLog;
import magopt.utils.MopFile;

/**
 * Interface of project
 * 
 * @author Werner Koppelsttter
 */
public interface IProject extends IParamContainer {

	/**
	 * Close connections of project
	 * 
	 * @throws RemoteException
	 */
	public void close() throws RemoteException;

	/**
	 * Is Project write protected. (If optimization exist)
	 * 
	 * @return Is write-protected
	 * @throws RemoteException
	 */
	public boolean isWriteProtected() throws RemoteException;

	/**
	 * Save Project to project-file
	 * 
	 * @throws Exception
	 * @throws RemoteException
	 */
	public void save() throws RemoteException, Exception;

	/**
	 * Save Project to given project-file
	 * 
	 * @param file
	 *            Project-file
	 * @throws Exception
	 * @throws RemoteException
	 */
	public void saveAs(String file) throws RemoteException, Exception;

	/**
	 * Save Project to given project-file
	 * 
	 * @param file
	 *            Project-file
	 * @param override
	 *            Override existing project
	 * @throws Exception
	 * @throws RemoteException
	 */
	public void saveAs(String file, boolean override) throws RemoteException, Exception;

	/**
	 * Save Project to given project-file
	 * 
	 * @param file
	 *            Project-file
	 * @throws Exception
	 * @throws RemoteException
	 */
	public void saveAs(File file) throws RemoteException, Exception;

	/**
	 * Save Project to given project-file
	 * 
	 * @param file
	 *            Project-file
	 * @param override
	 *            Override existing project
	 * @throws Exception
	 * @throws RemoteException
	 */
	public void saveAs(File file, boolean override) throws RemoteException, Exception;

	/**
	 * Get comment of current dataset
	 * 
	 * @return Comment
	 * @throws RemoteException
	 */
	public String getDSComment() throws RemoteException;

	/**
	 * Set comment of current dataset
	 * 
	 * @param comment
	 *            Comment
	 * @throws RemoteException
	 */
	public void setDSComment(String comment) throws RemoteException;

	/**
	 * Get event-log of current dataset
	 * 
	 * @return Event-log
	 * @throws RemoteException
	 */
	public IEventLog getEventLog() throws RemoteException;

	/**
	 * Get project-file
	 * 
	 * @return Project-file
	 * @throws RemoteException
	 */
	public String getFile() throws RemoteException;

	/**
	 * Get report-parameters
	 * 
	 * @return Report-parameter
	 * @throws RemoteException
	 */
	public IReportParameter getReportParameter() throws RemoteException;

	/**
	 * Get gui-handle from matlab
	 * 
	 * @return gui-handle
	 * @throws RemoteException
	 */
	// FIXME RMI
	public double getGuiHandle() throws RemoteException;

	/**
	 * Set gui-handle from matlab
	 * 
	 * @param guiHandle
	 *            gui-handle
	 * @throws RemoteException
	 */
	// FIXME RMI
	public void setGuiHandle(double guiHandle) throws RemoteException;

	/**
	 * Get comment-list of datasets
	 * 
	 * @return Comments
	 * @throws RemoteException
	 */
	// FIXME RMI
	public String[] getAllDSComment() throws RemoteException;

	/**
	 * Get project-file class
	 * 
	 * @return Project-file
	 * @throws RemoteException
	 */
	public MopFile getMopFile() throws RemoteException;

	/**
	 * Check structure which updates are required
	 * 
	 * @throws RemoteException
	 */
	public void check() throws RemoteException;

	/**
	 * Duplicate current dataset to the end of list
	 * 
	 * @throws CloneException
	 * @throws IOException
	 * @throws RemoteException
	 */
	public void duplicate() throws RemoteException, CloneException, IOException;

	/**
	 * Reduce current dataset of list
	 * 
	 * @throws RemoteException
	 */
	public void reduce() throws RemoteException;

	/**
	 * Regenerate project and update required functions
	 * 
	 * @param args
	 *            Run parameter
	 * @return Status of regeneration (0 .. finished successfully)
	 * @throws RemoteException
	 */
	public int regenerate(Map<String, Object> args) throws RemoteException;

	/**
	 * Get files directory of project
	 * 
	 * @return Project-directory
	 * @throws RemoteException
	 */
	public File getAllFilesDirectory() throws RemoteException;

	/**
	 * Get directory of dataset-files
	 * 
	 * @return Dataset-directory
	 * @throws RemoteException
	 */
	public File getFilesDirectory() throws RemoteException;

	/**
	 * Reset all functions in project
	 * 
	 * @throws RemoteException
	 */
	public void resetAllFunctions() throws RemoteException;

	/**
	 * Extract index to new project
	 * 
	 * @param index
	 *            Index to extract
	 * @param file
	 *            Target file of project
	 * @return New Project
	 * @throws Exception
	 * @throws RemoteException
	 */
	public IProject extractIndex(int index, File file) throws Exception, RemoteException;

	/**
	 * Restart python connection
	 * 
	 * @throws InterruptedException
	 * @throws ExecutionException
	 * @throws RemoteException
	 */
	public void restartJep() throws InterruptedException, ExecutionException, RemoteException;

	@Override
	public IProject getClientImpl() throws RemoteException;
}
