/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9713 $
 * $Date: 2020-03-19 09:28:58 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.magopt.structure;

import java.io.File;
import java.io.IOException;
import java.lang.reflect.Method;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.rmi.RemoteException;
import java.util.Hashtable;
import java.util.Map;

import javax.swing.JPanel;
import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLStreamWriter;

import at.lcm.bu21.general.utils.Clone;
import at.lcm.bu21.general.utils.CloneException;
import at.lcm.bu21.general.utils.FileUtils;
import at.lcm.bu21.magopt.docu.DocTexWriter;
import at.lcm.bu21.magopt.structure.client.CParamJavaFunction;
import at.lcm.bu21.magopt.structure.rmi.RParamJavaFunction;

/**
 * Parameter for user-defined java-function
 * 
 * @author Werner Koppelsttter
 */
public class ParamJavaFunction extends ParamFunction implements IParamJavaFunction {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/**
	 * Get empty map
	 * 
	 * @return Empty map
	 */
	public static Map<String, Object> getEmptyMap() {
		return new Hashtable<String, Object>();
	}

	/**
	 * Separate class-path and class-name
	 * 
	 * @param file
	 *            Class-file
	 * @return Array with class-path and class-name
	 */
	public static String[] separatePathName(String file) {
		String classPath = "";
		String className = "";
		int index;
		// boolean found = false;

		if (file != null) {
			classPath = file;
			if (classPath.toLowerCase().endsWith(".class")) {
				classPath = file.substring(0, classPath.length() - 6);
			}

			index = classPath.lastIndexOf(File.separator + "bin" + File.separator);
			if (index == -1) {
				className = classPath;
				classPath = "";
			} else {
				className = classPath.substring(index + 5).replace(File.separatorChar, '.');
				classPath = classPath.substring(0, index + 4);
			}
		}
		// separating path and binary-name
		// while (!found && (classPath.length() != 0)) {
		//
		// index = classPath.lastIndexOf(File.separator);
		// if (index == -1) {
		// if (className.isEmpty()) {
		// className = classPath;
		// } else {
		// className = classPath + "." + className;
		// }
		// classPath = "";
		// } else {
		// if (className.isEmpty()) {
		// className = classPath.substring(index + 1);
		// } else {
		// className = classPath.substring(index + 1) + "."
		// + className;
		// }
		// classPath = classPath.substring(0, index);
		// }
		//
		// try {
		// // FIXME URL form System Classloader
		// URL[] url = {
		// new URL("file:\\" + classPath + File.separator),
		// new URL("file:\\" + Project.getInstallPath()
		// + "System\\Java\\magopt.jar") };
		// URLClassLoader cl = new URLClassLoader(url);
		// cl.loadClass(className);
		// found = true;
		// } catch (Throwable ex) {
		// }
		// }

		// if (!found) {
		// classPath = className;
		// index = classPath.lastIndexOf(File.separator);
		// if (index == -1) {
		// className = classPath;
		// classPath = "";
		// } else {
		// className = classPath.substring(index + 1);
		// classPath = classPath.substring(0, index);
		// }
		// }
		return new String[] { classPath, className };
	}

	/**
	 * Constructor for loading a xml-file
	 * 
	 * @throws Exception
	 */
	ParamJavaFunction() throws Exception {
		this("XML_Load", "XML load constructor");
	}

	/**
	 * Constructor with name
	 * 
	 * @param name
	 *            Name of Parameter
	 * @throws Exception
	 */
	public ParamJavaFunction(String name) throws Exception {
		this(name, name);
	}

	/**
	 * Constructor with name and description
	 * 
	 * @param name
	 *            Name of Parameter
	 * @param description
	 *            Description of Parameter
	 * @throws Exception
	 */
	public ParamJavaFunction(String name, String description) throws Exception {
		this(name, description, null, null);
	}

	/**
	 * Constructor with name and description
	 * 
	 * @param name
	 *            Name of Parameter
	 * @param description
	 *            Description of Parameter
	 * @param file
	 *            File of function
	 * @throws Exception
	 */
	public ParamJavaFunction(String name, String description, String file) throws Exception {
		this(name, description, file, null);
	}

	/**
	 * Constructor with name and description
	 * 
	 * @param name
	 *            Name of Parameter
	 * @param description
	 *            Description of Parameter
	 * @param classPath
	 *            Class-path of Function
	 * @param className
	 *            Class-name of Function
	 * @throws Exception
	 */
	public ParamJavaFunction(String name, String description, String classPath, String className) throws Exception {
		super(name, description);
		if ((className == null) || className.isEmpty()) {
			String[] parts = ParamJavaFunction.separatePathName(classPath);
			this.setClassName(parts[1]);
			this.setClassPath(parts[0]);
		} else {
			this.setClassName(className);
			this.setClassPath(classPath);
		}
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (!super.equals(obj)) {
			return false;
		}
		if (this.getClass() != obj.getClass()) {
			return false;
		}
		ParamJavaFunction other = (ParamJavaFunction) obj;
		if (this.classLocation != other.classLocation) {
			return false;
		}
		if (this.classPath == null) {
			if (other.classPath != null) {
				return false;
			}
		} else if (!this.classPath.equals(other.classPath)) {
			return false;
		}
		if (this.className == null) {
			if (other.className != null) {
				return false;
			}
		} else if (!this.className.equals(other.className)) {
			return false;
		}
		return true;
	}

	@Override
	public ParamJavaFunction deepClone() throws CloneException {
		return (ParamJavaFunction) Clone.deepClone(this);
	}

	@Override
	void cloneParam(Parameter clone) throws CloneException, InstantiationException, IllegalAccessException {

		super.cloneParam(clone);

		if (clone instanceof ParamJavaFunction) {
			ParamJavaFunction other = (ParamJavaFunction) clone;

			other.className = this.className;
			other.classPath = this.classPath;
			other.classLocation = this.classLocation;
		}
	}

	@Override
	public String getFunctionID() {
		String id = "";

		try {
			id = this.getFunctionObject().getFunctionID();
		} catch (Exception e) {
			this.warning(e.getMessage(), e.getStackTrace());
		}

		return id;
	}

	@Override
	public void refreshDependency() {
		ParamJavaFnc jf;

		try {
			jf = this.getFunctionObject();
			if ((jf.getImports().length == 0) && (jf.getExports().length == 0)) {
				jf.setImportVars(jf.importVars());
				jf.setExportVars(jf.exportVars());
			}
			jf.refreshDependency();
		} catch (Exception e) {
			this.error(e.getMessage(), e.getStackTrace());
		}
	}

	@Override
	public String getClassName() {
		return this.className;
	}

	@Override
	public void setClassName(String className) {
		if (className == null) {
			className = "";
		}
		this.className = className;
	}

	@Override
	public String getLocation() {
		return this.getLocation(this.classLocation);
	}

	/**
	 * Get location-string of class
	 * 
	 * @return Location
	 */
	public String getLocationString() {
		return this.getLocationString(this.classLocation);
	}

	@Override
	public String getLibraryPackage() {
		String path = null;
		if (this.classLocation == Location.LIBRARY) {
			path = new File(this.classPath).getParentFile().getParentFile().getParent();
		}
		return path;
	}

	@Override
	void refreshClass() {
		String path = null;
		try {
			path = (new File(
					this.getFunctionObject().getClass().getProtectionDomain().getCodeSource().getLocation().getPath()))
							.getPath();
		} catch (Exception e) {
		}
		if ((path != null) && (path.length() != 0)) {
			this.setClassPath(path);
		}
	}

	/**
	 * Refresh class-path
	 */
	void refreshPath() {
		String path = this.getClassPath();
		if (path.contains(":")) {
			this.setClassPath(this.getClassPath());
		}
	}

	private Class<? extends ParamJavaFnc> getFunctionClass() throws ClassNotFoundException, MalformedURLException {
		Class<? extends ParamJavaFnc> c = null;
		URLClassLoader cl;

//		cl = new ChildFirstURLClassLoader(new URL[] { new File(this.getClassPath()).toURI().toURL() },
//				this.getClass().getClassLoader());
		cl = new URLClassLoader(new URL[] { new File(this.getClassPath()).toURI().toURL() },
				this.getClass().getClassLoader());
		c = (Class<? extends ParamJavaFnc>) cl.loadClass(this.className);

		return c;
	}

	public boolean isFunctionClassInstanceOf(Class<? extends Object> cFnc) {
		String cName;
		Class<? extends Object> c;

		cName = cFnc.getName();
		try {
			c = this.getFunctionClass();
			while ((c != null) && !cName.equals(c.getName())) {
				c = c.getSuperclass();
			}
		} catch (Exception e) {
			c = null;
		}

		return c != null;
	}

	/**
	 * Get function object
	 * 
	 * @return Function object
	 * @throws Exception
	 */
	public ParamJavaFnc getFunctionObject() throws Exception {
		ParamJavaFnc obj = null;
		try {
			// XXX
			// if ((this.className != null) && (this.className.length() != 0)) {
			obj = this.getFunctionClass().getConstructor().newInstance();
			obj.structure = this;
			// }
		} catch (Exception e) {
			throw new Exception(this.getName() + ": Class '" + this.className + "' not found!");
		}
		return obj;
	}

	@Override
	public String toString() {
		StringBuffer str = new StringBuffer();
		str.append("ParamJavaFunction\n");
		str.append("\n\tName          : ");
		str.append(this.getName());
		str.append("\n\tDescription   : ");
		str.append(this.getDescription());
		str.append("\n\tClassNamen    : ");
		str.append(this.getClassName());
		str.append("\n\tClassPath     : ");
		str.append(this.classPath);
		str.append("\n\tClassLocation : ");
		str.append(this.getLocationString());
		str.append("\n\tParams:       ");

		for (Parameter p : this.getParam()) {
			str.append("\n\t\tName: " + p.getName());
		}

		return str.toString();
	}

	@Override
	public int update(Map<String, Object> args) throws Exception {
		int status = 0;

		status = this.getFunctionObject().run(args);

		this.update(status);
		return status;
	}

	@Override
	void xml2obj(XMLStreamReader xmlReader) throws Exception {
		String element, tmp;
		// Integer length = 1;

		xmlReader.next();
		while (xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT) {
			element = xmlReader.getLocalName();
			if (element.equals("ParamFunction")) {
				super.xml2obj(xmlReader);
				// length = this.getLength();
			} else if (element.equals("ClassName")) {
				xmlReader.next();
				this.className = new String();
				while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
					this.className += xmlReader.getText();
					xmlReader.next();
				}
				xmlReader.next();
			} else if (element.equals("ClassPath")) {
				xmlReader.next();
				this.classPath = new String();
				while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
					this.classPath += xmlReader.getText();
					xmlReader.next();
				}
				xmlReader.next();
			} else if (element.equals("ClassLocation")) {
				xmlReader.next();
				tmp = new String();
				while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
					tmp += xmlReader.getText();
					xmlReader.next();
				}
				this.classLocation = Location.valueOf(tmp);
				xmlReader.next();
			} else {
				throw new Exception("Unrecognized property: " + xmlReader.getLocalName());
			}
		}
		xmlReader.next();

	}

	@Override
	void obj2xml(XMLStreamWriter xmlWriter) throws Exception {

		xmlWriter.writeStartElement("ParamJavaFunction");
		xmlWriter.writeAttribute("object", this.getClass().getSimpleName());
		xmlWriter.writeAttribute("name", this.getName());

		this.refreshPath();

		// super class
		super.obj2xml(xmlWriter);

		// Classname
		xmlWriter.writeStartElement("ClassName");
		if (this.className != null) {
			xmlWriter.writeCharacters(this.className);
		}
		xmlWriter.writeEndElement();

		// Classpath
		xmlWriter.writeStartElement("ClassPath");
		if (this.classPath != null) {
			xmlWriter.writeCharacters(this.classPath);
		}
		xmlWriter.writeEndElement();

		// ClassLocation
		xmlWriter.writeStartElement("ClassLocation");
		xmlWriter.writeCharacters(this.classLocation.name());
		xmlWriter.writeEndElement();

		xmlWriter.writeEndElement();
	}

	@Override
	public void reset() {
		if (!this.getUpdateRequired()) {
			this.setUpdateRequired(true);
		}
		try {
			this.getFunctionObject().reset();
			this.setResetFlag(false);
		} catch (Exception e) {
			this.error(this.getIdentifier(), e.getMessage(), e.getStackTrace());
		}
	}

	@Override
	void reportInfo(DocTexWriter content) {

		super.reportInfo(content);

		content.add("\tJava function & ");
		content.addString((this.getLocationString() + "#" + this.getClassPath() + "->" + this.getClassName())
				.replace(File.separatorChar, '/'));
		content.add(" \\\\\n");
	}

	@Override
	protected void generateDocFiles(File directory) throws Exception {
		DocTexWriter content;
		File texFile, file;
		String name;

		super.generateDocFiles(directory);
		// Generate DocFiles
		this.getFunctionObject().generateDocFiles(directory);

		// Generate Template
		file = this.getUniqueDocFileName(new File(directory, "Template.tex"));
		content = new DocTexWriter(file);
		name = this.getClassName().replace(".", File.separator);
		texFile = FileUtils.getFile(new File(this.getClassPath()).getParent(), "tex", name + "_Template",
				new File(name).getName() + ".tex");
		if (texFile.exists()) {
			try {
				content.readBody(texFile);
			} catch (IOException e) {
				this.warning("Cannot open tex-template-file");
			}
		}

		try {
			content = this.getFunctionObject().reportTex(content);
		} catch (Exception e) {
			this.warning("Cannot replace tex code in template");
		}

		if (content.length() > 0) {
			content.write();
			this.registerDocFile("Template", file.getPath(), "Template");
		}

	}

	@Override
	public IParamJavaFunction getRMIImpl() throws RemoteException {
		return new RParamJavaFunction(this);
	}

	@Override
	public IParamJavaFunction getClientImpl() {
		return new CParamJavaFunction(this);
	}

	@Override
	public boolean hasPreview() {
		boolean exist = false;

		if (this.getClassName() != "") {
			try {
				this.getFunctionClass().getMethod("preview", JPanel.class);
				exist = true;
			} catch (Exception e) {
			}
		}

		return exist;
	}

	@Override
	public void preview(JPanel panel) {
		ParamJavaFnc obj;
		Method method;

		if (this.hasPreview()) {
			try {
				obj = this.getFunctionObject();
				method = obj.getClass().getMethod("preview", JPanel.class);
				method.invoke(obj, panel);
			} catch (Exception e) {
				System.err.println(e.getMessage());
				e.printStackTrace();
			}
		}
	}

	@Override
	public boolean hasShow() {
		Class<?> cl;
		boolean exist = false;

		if (this.getClassName() != "") {
			try {
				this.getFunctionClass().getMethod("show");
				exist = true;
			} catch (Exception e) {
			}
		}

		return exist;
	}

	@Override
	public void show() {
		ParamJavaFnc obj;
		Method method;

		if (this.hasShow()) {
			try {
				obj = this.getFunctionObject();
				method = obj.getClass().getMethod("show");
				method.invoke(obj);
			} catch (Exception e) {
				e.printStackTrace();
			}

		}
	}
}
