/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9713 $
 * $Date: 2020-03-19 09:28:58 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.magopt.structure;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Hashtable;
import java.util.List;

import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLStreamWriter;

import at.lcm.bu21.general.dtypes.TData;

/**
 * Set of values for a parameter
 * 
 * @author Werner Koppelsttter
 */
class ParamSet implements Serializable {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** Data */
	private Hashtable<ParamSetKey, TData> data = new Hashtable<ParamSetKey, TData>();

	/** DocFiles */
	private ArrayList<DocFile> docFiles = new ArrayList<DocFile>();

	/** Temporary data */
	private transient Hashtable<String, Object> tmpData = null;

	@Override
	public boolean equals(Object obj) {
		TData valueThis, valueOther;
		if (this == obj) {
			return true;
		}
		if (obj == null) {
			return false;
		}
		if (this.getClass() != obj.getClass()) {
			return false;
		}
		ParamSet other = (ParamSet) obj;
		if (this.data == null) {
			if (other.data != null) {
				return false;
			}
		} else {
			for (ParamSetKey key : ParamSetKey.values()) {
				valueThis = this.data.get(key);
				valueOther = other.data.get(key);
				if (valueThis == null) {
					if (valueOther != null) {
						return false;
					}
				} else if (!valueThis.equals(valueOther)) {
					return false;
				}
			}
		}
		// if (this.docFiles == null) {
		// if (other.docFiles != null) {
		// return false;
		// }
		// } else if (!this.docFiles.equals(other.docFiles)) {
		// return false;
		// }
		return true;
	}

	/**
	 * Get data for the key of this parameter-set
	 * 
	 * @param key
	 *            Key of value
	 * @return Value
	 */
	TData getData(ParamSetKey key) {
		return this.data.get(key);
	}

	/**
	 * Set data for key of this parameter-set
	 * 
	 * @param key
	 *            Key of value
	 * @param value
	 *            Value
	 */
	void setData(ParamSetKey key, TData value) {
		this.data.put(key, value);
	}

	/**
	 * Remove data for the key in this parameter-set
	 * 
	 * @param key
	 *            Key of value
	 * @return Removed data
	 */
	TData rmData(ParamSetKey key) {
		return this.data.remove(key);
	}

	/**
	 * Get a list of all doc-files in this parameter-set
	 * 
	 * @return List of doc-files
	 */
	DocFile[] getDocFiles() {
		return this.docFiles.toArray(new DocFile[0]);
	}

	/**
	 * Get doc-file with the name
	 * 
	 * @param name
	 *            Name of doc-file
	 * @return Doc-file
	 */
	DocFile getDocFile(String name) {
		DocFile file = null;

		for (DocFile f : this.docFiles) {
			if (f.getName().compareTo(name) == 0) {
				file = f;
			}
		}

		return file;
	}

	/**
	 * Set/add doc-files to this parameter-set
	 * 
	 * @param doc
	 *            Doc-file to set
	 */
	void setDocFile(DocFile doc) {

		for (int k = 0; k < this.docFiles.size(); k++) {
			if ((doc != null)
					&& (this.docFiles.get(k).getName().compareTo(doc.getName()) == 0)) {
				this.docFiles.set(k, doc);
				doc = null;
			}
		}

		if (doc != null) {
			this.docFiles.add(doc);
		}
	}

	/**
	 * Reset all doc-files of this parameter-set
	 */
	void resetDocFiles() {
		for (DocFile doc : this.docFiles) {
			doc.setChecksum("");
		}
	}

	/**
	 * Unregister and remove all doc-files of this parameter-set
	 */
	void unregisterDocFiles() {
		this.docFiles.clear();
	}

	/**
	 * Unregister and remove doc-file with the name
	 * 
	 * @param name
	 *            Name of doc-file
	 */
	void unregisterDocFile(String name) {
		DocFile doc = this.getDocFile(name);
		if (doc != null) {
			this.docFiles.remove(doc);
		}
	}

	/**
	 * Get temporary data of parameter-set
	 * 
	 * @param key
	 *            Key of data
	 * @return Data
	 */
	Object getTmpData(String key) {
		Object obj = null;

		if (this.tmpData != null) {
			obj = this.tmpData.get(key);
		}

		return obj;
	}

	/**
	 * Set temporary data of parameter-set
	 * 
	 * @param key
	 *            Key of data
	 * @param obj
	 *            Data
	 */
	void setTmpData(String key, Object obj) {
		if (this.tmpData == null) {
			this.tmpData = new Hashtable<String, Object>();
		}
		this.tmpData.put(key, obj);
	}

	/**
	 * Remove temporary data from parameter-set
	 * 
	 * @param key
	 *            Key of data
	 * @return Removed data
	 */
	Object rmTmpData(String key) {
		Object obj = null;

		if (this.tmpData != null) {
			obj = this.tmpData.remove(key);
		}

		return obj;
	}

	/**
	 * Clear temporary data from parameter-set
	 */
	void clearTmpData() {
		if (this.tmpData != null) {
			this.tmpData.clear();
		}
	}

	/**
	 * Read object from xml-file
	 * 
	 * @param xmlReader
	 *            XMLReader to read xml-file
	 * @throws Exception
	 */
	void xml2obj(XMLStreamReader xmlReader) throws Exception {
		String element, key;
		DocFile doc;

		// xmlReader.next();
		while (xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT) {
			element = xmlReader.getLocalName();
			if (element.equals("Data")) {
				xmlReader.next();
				while (xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT) {
					key = xmlReader.getLocalName();
					xmlReader.next();
					this.data.put(ParamSetKey.valueOf(key),
							TData.xml2obj(xmlReader));
					xmlReader.next();
				}
				xmlReader.next();
			} else if (element.equals("DocFiles")) {
				xmlReader.next();
				while (xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT) {
					doc = new DocFile();
					doc.xml2obj(xmlReader);
					this.docFiles.add(doc);
				}
				xmlReader.next();
			} else {
				throw new Exception("Unrecognized property: "
						+ xmlReader.getLocalName());
			}
		}
		// xmlReader.next();
	}

	/**
	 * Write object to xml-file
	 * 
	 * @param xmlWriter
	 *            XMLWriter to write xml-file
	 * @throws Exception
	 */
	void obj2xml(XMLStreamWriter xmlWriter) throws Exception {
		ParamSetKey[] keys;

		// xmlWriter.writeStartElement("ParamSet");

		// Data
		xmlWriter.writeStartElement("Data");
		keys = this.data.keySet().toArray(new ParamSetKey[0]);
		Arrays.sort(keys);
		for (ParamSetKey entry : keys) {
			if (!entry.isTemporary()) {
				xmlWriter.writeStartElement(entry.name());
				this.data.get(entry).obj2xml(xmlWriter);
				xmlWriter.writeEndElement();
			}
		}
		xmlWriter.writeEndElement();

		// DocFiles
		xmlWriter.writeStartElement("DocFiles");
		for (DocFile entry : this.docFiles) {
			entry.obj2xml(xmlWriter);
		}
		xmlWriter.writeEndElement();

		// xmlWriter.writeEndElement();
	}

	/**
	 * Cleanup paramset
	 * 
	 * @param p
	 *            Parameter to dataset
	 */
	void cleanup(Parameter p) {
		List<ParamSetKey> list;
		ParamSetKey[] keys;

		list = p.listParamSetKey();
		keys = this.data.keySet().toArray(new ParamSetKey[0]);
		for (ParamSetKey key : keys) {
			if (!list.contains(key)) {
				this.data.remove(key);
			}
		}
	}
}
