/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9713 $
 * $Date: 2020-03-19 09:28:58 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.magopt.structure.formula;

import java.util.Arrays;
import java.util.Map;

import at.lcm.bu21.general.dtypes.TArray;
import at.lcm.bu21.general.dtypes.TBoolean;
import at.lcm.bu21.general.dtypes.TData;
import at.lcm.bu21.general.dtypes.TDataException;
import at.lcm.bu21.general.dtypes.TNull;
import at.lcm.bu21.general.dtypes.TNumeric;
import at.lcm.bu21.general.dtypes.TString;
import at.lcm.bu21.general.dtypes.TStruct;

/**
 * Node of syntax-tree to evaluate formulas
 *
 * @author Werner Koppelsttter
 */
public class SyntaxNode {

	/** Token of node */
	Token tok;

	/** Children tokens of node */
	SyntaxNode[] childs = null;

	/**
	 * Get numeric matrix from node (check type)
	 * 
	 * @param node
	 *            Node to evaluate
	 * @param vars
	 *            Map of variables
	 * @return Numeric matrix
	 * 
	 * @throws EvalException
	 * @throws UserException
	 */
	static TNumeric getNumeric(SyntaxNode node, Map<String, TData> vars) throws EvalException, UserException {
		TData d;

		d = node.eval(vars);
		if (d instanceof TNumeric) {
			return (TNumeric) d;
		} else {
			throw new EvalException("Is not numeric type", node.tok);
		}
	}

	/**
	 * Get scalar numeric value from node (check type and size)
	 * 
	 * @param node
	 *            Node to evaluate
	 * @param vars
	 *            Map of variables
	 * @return Scalar numeric
	 * 
	 * @throws EvalException
	 * @throws UserException
	 */
	private static TNumeric getScalar(SyntaxNode node, Map<String, TData> vars) throws EvalException, UserException {
		TNumeric num;

		num = SyntaxNode.getNumeric(node, vars);
		if (num.isScalar()) {
			return num;
		} else {
			throw new EvalException("Is not a scalar value", node.tok);
		}
	}

	/**
	 * Get real numeric value from node (check type and real)
	 * 
	 * @param node
	 *            Node to evaluate
	 * @param vars
	 *            Map of variables
	 * 
	 * @return Real numeric
	 * 
	 * @throws EvalException
	 * @throws UserException
	 */
	static TNumeric getReal(SyntaxNode node, Map<String, TData> vars) throws EvalException, UserException {
		TNumeric num;

		num = SyntaxNode.getNumeric(node, vars);
		if (num.isReal()) {
			return num;
		} else {
			throw new EvalException("Is not a real value", node.tok);
		}
	}

	/**
	 * Get real scalar numeric value from node (check type, size and real)
	 * 
	 * @param node
	 *            Node to evaluate
	 * @param vars
	 *            Map of variables
	 * 
	 * @return Real value
	 * 
	 * @throws EvalException
	 * @throws UserException
	 */
	static double getScalarReal(SyntaxNode node, Map<String, TData> vars) throws EvalException, UserException {
		TNumeric num;

		num = SyntaxNode.getScalar(node, vars);
		if (num.isReal()) {
			return num.getReal(0, 0);
		} else {
			throw new EvalException("Is not a real value", node.tok);
		}
	}

	/**
	 * Get array from node (check type)
	 * 
	 * @param node
	 *            Node to evaluate
	 * @param vars
	 *            Map of variables
	 * 
	 * @return Array
	 * 
	 * @throws EvalException
	 * @throws UserException
	 */
	private static TArray getArray(SyntaxNode node, Map<String, TData> vars) throws EvalException, UserException {
		TData d;

		d = node.eval(vars);
		if (d instanceof TArray) {
			return (TArray) d;
		} else {
			throw new EvalException("Is not array type", node.tok);
		}
	}

	/**
	 * Get struct from node (check type)
	 * 
	 * @param node
	 *            Node to evaluate
	 * @param vars
	 *            Map of variables
	 * 
	 * @return Struct
	 * 
	 * @throws EvalException
	 * @throws UserException
	 */
	private static TStruct getStruct(SyntaxNode node, Map<String, TData> vars) throws EvalException, UserException {
		TData d;

		d = node.eval(vars);
		if (d instanceof TStruct) {
			return (TStruct) d;
		} else {
			throw new EvalException("Is not struct type", node.tok);
		}
	}

	/**
	 * Get boolean from node (check type)
	 * 
	 * @param node
	 *            Node to evaluate
	 * @param vars
	 *            Map of variables
	 * 
	 * @return Boolean value
	 * 
	 * @throws EvalException
	 * @throws UserException
	 */
	private static boolean getBoolean(SyntaxNode node, Map<String, TData> vars) throws EvalException, UserException {
		TData d;

		d = node.eval(vars);
		if (d instanceof TBoolean) {
			return ((TBoolean) d).getBoolean();
		} else {
			throw new EvalException("Is not a boolean type", node.tok);
		}
	}

	/**
	 * Get string from node (check type)
	 * 
	 * @param node
	 *            Node to evaluate
	 * @param vars
	 *            Map of variables
	 * 
	 * @return String value
	 * 
	 * @throws EvalException
	 * @throws UserException
	 */
	static String getString(SyntaxNode node, Map<String, TData> vars) throws EvalException, UserException {
		TData d;

		d = node.eval(vars);
		if (d instanceof TString) {
			return ((TString) d).getString();
		} else {
			throw new EvalException("Is not a string type", node.tok);
		}
	}

	/**
	 * Check bounds (numeric and array
	 * 
	 * @param index
	 *            Index used
	 * @param len
	 *            Length of dimension
	 * @param t
	 *            Token of index
	 * 
	 * @throws EvalException
	 */
	static void checkBounds(int index, int len, Token t) throws EvalException {
		if ((index < 0) || (len <= index)) {
			throw new EvalException("Index out of bounds", t);
		}
	}

	/**
	 * Get indices (start and end)
	 * 
	 * @param node
	 *            Node of index
	 * @param maxIndex
	 *            Maximal possible index
	 * @param vars
	 *            Map of variables
	 * @return EvalException
	 * @throws TDataException
	 * @throws UserException
	 * @throws EvalException
	 */
	private static int[] getIndices(SyntaxNode node, int maxIndex, Map<String, TData> vars)
			throws TDataException, EvalException, UserException {
		int[] indizes = new int[2];

		SyntaxNode.replaceEnd(node, maxIndex);
		if (node.tok.kind == TokenKind.COLON) {
			if (node.childs == null) {
				indizes[0] = 0;
				indizes[1] = maxIndex;
			} else {
				indizes[0] = SyntaxNode.getInt(SyntaxNode.getScalarReal(node.childs[0], vars));
				indizes[1] = SyntaxNode.getInt(SyntaxNode.getScalarReal(node.childs[1], vars));
			}

		} else {
			indizes[0] = indizes[1] = SyntaxNode.getInt(SyntaxNode.getScalarReal(node, vars));
		}

		return indizes;
	}

	/**
	 * Get integer type of double
	 * 
	 * @param val
	 *            Value to cast (round)
	 * @return Integer value
	 */
	static int getInt(double val) {
		return (int) Math.round(val);
	}

	/**
	 * Replace end token to number of maxIndex
	 * 
	 * @param node
	 *            Node to replace recursive
	 * @param maxIndex
	 *            Maximal possible index
	 * @throws TDataException
	 */
	private static void replaceEnd(SyntaxNode node, int maxIndex) throws TDataException {
		switch (node.tok.kind) {
		case END:
			node.tok.kind = TokenKind.NUMBER;
			node.tok.val = new TNumeric(maxIndex);
			break;
		case RBRACK:
		case RBRACE:
		case ARRAY:
		case NUMERIC:
			break;
		default:
			if (node.childs != null) {
				for (SyntaxNode n : node.childs) {
					SyntaxNode.replaceEnd(n, maxIndex);
				}
			}
			break;

		}
	}

	/**
	 * Create node for token
	 * 
	 * @param tok
	 *            Token to represent
	 */
	SyntaxNode(Token tok) {
		this.tok = tok;
	}

	/**
	 * Add child to node
	 * 
	 * @param sn
	 *            Add child to list
	 */
	void addChild(SyntaxNode sn) {
		if (this.childs == null) {
			this.childs = new SyntaxNode[1];
		} else {
			this.childs = Arrays.copyOf(this.childs, this.childs.length + 1);
		}
		this.childs[this.childs.length - 1] = sn;
	}

	/**
	 * Evaluate node
	 * 
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws EvalException
	 * @throws UserException
	 */
	TData eval(Map<String, TData> vars) throws EvalException, UserException {
		TData data;

		switch (this.tok.kind) {
		case IF:
			data = this.if_(vars);
			break;
		case OR:
			data = this.or(vars);
			break;
		case AND:
			data = this.and(vars);
			break;
		case EQL:
			data = this.eql(vars);
			break;
		case NEQ:
			data = this.neq(vars);
			break;
		case GTR:
			data = this.gtr(vars);
			break;
		case GEQ:
			data = this.geq(vars);
			break;
		case LSS:
			data = this.lss(vars);
			break;
		case LEQ:
			data = this.leq(vars);
			break;
		case CALL:
			data = this.call(vars);
			break;
		case PLUS:
			data = this.plus(vars);
			break;
		case MINUS:
			data = this.minus(vars);
			break;
		case TIMES:
			data = this.times(vars);
			break;
		case DOTTIMES:
			data = this.dottimes(vars);
			break;
		case SLASH:
			data = this.slash(vars);
			break;
		case DOTSLASH:
			data = this.dotslash(vars);
			break;
		case PERCENT:
			data = this.percent(vars);
			break;
		case DOTPERCENT:
			data = this.dotpercent(vars);
			break;
		case CARET:
			data = this.caret(vars);
			break;
		case DOTCARET:
			data = this.dotcaret(vars);
			break;
		case ACCENT:
			data = this.accent(vars);
			break;
		case LBRACK:
			data = this.accessNumeric(vars);
			break;
		case LBRACE:
			data = this.accessArray(vars);
			break;
		case VAR:
			data = vars.get(this.tok.str);
			if (data == null) {
				throw new EvalException("Undefined variable " + this.tok.str, this.tok);
			}
			break;
		case PERIOD:
			data = this.struct(vars);
			break;
		case METHOD:
			data = this.method(vars);
			break;
		case NUMERIC:
			data = this.defineNumeric(vars);
			break;
		case ARRAY:
			data = this.defineArray(vars);
			break;
		case IDENT:
			data = new TString(this.tok.str);
			break;
		case NUMBER:
			data = this.tok.val;
			break;
		case STRING:
			data = new TString(this.tok.str);
			break;
		case PI:
			try {
				data = new TNumeric(Math.PI);
			} catch (TDataException e) {
				throw new EvalException(e.getMessage(), this.tok, e.getStackTrace());
			}
			break;
		case E:
			try {
				data = new TNumeric(Math.E);
			} catch (TDataException e) {
				throw new EvalException(e.getMessage(), this.tok, e.getStackTrace());
			}
			break;
		case I:
			try {
				data = new TNumeric(0, 1);
			} catch (TDataException e) {
				throw new EvalException(e.getMessage(), this.tok, e.getStackTrace());
			}
			break;
		case TRUE:
			data = new TBoolean(true);
			break;
		case FALSE:
			data = new TBoolean(false);
			break;
		case NULL:
			data = new TNull();
			break;
		case FUNCTION:
			try {
				data = EvalFunction.function_(this, vars);
			} catch (TDataException e) {
				throw new EvalException(e.getMessage(), this.tok, e.getStackTrace());
			}
			break;
		default:
			throw new EvalException("Undefined token", this.tok);
		}

		return data;
	}

	/**
	 * If-else token
	 * 
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws EvalException
	 * @throws UserException
	 */
	private TData if_(Map<String, TData> vars) throws EvalException, UserException {
		if (SyntaxNode.getBoolean(this.childs[0], vars)) {
			return this.childs[1].eval(vars);
		} else {
			return this.childs[2].eval(vars);
		}
	}

	/**
	 * Or token
	 * 
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws EvalException
	 * @throws UserException
	 */
	private TBoolean or(Map<String, TData> vars) throws EvalException, UserException {
		boolean bool;

		bool = SyntaxNode.getBoolean(this.childs[0], vars);
		bool |= SyntaxNode.getBoolean(this.childs[1], vars);

		return new TBoolean(bool);
	}

	/**
	 * And token
	 * 
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws EvalException
	 * @throws UserException
	 */
	private TBoolean and(Map<String, TData> vars) throws EvalException, UserException {
		boolean bool;

		bool = SyntaxNode.getBoolean(this.childs[0], vars);
		bool &= SyntaxNode.getBoolean(this.childs[1], vars);

		return new TBoolean(bool);
	}

	/**
	 * Equal token
	 * 
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws EvalException
	 * @throws UserException
	 */
	private TBoolean eql(Map<String, TData> vars) throws EvalException, UserException {
		return new TBoolean(
				SyntaxNode.getScalarReal(this.childs[0], vars) == SyntaxNode.getScalarReal(this.childs[1], vars));
	}

	/**
	 * Not equal token
	 * 
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws EvalException
	 * @throws UserException
	 */
	private TBoolean neq(Map<String, TData> vars) throws EvalException, UserException {
		return new TBoolean(
				SyntaxNode.getScalarReal(this.childs[0], vars) != SyntaxNode.getScalarReal(this.childs[1], vars));
	}

	/**
	 * Less token
	 * 
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws EvalException
	 * @throws UserException
	 */
	private TBoolean lss(Map<String, TData> vars) throws EvalException, UserException {
		return new TBoolean(
				SyntaxNode.getScalarReal(this.childs[0], vars) < SyntaxNode.getScalarReal(this.childs[1], vars));
	}

	/**
	 * Less or equal token
	 * 
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws EvalException
	 * @throws UserException
	 */
	private TBoolean leq(Map<String, TData> vars) throws EvalException, UserException {
		return new TBoolean(
				SyntaxNode.getScalarReal(this.childs[0], vars) <= SyntaxNode.getScalarReal(this.childs[1], vars));
	}

	/**
	 * Greater token
	 * 
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws EvalException
	 * @throws UserException
	 */
	private TBoolean gtr(Map<String, TData> vars) throws EvalException, UserException {
		return new TBoolean(
				SyntaxNode.getScalarReal(this.childs[0], vars) > SyntaxNode.getScalarReal(this.childs[1], vars));
	}

	/**
	 * Greater or equal token
	 * 
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws EvalException
	 * @throws UserException
	 */
	private TBoolean geq(Map<String, TData> vars) throws EvalException, UserException {
		return new TBoolean(
				SyntaxNode.getScalarReal(this.childs[0], vars) >= SyntaxNode.getScalarReal(this.childs[1], vars));
	}

	/**
	 * Call token
	 * 
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws EvalException
	 * @throws UserException
	 */
	private TBoolean call(Map<String, TData> vars) throws EvalException, UserException {
		return new TBoolean(!SyntaxNode.getBoolean(this.childs[0], vars));
	}

	/**
	 * Plus token
	 * 
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws EvalException
	 * @throws UserException
	 */
	private TNumeric plus(Map<String, TData> vars) throws EvalException, UserException {

		try {
			return SyntaxNode.getNumeric(this.childs[0], vars).add(SyntaxNode.getNumeric(this.childs[1], vars));
		} catch (TDataException e) {
			throw new EvalException(e.getMessage(), this.tok, e.getStackTrace());
		}
	}

	/**
	 * Minus token
	 * 
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws EvalException
	 * @throws UserException
	 */
	private TNumeric minus(Map<String, TData> vars) throws EvalException, UserException {
		try {
			if (this.childs.length == 1) {
				return SyntaxNode.getNumeric(this.childs[0], vars).multiply(new TNumeric(-1));
			} else {
				return SyntaxNode.getNumeric(this.childs[0], vars)
						.add(SyntaxNode.getNumeric(this.childs[1], vars).multiply(new TNumeric(-1)));
			}
		} catch (TDataException e) {
			throw new EvalException(e.getMessage(), this.tok, e.getStackTrace());
		}
	}

	/**
	 * Times token
	 * 
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws EvalException
	 * @throws UserException
	 */
	private TData times(Map<String, TData> vars) throws EvalException, UserException {
		try {
			return SyntaxNode.getNumeric(this.childs[0], vars).multiply(SyntaxNode.getNumeric(this.childs[1], vars));
		} catch (TDataException e) {
			throw new EvalException(e.getMessage(), this.tok, e.getStackTrace());
		}
	}

	/**
	 * Dot-times token
	 * 
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws EvalException
	 * @throws UserException
	 */
	private TData dottimes(Map<String, TData> vars) throws EvalException, UserException {
		try {
			return SyntaxNode.getNumeric(this.childs[0], vars)
					.multiplyElem(SyntaxNode.getNumeric(this.childs[1], vars));
		} catch (TDataException e) {
			throw new EvalException(e.getMessage(), this.tok, e.getStackTrace());
		}
	}

	/**
	 * Slash token
	 * 
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws EvalException
	 * @throws UserException
	 */
	private TData slash(Map<String, TData> vars) throws EvalException, UserException {
		TNumeric num, div;

		num = SyntaxNode.getNumeric(this.childs[0], vars);
		div = SyntaxNode.getNumeric(this.childs[1], vars);
		if (!div.isScalar() && !num.isScalar()) {
			throw new EvalException("One side of divison has to be a scalar value", this.tok);
		}

		try {
			return num.multiply(div.invertElem());
		} catch (TDataException e) {
			throw new EvalException(e.getMessage(), this.tok, e.getStackTrace());
		}
	}

	/**
	 * Dot-slash token
	 * 
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws EvalException
	 * @throws UserException
	 */
	private TData dotslash(Map<String, TData> vars) throws EvalException, UserException {
		TNumeric num, div;

		num = SyntaxNode.getNumeric(this.childs[0], vars);
		div = SyntaxNode.getNumeric(this.childs[1], vars);

		try {
			return num.multiplyElem(div.invertElem());
		} catch (TDataException e) {
			throw new EvalException(e.getMessage(), this.tok, e.getStackTrace());
		}
	}

	/**
	 * Percent token
	 * 
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws EvalException
	 * @throws UserException
	 */
	private TData percent(Map<String, TData> vars) throws EvalException, UserException {
		TNumeric num1, num2;

		num1 = SyntaxNode.getReal(this.childs[0], vars);
		num2 = SyntaxNode.getReal(this.childs[1], vars);

		return this.moduloScalar(num1, num2);
	}

	/**
	 * Dot-percent token
	 * 
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws EvalException
	 * @throws UserException
	 */
	private TData dotpercent(Map<String, TData> vars) throws EvalException, UserException {
		TNumeric num1, num2;
		double[][] res;

		num1 = SyntaxNode.getReal(this.childs[0], vars);
		num2 = SyntaxNode.getReal(this.childs[1], vars);

		if (num2.isScalar() || num1.isScalar()) {
			return this.moduloScalar(num1, num2);
		} else if ((num1.size()[0] != num2.size()[0]) || (num1.size()[1] != num2.size()[1])) {
			throw new EvalException("matrix dimension error", this.tok);
		} else {
			res = new double[num1.size()[0]][num1.size()[1]];
			for (int row = 0; row < res.length; row++) {
				for (int col = 0; col < res[0].length; col++) {
					res[row][col] = num1.getReal(row, col) % num2.getReal(row, col);
				}
			}
			try {
				return new TNumeric(res);
			} catch (TDataException e) {
				throw new EvalException(e.getMessage(), this.tok, e.getStackTrace());
			}
		}
	}

	/**
	 * Modulo for real numeric (one has to be scalar)
	 * 
	 * @param num1
	 *            Numeric left side
	 * @param num2
	 *            Numeric right side
	 * @return Evaluated data
	 * 
	 * @throws EvalException
	 */
	private TNumeric moduloScalar(TNumeric num1, TNumeric num2) throws EvalException {
		double val;
		double[][] res;

		if (num2.isScalar()) {
			val = num2.getReal(0, 0);
			res = new double[num1.size()[0]][num1.size()[1]];
			for (int row = 0; row < res.length; row++) {
				for (int col = 0; col < res[0].length; col++) {
					res[row][col] = num1.getReal(row, col) % val;
				}
			}
		} else if (num1.isScalar()) {
			val = num1.getReal(0, 0);
			res = new double[num2.size()[0]][num2.size()[1]];
			for (int row = 0; row < res.length; row++) {
				for (int col = 0; col < res[0].length; col++) {
					res[row][col] = val % num2.getReal(row, col);
				}
			}
		} else {
			throw new EvalException("One side of modulo has to be a scalar value", this.tok);
		}

		try {
			return new TNumeric(res);
		} catch (TDataException e) {
			throw new EvalException(e.getMessage(), this.tok, e.getStackTrace());
		}
	}

	/**
	 * Caret token
	 * 
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws EvalException
	 * @throws UserException
	 */
	private TData caret(Map<String, TData> vars) throws EvalException, UserException {

		TNumeric num1, num2;

		num1 = SyntaxNode.getReal(this.childs[0], vars);
		num2 = SyntaxNode.getReal(this.childs[1], vars);

		return this.powScalar(num1, num2);
	}

	/**
	 * Dot-caret token
	 * 
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws EvalException
	 * @throws UserException
	 */
	private TData dotcaret(Map<String, TData> vars) throws EvalException, UserException {
		TNumeric num1, num2;
		double[][] res;

		num1 = SyntaxNode.getReal(this.childs[0], vars);
		num2 = SyntaxNode.getReal(this.childs[1], vars);

		if (num2.isScalar() || num1.isScalar()) {
			return this.powScalar(num1, num2);
		} else if ((num1.size()[0] != num2.size()[0]) || (num1.size()[1] != num2.size()[1])) {
			throw new EvalException("matrix dimension error", this.tok);
		} else {
			res = new double[num1.size()[0]][num1.size()[1]];
			for (int row = 0; row < res.length; row++) {
				for (int col = 0; col < res[0].length; col++) {
					res[row][col] = Math.pow(num1.getReal(row, col), num2.getReal(row, col));
				}
			}
			try {
				return new TNumeric(res);
			} catch (TDataException e) {
				throw new EvalException(e.getMessage(), this.tok, e.getStackTrace());
			}
		}
	}

	/**
	 * Power for real numeric (one has to be scalar)
	 * 
	 * @param num1
	 *            Numeric left side
	 * @param num2
	 *            Numeric right side
	 * @return Evaluated data
	 * 
	 * @throws EvalException
	 */
	private TNumeric powScalar(TNumeric num1, TNumeric num2) throws EvalException {
		double val;
		double[][] res;

		if (num2.isScalar()) {
			val = num2.getReal(0, 0);
			res = new double[num1.size()[0]][num1.size()[1]];
			for (int row = 0; row < res.length; row++) {
				for (int col = 0; col < res[0].length; col++) {
					res[row][col] = Math.pow(num1.getReal(row, col), val);
				}
			}
		} else if (num1.isScalar()) {
			val = num1.getReal(0, 0);
			res = new double[num2.size()[0]][num2.size()[1]];
			for (int row = 0; row < res.length; row++) {
				for (int col = 0; col < res[0].length; col++) {
					res[row][col] = Math.pow(val, num2.getReal(row, col));
				}
			}
		} else {
			throw new EvalException("One side of power has to be a scalar value", this.tok);
		}

		try {
			return new TNumeric(res);
		} catch (TDataException e) {
			throw new EvalException(e.getMessage(), this.tok, e.getStackTrace());
		}
	}

	/**
	 * Accent token
	 * 
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws EvalException
	 * @throws UserException
	 */
	private TData accent(Map<String, TData> vars) throws EvalException, UserException {
		return SyntaxNode.getNumeric(this.childs[0], vars).transpose();
	}

	/**
	 * Access numeric token
	 * 
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws EvalException
	 * @throws UserException
	 */
	private TNumeric accessNumeric(Map<String, TData> vars) throws EvalException, UserException {
		TNumeric num;
		int[] rows, cols;

		num = SyntaxNode.getNumeric(this.childs[0], vars);
		try {
			if (this.childs.length == 3) {
				rows = SyntaxNode.getIndices(this.childs[1], num.size()[0] - 1, vars);
				cols = SyntaxNode.getIndices(this.childs[2], num.size()[1] - 1, vars);
			} else {
				if (num.getNumRows() == 1) {
					rows = new int[] { 0, 0 };
					cols = SyntaxNode.getIndices(this.childs[1], num.size()[1] - 1, vars);
				} else if (num.getNumColumns() == 1) {
					rows = SyntaxNode.getIndices(this.childs[1], num.size()[0] - 1, vars);
					cols = new int[] { 0, 0 };
				} else {
					rows = SyntaxNode.getIndices(this.childs[1], num.size()[0] - 1, vars);
					cols = new int[] { 0, num.size()[1] - 1 };
				}
			}

			return num.getSubMatrix(rows[0], rows[1], cols[0], cols[1]);
		} catch (TDataException e) {
			throw new EvalException(e.getMessage(), this.tok, e.getStackTrace());
		}
	}

	/**
	 * Access array token
	 * 
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws EvalException
	 * @throws UserException
	 */
	private TData accessArray(Map<String, TData> vars) throws EvalException, UserException {
		TArray arr;
		int row, col;

		arr = SyntaxNode.getArray(this.childs[0], vars);
		row = SyntaxNode.getInt(SyntaxNode.getScalarReal(this.childs[1], vars));
		try {
			if (this.childs.length == 3) {
				col = SyntaxNode.getInt(SyntaxNode.getScalarReal(this.childs[2], vars));
				return arr.get(row, col);
			} else {
				if (arr.size()[0] == 1) {
					return arr.get(0, row);
				} else if (arr.size()[1] == 1) {
					return arr.get(row, 0);
				} else {
					return arr.getRow(row);
				}
			}
		} catch (ArrayIndexOutOfBoundsException e) {
			throw new EvalException(e.getMessage(), this.tok, e.getStackTrace());
		} catch (TDataException e) {
			throw new EvalException(e.getMessage(), this.tok, e.getStackTrace());
		}
	}

	/**
	 * Access struct token
	 * 
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws EvalException
	 * @throws UserException
	 */
	private TData struct(Map<String, TData> vars) throws EvalException, UserException {
		TStruct struct;
		String key;
		TData data;

		struct = SyntaxNode.getStruct(this.childs[0], vars);
		key = SyntaxNode.getString(this.childs[1], vars);
		data = struct.get(key);
		if (data == null) {
			throw new EvalException("Key '" + key + "'of struct undefined", this.tok);
		} else {
			return data;
		}
	}

	/**
	 * Method token
	 * 
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws EvalException
	 * @throws UserException
	 */
	private TData method(Map<String, TData> vars) throws EvalException, UserException {
		TData obj;
		String name;
		TData[] params;

		obj = this.childs[0].eval(vars);
		name = this.childs[1].tok.str;

		// eval parameter
		params = new TData[this.childs.length - 2];
		for (int k = 0; k < params.length; k++) {
			params[k] = this.childs[k + 2].eval(vars);
		}

		try {
			return EvalMethod.evalMethod(obj, name, params);
		} catch (Exception ex) {
			throw new EvalException("Exception by eval method '" + name + "': " + ex.getMessage(), this.tok);
		}
	}

	/**
	 * Define numeric token
	 * 
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws EvalException
	 * @throws UserException
	 */
	private TNumeric defineNumeric(Map<String, TData> vars) throws EvalException, UserException {
		TNumeric n;
		double[][] re;
		double[][] im;
		int row, col, numRow = 0, numCol = 0;

		re = new double[this.childs.length][];
		im = new double[this.childs.length][];
		row = 0;
		for (SyntaxNode rNode : this.childs) {
			re[row] = new double[rNode.childs.length];
			im[row] = new double[rNode.childs.length];
			col = 0;
			for (SyntaxNode cNode : rNode.childs) {
				n = SyntaxNode.getNumeric(cNode, vars);
				// resize rows and check number of rows
				if (col == 0) {
					numRow = n.getNumRows();
					if (numRow > 1) {
						re = Arrays.copyOf(re, (re.length + numRow) - 1);
						im = Arrays.copyOf(im, (im.length + numRow) - 1);
						for (int r = row + 1; r < (row + numRow); r++) {
							re[r] = new double[re[row].length];
							im[r] = new double[im[row].length];
						}
					}
				} else if (n.getNumRows() != numRow) {
					throw new EvalException("Rows have not the same length", this.tok);
				}
				// resize columns
				numCol = n.getNumColumns();
				if (numCol > 1) {
					for (int r = row; r < (row + numRow); r++) {
						re[r] = Arrays.copyOf(re[r], (re[r].length + numCol) - 1);
						im[r] = Arrays.copyOf(im[r], (im[r].length + numCol) - 1);
					}
				}

				// insert values
				for (int r = 0; r < numRow; r++) {
					for (int c = 0; c < numCol; c++) {
						re[row + r][col + c] = n.getReal(r, c);
						im[row + r][col + c] = n.getImag(r, c);
					}
				}
				col += numCol;
			}
			// check number of columns
			if (re[row].length != re[0].length) {
				throw new EvalException("Colomns have not the same length", this.tok);
			}
			row += numRow;
		}

		try {
			return new TNumeric(re, im);
		} catch (TDataException e) {
			throw new EvalException(e.getMessage(), this.tok, e.getStackTrace());
		}
	}

	/**
	 * Define array token
	 * 
	 * @param vars
	 *            Map of variables
	 * @return Evaluated data
	 * 
	 * @throws EvalException
	 * @throws UserException
	 */
	private TArray defineArray(Map<String, TData> vars) throws EvalException, UserException {
		TArray arr;
		SyntaxNode rNode;

		arr = new TArray();
		for (int row = 0; row < this.childs.length; row++) {
			rNode = this.childs[row];
			for (int col = 0; col < rNode.childs.length; col++) {
				arr.set(row, col, rNode.childs[col].eval(vars));
			}
			if (rNode.childs.length != this.childs[0].childs.length) {
				throw new EvalException("Lines have not the same length", this.tok);
			}
		}

		return arr;
	}

	@Override
	public String toString() {
		return this.tok.kind.getDescription();
	}
}
