/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9713 $
 * $Date: 2020-03-19 09:28:58 +0100 (Thu, 19 Mar 2020) $
 */

package at.lcm.bu21.magopt.structure.ipython;

import java.util.ArrayList;
import java.util.Arrays;

import org.json.simple.JSONAware;
import org.json.simple.JSONObject;
import org.json.simple.JSONValue;

/**
 * 
 * 
 * @author gweidenholzer
 */
public class Message {

	static final byte[] DELIM = "<IDS|MSG>".getBytes();

	MessageHeader header;
	private MessageHeader parent;

	private JSONAware content;
	private JSONAware metadata;
	byte[][] ident;
	byte[][] buffers;

	/**
	 * Build a message of the given type, with content, parent and metadata
	 * 
	 * @param msg_type
	 *            the message type
	 * @param content
	 *            message content
	 * @param parent
	 *            message parent header
	 * @param metadata
	 *            metadata
	 */
	public Message(KernelConnection kernel, String msg_type, JSONAware content,
			MessageHeader parent, JSONAware metadata, byte[][] ident,
			byte[][] buffers) {

		this.header = new MessageHeader(kernel, msg_type);
		this.content = content;
		this.parent = parent;
		this.metadata = metadata;
		this.ident = ident;
		this.buffers = buffers;

	}

	/**
	 * Constructor using serialized data
	 * 
	 * @param serialized
	 *            serialized message in the form of a list of byte arrays
	 * @throws Exception
	 */
	public Message(KernelConnection kernel, byte[][] serialized)
			throws Exception {

		ArrayList<byte[]> id = new ArrayList<byte[]>();

		// read idents until delimiter found
		int idx = 0;
		for (; idx < serialized.length; idx++) {
			byte[] data = serialized[idx];
			if (Arrays.equals(data, Message.DELIM)) {
				idx++;
				break;
			} else {
				id.add(data);
			}
		}
		this.ident = id.toArray(new byte[0][]);

		// read signature
		byte[] signature = serialized[idx++];

		// read and check signed data
		byte[][] signed = Arrays.copyOfRange(serialized, idx, idx + 5);
		boolean check = kernel.checkSignature(signature, signed);
		if (!check) {
			throw new Exception("Invalid signature!");
		}

		// read header
		this.header = new MessageHeader(serialized[idx++]);

		// read parent header
		try {
			this.parent = new MessageHeader(serialized[idx++]);
		} catch (Exception e) {
			this.parent = null;
		}

		// read metadata
		this.metadata = (JSONAware) Message.unpack(serialized[idx++]);

		// read message content
		this.content = (JSONAware) Message.unpack(serialized[idx++]);

		// read extra buffers
		if ((serialized.length - idx) > 1) {
			this.buffers = Arrays.copyOfRange(serialized, idx,
					serialized.length);
		}
	}

	/**
	 * Serialize a message into a list of byte arrays
	 * 
	 * @throws Exception
	 */
	public byte[][] serialize(KernelConnection session) throws Exception {

		ArrayList<byte[]> serialized = new ArrayList<byte[]>();

		// serialize idents
		if (this.ident != null) {
			for (int k = 0; k < this.ident.length; k++) {
				serialized.add(this.ident[k]);
			}
		}

		// add delimiter
		serialized.add(Message.DELIM);

		// collect data to be signed
		byte[][] toSign = new byte[4][];

		toSign[0] = Message.pack(this.header);
		toSign[1] = Message.pack(this.parent);
		toSign[2] = Message.pack(this.metadata);

		// serialize content
		if (this.content != null) {
			toSign[3] = Message.pack(this.content);
		} else {
			toSign[3] = Message.pack(new JSONObject());
		}

		// sign message
		byte[] signature = session.sign(toSign);
		serialized.add(signature);

		// add signed data
		for (int k = 0; k < toSign.length; k++) {
			serialized.add(toSign[k]);
		}

		// add extra buffers
		if (this.buffers != null) {
			for (int k = 0; k < this.buffers.length; k++) {
				serialized.add(this.buffers[k]);
			}
		}

		return serialized.toArray(new byte[0][]);

	}

	/**
	 * Pack message data into a byte array
	 * 
	 * @param msg
	 *            message data to pack
	 * @return
	 * @throws Exception
	 */
	static byte[] pack(JSONAware msg) throws Exception {

		if (msg == null) {
			msg = new JSONObject();
		}
		String json = msg.toJSONString();
		return json.getBytes("utf-8");
	}

	/**
	 * Pack message data into a byte array
	 * 
	 * @param data
	 *            message data to pack
	 * @return
	 * @throws Exception
	 */
	static Object unpack(byte[] data) throws Exception {

		String str = new String(data, "utf-8");
		return JSONValue.parse(str);

	}

}
