/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9713 $
 * $Date: 2020-03-19 09:28:58 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.magopt.structure.python;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.UnknownHostException;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.Timer;
import java.util.TimerTask;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.jezhumble.javasysmon.JavaSysMon;

import net.razorvine.pyro.PyroException;
import net.razorvine.pyro.PyroProxy;

/**
 * Python proxy with timeout
 *
 * @author Werner Koppelsttter
 */
public class PythonProxy {

	/** Connection */
	protected PyroProxy proxy;

	/** Timeout for calls */
	protected long timeout = 0;

	/** Watchdog */
	private Timer timer = null;

	/** Regular expression pattern (to get pid of server) */
	Pattern pattern = null;

	/**
	 * Open connection without timeout
	 * 
	 * @param hostname
	 *            Name of host
	 * @param port
	 *            Port-number
	 * @param objectid
	 *            Object identifier
	 * @throws UnknownHostException
	 * @throws IOException
	 */
	public PythonProxy(String hostname, int port, String objectid)
			throws UnknownHostException, IOException {
		this.proxy = new PyroProxy(hostname, port, objectid);
	}

	/**
	 * Open connection with timeout
	 * 
	 * @param hostname
	 *            Name of host
	 * @param port
	 *            Port-number
	 * @param objectid
	 *            Object identifier
	 * @param timeout
	 *            Timeout of calls
	 * @throws UnknownHostException
	 * @throws IOException
	 */
	public PythonProxy(String hostname, int port, String objectid, long timeout)
			throws UnknownHostException, IOException {
		this.proxy = new PyroProxy(hostname, port, objectid);
		this.timeout = timeout;
	}

	/**
	 * Set hmac-key
	 * 
	 * @param hmacKey
	 *            hmac-key
	 */
	public void setHmacKey(byte[] hmacKey) {
		this.proxy.pyroHmacKey = hmacKey;
	}

	/**
	 * Call method with arguments
	 * 
	 * @param method
	 *            Methode name
	 * @param arguments
	 *            Arguments for method
	 * @return Return value of method
	 * @throws Exception
	 */
	public Object call(String method, Object... arguments) throws Exception {
		Object obj = null;

		this.startWD();
		try {
			obj = this.proxy.call(method, arguments);
		} catch (PyroException ex) {
			PythonProxy.modifyPyroException(ex);
			throw ex;
		} finally {
			this.stopWD();
		}

		return obj;
	}

	/**
	 * Get port of connection
	 * 
	 * @return Port number
	 */
	public int getPort() {
		return this.proxy.port;
	}

	/**
	 * Close conneciton
	 */
	public void close() {
		this.proxy.close();
	}

	/**
	 * Modify StackTrace list form PyroException (add python stack)
	 * 
	 * @param ex
	 *            Pyro-exception
	 * @throws PyroException
	 */
	private static void modifyPyroException(PyroException ex) {
		LinkedList<StackTraceElement> stack;
		String[] stackline, elems;
		File file;
		String str, methode, cl;
		int line;
		int k = 1;
		StackTraceElement st;

		stack = new LinkedList<StackTraceElement>(
				Arrays.asList(ex.getStackTrace()));
		stackline = ex._pyroTraceback.split("\n");
		while ((k < stackline.length)
				&& stackline[k].trim().startsWith("File")) {
			elems = stackline[k].split(",");

			// file
			str = elems[0].trim();
			str = str.substring(6, str.length() - 1);
			file = new File(str);
			// class
			cl = file.getName();
			cl = cl.substring(0, cl.lastIndexOf("."));

			// line
			str = elems[1].trim();
			str = str.substring(5);
			line = Integer.parseInt(str);

			// methode
			if (elems.length > 2) {
				str = elems[2].trim();
				methode = str.substring(3);
			} else {
				methode = "";
			}

			st = new StackTraceElement(cl, methode, file.getName(), line);
			stack.addFirst(st);
			k += 2;
		}
		ex.setStackTrace(stack.toArray(new StackTraceElement[0]));
	}

	/**
	 * Start watchdog
	 */
	protected void startWD() {
		if (this.timeout > 0) {
			this.timer = new Timer();

			this.timer.schedule(new TimerTask() {
				@Override
				public void run() {
					try {
						PythonProxy.this.timeout();
					} catch (Exception e) {
						System.err.println(e.getMessage());
						e.printStackTrace();
					}
				}
			}, this.timeout);
		}
	}

	/**
	 * Stop watchdog
	 * 
	 * @throws Exception
	 */
	protected void stopWD() throws Exception {
		if (this.timeout > 0) {
			if (this.timer == null) {
				throw new Exception("Timeout by calling python command");
			} else {
				this.timer.cancel();
				this.timer = null;
			}
		}
	}

	/**
	 * Timeout of command (connextion). Kill connection
	 */
	protected void timeout() {
		ProcessBuilder builder;
		Process proc;
		BufferedReader in = null;
		String line;
		Matcher matcher;
		int pid = 0;
		JavaSysMon sys;

		// get pid of server
		if (System.getProperty("os.name").equalsIgnoreCase("Linux")) {
			// linux
			if (this.pattern == null) {
				this.pattern = Pattern.compile(
						"\\s*.+\\s+\\d+\\s+\\d+\\s+.+:" + this.getPort()
								+ "\\s+.+:\\d+\\s+.+\\s+(\\d+)/.+\\s*");
			}

			builder = new ProcessBuilder("netstat", "-n", "-p"); // -a listening
		} else {
			// windows
			if (this.pattern == null) {
				this.pattern = Pattern
						.compile("\\s*.+\\s+\\[?.*\\]?:" + this.getPort()
								+ "\\s+\\[?.+\\]?:\\d+\\s+.+\\s+(\\d+)\\s*");
			}

			builder = new ProcessBuilder("netstat", "-o", "-n"); // -l listening
		}

		try {
			proc = builder.start();
			in = new BufferedReader(
					new InputStreamReader(proc.getInputStream()));
			while ((line = in.readLine()) != null) {
				matcher = this.pattern.matcher(line);
				if (matcher.matches()) {
					pid = Integer.parseInt(matcher.group(1));
				}
			}
		} catch (IOException e) {
			System.err.println(e.getMessage());
			e.printStackTrace();
		} finally {
			if (in != null) {
				try {
					in.close();
				} catch (IOException e) {
					System.err.println(e.getMessage());
					e.printStackTrace();
				}
			}
		}

		// kill process
		if (pid > 0) {
			System.out.println("Kill process " + pid);
			this.timer.cancel();
			this.timer = null;
			sys = new JavaSysMon();
			sys.killProcessTree(pid, false);
			sys.killProcess(pid);
		}
	}
}
