/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9714 $
 * $Date: 2020-03-19 09:37:13 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.magopt.utils;

import java.io.File;
import java.io.IOException;
import java.lang.reflect.Method;
import java.net.ServerSocket;
import java.net.URL;
import java.net.URLClassLoader;
import java.rmi.RemoteException;

import at.lcm.bu21.magopt.structure.IProject;
import at.lcm.bu21.magopt.structure.Parameter;
import matlabcontrol.MatlabInvocationException;
import matlabcontrol.MatlabProxy;
import matlabcontrol.MatlabProxyFactory;
import matlabcontrol.MatlabProxyFactoryOptions;

/**
 * Get Matlab-proxy with settings from MagOpt-config
 * 
 * @author Werner Koppelsttter
 */
public class MatlabControl {

	/** Port of proxy */
	private static int port = 0;

	/** Matlab-proxy */
	private static MatlabProxy proxy = null;

	/**
	 * Check if is running inside Matlab
	 * 
	 * @return Is inside
	 */
	public static boolean isInsideMatlab() {
		boolean isInside = false;
		Class<?> matlabClass;
		Method isAvailableMethod;

		try {
			// call matlab method to check Matlab is avaiable
			matlabClass = Class.forName("com.mathworks.jmi.Matlab");
			isAvailableMethod = matlabClass.getMethod("isMatlabAvailable");
			isInside = (Boolean) isAvailableMethod.invoke(null);
		} catch (Throwable e) {
		}

		return isInside;
	}

	/**
	 * Get Matlab-proxy
	 * 
	 * @return Matlab-proxy
	 */
	public static MatlabProxy getProxy() {
		return MatlabControl.getProxy((String) null);
	}

	/**
	 * Get Matlab-proxy
	 * 
	 * @param param
	 *            Parameter
	 * @return Matlab-proxy
	 */
	public static MatlabProxy getProxy(Parameter param) {
		String path = null;

		if (param != null) {
			param = param.getRoot();
			if (param instanceof IProject) {
				try {
					path = ((IProject) param).getMopFile().getParent();
				} catch (RemoteException e) {
					e.printStackTrace();
				}
			}
		}

		return MatlabControl.getProxy(path);
	}

	/**
	 * Get Matlab-proxy
	 * 
	 * @param path
	 *            Matlab-path
	 * @return Matlab-proxy
	 */
	public static MatlabProxy getProxy(String path) {
		String str;
		MatlabProxyFactoryOptions.Builder option;

		if ((MatlabControl.proxy == null) || !MatlabControl.proxy.isConnected()) {
			option = new MatlabProxyFactoryOptions.Builder();

			str = MagOptConfig.getConfig("MatlabPath");
			if (str != null) {
				option.setMatlabLocation(str);
			}

			str = MagOptConfig.getConfig("MatlabProxyPort");
			if (str != null) {
				MatlabControl.port = Integer.parseInt(str);
			} else {
				ServerSocket ss = null;
				try {
					ss = new ServerSocket(0);
					MatlabControl.port = ss.getLocalPort();
				} catch (IOException e) {
				} finally {
					if (ss != null) {
						try {
							ss.close();
						} catch (IOException e) {
						}
					}
				}
			}
			if (MatlabControl.port != 0) {
				option.setPort(MatlabControl.port);
			}

			str = MagOptConfig.getConfig("MatlabLogFile");
			if (str != null) {
				option.setLogFile(str);
			}

			str = MagOptConfig.getConfig("MatlabStartupPath");
			if (str != null) {
				option.setMatlabStartingDirectory(new File(str));
			} else {
				str = MagOptConfig.getConfig("MatlabDebugPort");
				if (str != null) {
					option.setJavaDebugger(Integer.parseInt(str));
				}
			}

			try {
				MatlabControl.proxy = new MatlabProxyFactory(option.build()).getProxy();
				if (!MatlabControl.proxy.isRunningInsideMatlab()
						&& (MagOptConfig.getConfig("MatlabStartupPath") == null)) {
					MatlabControl.proxy.feval("javaaddpath",
							new File(MatlabControl.class.getProtectionDomain().getCodeSource().getLocation().getFile())
									.getPath());

					MatlabControl.proxy.feval("at.lcm.bu21.magopt.utils.MatlabControl.updateMPath");
					MatlabControl.proxy.eval(
							"at.lcm.bu21.magopt.utils.MatlabControl.updateJar(java.lang.ClassLoader.getSystemClassLoader())");

					MatlabControl.proxy.feval("javarmpath",
							new File(MatlabControl.class.getProtectionDomain().getCodeSource().getLocation().getFile())
									.getPath());
				}
			} catch (Exception e) {
				e.printStackTrace();
			}
		}

		if (!MatlabControl.proxy.isRunningInsideMatlab()) {
			if ((path != null) && !path.isEmpty()) {
				try {
					MatlabControl.proxy.eval("cd '" + path + "'");
				} catch (MatlabInvocationException e) {
					e.printStackTrace();
				}
			}
		}

		return MatlabControl.proxy;
	}

	/**
	 * Update Matlab-path
	 * 
	 * @throws MatlabInvocationException
	 */
	public static void updateMPath() throws MatlabInvocationException {
		MatlabProxy proxy;
		String path;

		proxy = MatlabControl.getProxy();
		path = ";" + ((String) proxy.returningEval("path", 1)[0]).toLowerCase() + ";";

		for (File f : MagOptConfig.getMPath()) {
			if (!path.contains(";" + f.getPath().toLowerCase() + ";")) {
				proxy.feval("addpath", f.getPath());
			}
		}
	}

	/**
	 * Update class-path of class-loader
	 * 
	 * @param cl
	 *            Class-loader to update
	 * @throws MatlabInvocationException
	 */
	public static void updateJar(URLClassLoader cl) throws MatlabInvocationException {
		URL[] urls;

		// modify library-path
		MagOptConfig.updateLib();

		try {
			urls = cl.getURLs();
			Method m;
			m = URLClassLoader.class.getDeclaredMethod("addURL", URL.class);
			m.setAccessible(true);
			for (File f : MagOptConfig.getJavaPath()) {
				for (URL u : urls) {
					if ((f != null) && (f.toURI().toURL().getPath().compareToIgnoreCase(u.getPath()) == 0)) {
						f = null;
					}
				}

				if (f != null) {
					m.invoke(cl, f.toURI().toURL());
				}
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	/**
	 * Update waitbar of matlab gui
	 * 
	 * @param handle
	 *            Handle of gui
	 * @param count
	 *            Value of waitbar
	 * @param max
	 *            Max Value of waitbar
	 */
	// FIXME waitbar
	public static void updateWaitbar(double handle, int count, int max) {
		if (handle != -1) {
			MatlabControl.proxy = MatlabControl.getProxy();
			try {
				MatlabControl.proxy.feval("MagOptGui",
						new Object[] { "StructureControl_Regenerate", handle, (count - max) / count });
			} catch (MatlabInvocationException e) {
				e.printStackTrace();
			}

		}
	}
}
