/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9714 $
 * $Date: 2020-03-19 09:37:13 +0100 (Thu, 19 Mar 2020) $
 */
package magopt.gui.component.browser;

import java.awt.Color;
import java.awt.Component;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Image;
import java.net.MalformedURLException;

import javax.swing.ImageIcon;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTree;
import javax.swing.border.Border;
import javax.swing.border.EmptyBorder;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.TreeCellRenderer;

import magopt.utils.ComponentInfo;

/**
 * TreeCellRenderer for component-browser.
 * 
 * @author Werner Koppelsttter
 */
public class ComponentBrowserTreeCellRenderer extends JPanel implements
		TreeCellRenderer {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** Font of nodes */
	private Font nodeFont;

	/** Font of leaf-nodes */
	private Font leafFont;

	/** Border for leaf-nodes */
	private Border border;

	/** Is node is selected */
	private boolean selected;

	/** Is node is leaf */
	private boolean leaf;

	/** Icon of node */
	private JLabel icon;

	/** Icon for closed path */
	private ImageIcon down = new ImageIcon(this.getClass().getResource(
			"/magopt/gui/component/browser/dropdown.png"));

	/** Icon for expanded path */
	private ImageIcon up = new ImageIcon(this.getClass().getResource(
			"/magopt/gui/component/browser/dropup.png"));

	/** Background color of non selected node */
	Color nonSelectedColor = null;

	/** Background color of selected node */
	Color selectedColor;

	/**
	 * Create TreeCellRenderer for component-browser
	 */
	public ComponentBrowserTreeCellRenderer() {
		Font font;

		this.border = new EmptyBorder(5, 5, 5, 5);
		this.setOpaque(false);
		this.setLayout(new VerticalLayout(5));

		font = new JLabel().getFont();
		this.nodeFont = new Font(font.getName(), Font.BOLD, font.getSize() + 3);
		// this.leafFont = this.nodeFont;
		this.leafFont = new Font(font.getName(), Font.BOLD, font.getSize());

		// this.nonSelectedColor = new Color(255, 255, 255);
		this.selectedColor = new Color(51, 153, 255, 64);
	}

	/**
	 * Paints the value. The background is filled based on selected.
	 */
	@Override
	public void paint(Graphics g) {

		// split between label and icon
		int imageOffset = this.getHeight();
		if (this.icon != null) {
			imageOffset -= this.icon.getHeight() + this.getInsets().bottom + 2;
		}

		if (this.selected) {
			// selection behind label
			if ((this.selectedColor != null)) {
				g.setColor(this.selectedColor);
				g.fillRect(0, 0, this.getWidth(), imageOffset);
			}
		} else {
			// selection behind panel
			if ((this.nonSelectedColor != null)) {
				g.setColor(this.nonSelectedColor);
				g.fillRect(0, 0, this.getWidth(), this.getHeight());
			}
		}

		// if (this.leaf) {
		// g.setColor(new Color(0, 157, 224));
		// g.fillRect(0, 0, this.getWidth(), imageOffset);
		// } else {
		// g.setColor(new Color(216, 216, 216));
		// g.fillRect(0, 0, this.getWidth(), this.getHeight());
		// }

		super.paint(g);

		if (this.selected && (this.icon != null)) {
			// selection before icon
			if ((this.selectedColor != null)) {
				g.setColor(this.selectedColor);
				g.fillRect(0, imageOffset, this.getWidth(), this.getHeight()
						- imageOffset);
			}
		}
	}

	/**
	 * Set the style of the current node
	 * 
	 * @param tree
	 *            Tree to insert
	 * @param value
	 *            Object to display
	 * @param sel
	 *            If node is selected
	 * @param expanded
	 *            If node if expanded
	 * @param leaf
	 *            If node is a leaf
	 * @param row
	 *            Number of row
	 * @param hasFocus
	 *            If node has focus
	 * @return Component to draw node
	 */
	@Override
	public Component getTreeCellRendererComponent(JTree tree, Object value,
			boolean sel, boolean expanded, boolean leaf, int row,
			boolean hasFocus) {
		JLabel label;
		ComponentInfo item = null;
		Image img;
		Object obj;
		String name;
		ImageIcon icon;
		int width, height;

		obj = ((DefaultMutableTreeNode) value).getUserObject();
		if (obj instanceof ComponentInfo) {
			item = (ComponentInfo) obj;
			name = item.getSimpleName();
		} else {
			name = (String) obj;
		}

		this.selected = sel;
		this.leaf = item != null;

		this.removeAll();
		// this.setToolTipText(name);
		label = new JLabel();
		label.setText(name);
		// label.setToolTipText(name);
		if (item == null) {
			label.setFont(this.nodeFont);
			this.add(label);

			this.icon = null;

			if (expanded) {
				label.setIcon(this.up);
			} else {
				label.setIcon(this.down);

			}

			this.setBorder(null);
			this.setBorder(this.border);
		} else {
			label.setFont(this.leafFont);
			this.add(label);

			if (item.getImage() != null) {
				this.icon = new JLabel();
				// this.icon.setToolTipText(item.getName());
				try {
					icon = new ImageIcon(item.getImage().toURI().toURL(),
							item.getName());
					if (icon.getIconWidth() > icon.getIconHeight()) {
						width = 100;
						height = ((100 * icon.getIconHeight()) / icon
								.getIconWidth());
					} else {
						width = (100 * icon.getIconWidth())
								/ icon.getIconHeight();
						height = 100;
					}
					img = icon.getImage().getScaledInstance(width, height,
							Image.SCALE_SMOOTH);
					this.icon.setIcon(new ImageIcon(img));
				} catch (MalformedURLException e) {
					e.printStackTrace();
				}
				this.add(this.icon);
			} else {
				this.icon = null;
			}

			this.setBorder(this.border);
		}

		return this;
	}
}
