/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9714 $
 * $Date: 2020-03-19 09:37:13 +0100 (Thu, 19 Mar 2020) $
 */
package magopt.gui.tree;

import java.io.Serializable;
import java.rmi.RemoteException;

import javax.swing.event.EventListenerList;
import javax.swing.event.TreeModelEvent;
import javax.swing.event.TreeModelListener;
import javax.swing.tree.TreeModel;
import javax.swing.tree.TreePath;

import at.lcm.bu21.magopt.structure.IParamContainer;
import at.lcm.bu21.magopt.structure.IParameter;
import at.lcm.bu21.magopt.structure.IProject;

/**
 * This class represents the treeModel for the ProjectDnDTree. The model
 * represent the project data structure
 * 
 * @author Markus Priglinger
 */
public class ProjectTreeModel implements Serializable, TreeModel {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** Project to display */
	private IProject _project = null;

	/** Listeners. */
	protected EventListenerList _listenerList = new EventListenerList();

	/**
	 * Set Project or display
	 * 
	 * @param project
	 *            Project to display
	 */
	public void setProject(IProject project) {

		this._project = project;

		// refresh tree of model
		for (TreeModelListener listener : this._listenerList
				.getListeners(TreeModelListener.class)) {
			listener.treeStructureChanged(new TreeModelEvent(this,
					new TreePath("")));
		}
	}

	/**
	 * generates and fires a TreeModelEvent for TreeModelListeners
	 * 
	 * @param treePath
	 *            path to changed node
	 * */
	public void fireTreeNodeChanged(TreePath treePath) {

		TreePath[] treePathArray = new TreePath[1];
		treePathArray[0] = treePath;

		TreeModelEvent tme = new TreeModelEvent(this, treePathArray);

		for (TreeModelListener listener : this._listenerList
				.getListeners(TreeModelListener.class)) {
			listener.treeNodesChanged(tme);
		}
	}

	/**
	 * generates and fires a TreeModelEvent for TreeModelListeners
	 * 
	 * @param treePath
	 *            path to inserted node
	 * @param childIndex
	 *            index of new child
	 * @param child
	 *            new child
	 * */
	public void fireTreeNodeInserted(TreePath treePath, int childIndex,
			Object child) {

		int[] childIndexArray = new int[1];
		Object[] childArray = new Object[1];
		TreePath[] treePathArray = new TreePath[1];

		childIndexArray[0] = childIndex;
		childArray[0] = child;
		treePathArray[0] = treePath;

		TreeModelEvent tme = new TreeModelEvent(this, treePathArray,
				childIndexArray, childArray);

		for (TreeModelListener listener : this._listenerList
				.getListeners(TreeModelListener.class)) {
			listener.treeNodesInserted(tme);
		}
	}

	/**
	 * generates and fires a TreeModelEvent for TreeModelListeners
	 * 
	 * @param treePath
	 *            path to removed node
	 * @param childIndex
	 *            index of removed child
	 * @param child
	 *            removed child
	 * */
	public void fireTreeNodeRemoved(TreePath treePath, int childIndex,
			Object child) {

		int[] childIndexArray = new int[1];
		Object[] childArray = new Object[1];
		TreePath[] treePathArray = new TreePath[1];

		childIndexArray[0] = childIndex;
		childArray[0] = child;
		treePathArray[0] = treePath;

		TreeModelEvent tme = new TreeModelEvent(this, treePathArray,
				childIndexArray, childArray);

		for (TreeModelListener listener : this._listenerList
				.getListeners(TreeModelListener.class)) {
			listener.treeNodesRemoved(tme);
		}
	}

	@Override
	public Object getRoot() {
		return this._project;
	}

	@Override
	public Object getChild(Object parent, int index) {

		IParameter child = null;
		IParameter[] param;

		try {
			if (parent instanceof IParamContainer) {
				param = ((IParamContainer) parent).getParam();
				if ((index < param.length) && (index >= 0)) {
					child = param[index];
				}
			}
		} catch (RemoteException e) {
			;
		}
		return child;
	}

	@Override
	public int getChildCount(Object parent) {
		int count = 0;

		try {
			if (parent instanceof IParamContainer) {
				count = ((IParamContainer) parent).getParam().length;
			}
		} catch (RemoteException e) {
			;
		}
		return count;
	}

	@Override
	public boolean isLeaf(Object node) {
		boolean leaf = true;

		if (node instanceof IParamContainer) {
			leaf = this.getChildCount(node) == 0;
		}
		return leaf;
	}

	@Override
	public void valueForPathChanged(TreePath path, Object newValue) {
	}

	@Override
	public int getIndexOfChild(Object parent, Object child) {
		IParameter[] param;
		int index = -1;

		if ((child != null) && (parent != null)) {
			try {
				if (parent.equals(((IParameter) child).getParentParameter())) {
					param = ((IParamContainer) parent).getParam();
					for (int k = 0; k < param.length; k++) {
						if (child.equals(param[k])) {
							index = k;
							break;
						}
					}
				}
			} catch (RemoteException e) {
				;
			}
		}
		return index;
	}

	@Override
	public void addTreeModelListener(TreeModelListener l) {

		this._listenerList.add(TreeModelListener.class, l);
	}

	@Override
	public void removeTreeModelListener(TreeModelListener l) {

		this._listenerList.remove(TreeModelListener.class, l);
	}
}
