/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 247 $
 * $Date: 2020-03-19 10:13:52 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.optim.core;

import java.io.File;
import java.io.FileInputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Arrays;

import at.lcm.bu21.general.dtypes.TArray;
import at.lcm.bu21.general.dtypes.TData;
import at.lcm.bu21.general.dtypes.TDataException;
import at.lcm.bu21.general.dtypes.TNumeric;
import at.lcm.bu21.general.dtypes.TString;
import at.lcm.bu21.general.dtypes.TStruct;
import at.lcm.bu21.magopt.structure.Project;
import at.lcm.bu21.optim.OptimizationArchive;
import magopt.utils.MopFile;

/**
 * Optimization archive
 * 
 * @author Werner Koppelsttter
 */
public class Archive implements Serializable {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** Related mop-file */
	private MopFile mop;

	/** Name of archive */
	private String name;

	/** Base-index of project */
	private int basedIndex = 0;

	/** Names of all defined variables */
	private String[] variables = new String[0];

	/** Names of all defined results */
	private String[] results = new String[0];

	/** List of problem definitions */
	private ArrayList<TStruct> problems = new ArrayList<TStruct>();

	/** Current loaded problem */
	private Problem loadedProblem = null;

	/** Memory with all calculated individuals */
	private Memory memory = new Memory(this);

	/** Memory for recompute mapping (sensitivity) results */
	private Memory recomputeMemory = new Memory(this);

	/** based project */
	transient private Project project = null;

	/**
	 * Convert form int-array to double-array
	 * 
	 * @param iValues
	 *            Int-array
	 * @return Double-array
	 */
	public static double[] int2double(int[] iValues) {
		double[] dValues;

		dValues = new double[iValues.length];
		for (int k = 0; k < iValues.length; k++) {
			dValues[k] = iValues[k];
		}

		return dValues;
	}

	/**
	 * Convert form double-array to int-array
	 * 
	 * @param dValues
	 *            Double-array
	 * @return Int-array
	 */
	public static int[] double2int(double[] dValues) {
		int[] iValues;

		iValues = new int[dValues.length];
		for (int k = 0; k < dValues.length; k++) {
			iValues[k] = (int) dValues[k];
		}

		return iValues;
	}

	/**
	 * List archive-names
	 * 
	 * @param mop
	 *            Project file
	 * @return List of names
	 * @throws Exception
	 */
	public static String[] listArchives(MopFile mop) throws Exception {
		return new OptimizationArchive(mop).getArchives();
	}

	/**
	 * Load archive
	 * 
	 * @param mop
	 *            Project-file
	 * @param name
	 *            Name of archive
	 * @return Archive
	 * @throws Exception
	 */
	public static Archive load(MopFile mop, String name) throws Exception {
		return new Archive(mop, name, new OptimizationArchive(mop).read(name));
	}

	/**
	 * Load archive
	 * 
	 * @param project
	 *            Project
	 * @param name
	 *            Name of archive
	 * @return Archive
	 * @throws Exception
	 */
	public static Archive load(Project project, String name) throws Exception {
		Archive archive;

		archive = new Archive(project, name, new OptimizationArchive(project.getMopFile()).read(name));

		return archive;
	}

	/**
	 * Delete Archive
	 * 
	 * @param mop
	 *            Project-file
	 * @param name
	 *            Name of archive
	 * @throws Exception
	 */
	public static void delete(MopFile mop, String name) throws Exception {
		new OptimizationArchive(mop).remove(name);
	}

	/**
	 * Create archive to mop-file
	 * 
	 * @param mop
	 *            Mop-file
	 * @param name
	 *            Name of archive
	 */
	public Archive(MopFile mop, String name) {
		this.mop = mop;
		this.name = name;
	}

	/**
	 * Create archive to mop-file
	 * 
	 * @param project
	 *            Project
	 * @param name
	 *            Name of archive
	 */
	public Archive(Project project, String name) {
		this(project.getMopFile(), name);
		this.project = project;
	}

	/**
	 * Resume archive
	 * 
	 * @param mop
	 *            Mop-file to archive
	 * @param name
	 *            Name of archive
	 * @param resumeData
	 *            Resume data
	 * @throws TDataException
	 */
	private Archive(MopFile mop, String name, TStruct resumeData) throws TDataException {
		this(mop, name);
		this.setResumeData(resumeData);
	}

	/**
	 * Resume archive
	 * 
	 * @param project
	 *            Mop-file to archive
	 * @param name
	 *            Name of archive
	 * @param resumeData
	 *            Resume data
	 * @throws TDataException
	 */
	private Archive(Project project, String name, TStruct resumeData) throws TDataException {
		this(project.getMopFile(), name, resumeData);
		this.project = project;
	}

	/**
	 * Save archive
	 * 
	 * @throws Exception
	 */
	public void save() throws Exception {
		new OptimizationArchive(this.mop).write(this.getName(), this.getResumeData());
	}

	/**
	 * Duplicate archive
	 * 
	 * @param name
	 *            Name of new archive
	 * @return New archive
	 * @throws Exception
	 */
	public Archive duplicate(String name) throws Exception {
		Archive archive = null;
		Problem problem;

		if (!Arrays.asList(Archive.listArchives(this.mop)).contains(name)) {
			archive = new Archive(this.mop, name, this.getResumeData().deepClone());

			// clear solutions, mappings, and sensitivity analyses
			for (String prob : this.listProblems()) {
				problem = archive.readProblem(prob);
				problem.clearSolution();
				problem.clearMapping();
				problem.clearSerializedErrorsClassiffier();
				problem.clearSensitivityAnalyses();
			}
			archive.memory.clear();
			archive.recomputeMemory.clear();

			archive.save();
		}

		return archive;
	}

	/**
	 * Rename archive
	 * 
	 * @param name
	 *            Name of new archive
	 * @throws Exception
	 */
	public void rename(String name) throws Exception {
		String oldName;

		if (!Arrays.asList(Archive.listArchives(this.mop)).contains(name)) {
			oldName = this.name;
			this.name = name;
			this.save();
			Archive.delete(this.mop, oldName);
		}
	}

	/**
	 * Get name of archive
	 * 
	 * @return Name
	 */
	public String getName() {
		return this.name;
	}

	/**
	 * Get related mop-file
	 * 
	 * @return Mop-file
	 */
	public MopFile getMop() {
		return this.mop;
	}

	/**
	 * Get based project of optimization
	 * 
	 * @return Project
	 */
	public Project getProject() {
		if (this.project == null) {
			try {
				this.project = Project.load(this.mop);
			} catch (Exception e) {
				System.err.println(e.getMessage());
				e.printStackTrace();
				this.project = null;
			}
		}

		return this.project;
	}

	/**
	 * Get based project index
	 * 
	 * @return Index of project
	 */
	public int getBasedIndex() {
		return this.basedIndex;
	}

	/**
	 * Set based project index
	 * 
	 * @param index
	 *            Index of project
	 */
	public void setBasedIndex(int index) {
		if (!this.isLocked()) {
			this.basedIndex = index;
		}
	}

	/**
	 * Get name of used variables
	 * 
	 * @return Array of names
	 */
	public String[] getVariables() {
		return this.variables;
	}

	/**
	 * Set variables names
	 * 
	 * @param variables
	 *            List of variable names
	 */
	void setVariables(String[] variables) {
		if (!this.isLocked()) {
			this.variables = variables;
		}
	}

	/**
	 * Get name of used results
	 * 
	 * @return Array of names
	 */
	public String[] getResults() {
		return this.results;
	}

	/**
	 * Set results names
	 * 
	 * @param results
	 *            List of result names
	 */
	void setResults(String[] results) {
		if (!this.isLocked()) {
			this.results = results;
		}
	}

	/**
	 * Get memory of archive
	 * 
	 * @return Memory with individuals
	 */
	public Memory getMemory() {
		return this.memory;
	}

	/**
	 * Get recompute memory of archive
	 * 
	 * @return Memory with individuals
	 */
	public Memory getRecomputeMemory() {
		return this.recomputeMemory;
	}

	/**
	 * Get resume data
	 * 
	 * @return Resume data
	 * @throws Exception
	 */
	TStruct getResumeData() throws Exception {
		TStruct resumeData;
		TArray data;

		resumeData = new TStruct();

		// based dataset-identifier
		resumeData.put("baseIndex", new TNumeric(this.basedIndex));

		// list of variables
		data = new TArray(1, this.variables.length);
		resumeData.put("variables", data);
		for (int k = 0; k < this.variables.length; k++) {
			data.set(0, k, new TString(this.variables[k]));
		}

		// list of results
		data = new TArray(1, this.results.length);
		resumeData.put("results", data);
		for (int k = 0; k < this.results.length; k++) {
			data.set(0, k, new TString(this.results[k]));
		}

		// memory
		resumeData.put("memory", this.memory.getResumeData());

		// recompute memory
		resumeData.put("recomputeMemory", this.recomputeMemory.getResumeData());

		// problem
		this.writeProblem(this.loadedProblem);
		data = new TArray(1, this.problems.size());
		resumeData.put("problems", data);
		for (int k = 0; k < this.problems.size(); k++) {
			data.set(0, k, this.problems.get(k));
		}

		return resumeData;
	}

	/**
	 * Set resume data
	 * 
	 * @param resumeData
	 *            Resume Data
	 * @throws TDataException
	 */
	private void setResumeData(TStruct resumeData) throws TDataException {
		TNumeric num;
		TArray array;

		// based dataset-identifier
		num = (TNumeric) resumeData.get("baseIndex");
		if (num != null) {
			this.basedIndex = (int) num.getReal(0, 0);
		}

		// list of variables
		array = (TArray) resumeData.get("variables");
		if (array != null) {
			this.variables = new String[array.size()[1]];
			for (int k = 0; k < this.variables.length; k++) {
				this.variables[k] = ((TString) array.get(0, k)).getString();
			}
		}

		// list of results
		array = (TArray) resumeData.get("results");
		if (array != null) {
			this.results = new String[array.size()[1]];
			for (int k = 0; k < this.results.length; k++) {
				this.results[k] = ((TString) array.get(0, k)).getString();
			}
		}

		// memory
		this.memory.setResumeData((TArray) resumeData.get("memory"));

		// recompute memory
		this.recomputeMemory.setResumeData((TArray) resumeData.get("recomputeMemory"));

		// problem
		array = (TArray) resumeData.get("problems");
		if (array != null) {
			this.problems.clear();
			for (int k = 0; k < array.size()[1]; k++) {
				this.problems.add((TStruct) array.get(0, k));
			}
		}
	}

	/**
	 * List names of existing problems
	 * 
	 * @return Problem names
	 */
	public String[] listProblems() {
		String[] names;

		names = new String[this.problems.size()];

		for (int k = 0; k < this.problems.size(); k++) {
			names[k] = ((TString) this.problems.get(k).get("name")).getString();
		}

		return names;
	}

	/**
	 * Write problem to resume-data-list
	 * 
	 * @param prob
	 *            Problem to write
	 * @throws Exception
	 */
	void writeProblem(Problem prob) throws Exception {
		if (prob != null) {
			for (int k = 0; k < this.problems.size(); k++) {
				TStruct probStruct = this.problems.get(k);
				if (((TString) probStruct.get("name")).getString().compareTo(prob.getName()) == 0) {
					this.problems.set(k, prob.getResumeData());
				}
			}
		}
	}

	/**
	 * Read problem from resume-data-list
	 * 
	 * @param name
	 *            Name of problem to read
	 * @return Problem
	 * @throws Exception
	 */
	public Problem readProblem(String name) throws Exception {

		this.writeProblem(this.loadedProblem);
		this.loadedProblem = null;

		for (int k = 0; k < this.problems.size(); k++) {
			TStruct prob = this.problems.get(k);
			if (name.compareTo(((TString) prob.get("name")).getString()) == 0) {
				this.loadedProblem = new Problem(this, prob);
			}
		}

		return this.loadedProblem;
	}

	/**
	 * Create Problem for archive
	 * 
	 * @param name
	 *            Name of new Problem
	 * @return Problem if not already exists
	 * @throws Exception
	 */
	public Problem createProblem(String name) throws Exception {
		this.writeProblem(this.loadedProblem);
		this.loadedProblem = null;

		if (!Arrays.asList(this.listProblems()).contains(name)) {
			this.loadedProblem = new Problem(this, name);
			this.problems.add(this.loadedProblem.getResumeData());
		}

		return this.loadedProblem;
	}

	/**
	 * Delete problem form archive
	 * 
	 * @param name
	 *            Name of archive to remove
	 */
	public void deleteProblem(String name) {
		if ((this.loadedProblem != null) && (this.loadedProblem.getName().compareTo(name) == 0)) {
			this.loadedProblem = null;
		}

		for (int k = this.problems.size() - 1; k >= 0; k--) {
			TStruct prob = this.problems.get(k);
			if (name.compareTo(((TString) prob.get("name")).getString()) == 0) {
				this.problems.remove(k);
			}
		}

	}

	/**
	 * Rename problem (called from problem)
	 * 
	 * @param oldName
	 *            Old name
	 * @param newName
	 *            New name
	 */
	void renameProblem(String oldName, String newName) {
		TStruct prob = null;

		if (!Arrays.asList(this.listProblems()).contains(this.name)) {
			for (int k = 0; k < this.problems.size(); k++) {
				prob = this.problems.get(k);
				if (oldName.compareTo(((TString) prob.get("name")).getString()) == 0) {
					prob.put("name", new TString(newName));
				}
			}
		}

	}

	/**
	 * Check if archive is locked (no changes of number of variables or results)
	 * 
	 * @return Is archive locked
	 */
	public boolean isLocked() {
		return (this.problems.size() > 1) || !this.memory.isEmpty();
	}

	/**
	 * @param name
	 *            Name of new problem
	 * @param xml
	 *            XML-file with settings
	 * @return Warning message
	 * @throws Exception
	 */
	public String importProblemSettings(String name, File xml) throws Exception {
		Problem prob;
		TStruct settings = null;
		FileInputStream in;
		String message;

		// load settings
		if (!xml.getName().toLowerCase().endsWith(".xml")) {
			xml = new File(xml.getAbsolutePath() + ".xml");
		}

		in = new FileInputStream(xml);
		settings = (TStruct) TData.read(in);
		in.close();

		// create problem and set settings
		prob = this.createProblem(name);
		message = prob.importSettings(settings);

		// save changes
		this.save();

		return message;

	}

	@Override
	public String toString() {
		StringBuffer str;

		str = new StringBuffer();
		str.append("Archive: " + this.name + "\n");
		str.append("\tProject: " + this.mop.getAbsolutePath() + " @index " + this.basedIndex + "\n");
		str.append("\tVariables:\n");
		for (int k = 0; k < this.variables.length; k++) {
			str.append("\t\t" + this.variables[k] + "\n");
		}
		str.append("\tResults:\n");
		for (int k = 0; k < this.results.length; k++) {
			str.append("\t\t" + this.results[k] + "\n");
		}
		str.append("\tProblems:\n");
		for (int k = 0; k < this.problems.size(); k++) {
			str.append("\t\t" + ((TString) this.problems.get(k).get("name")).getString() + "\n");
		}
		str.append("\tMemory: #" + this.memory.getNumIndividuals() + " individuals\n");

		return str.toString();
	}
}
