/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 247 $
 * $Date: 2020-03-19 10:13:52 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.optim.core.mapping.utils;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import at.lcm.bu21.optim.core.Individual;
import at.lcm.bu21.optim.core.ResultDesc;
import at.lcm.bu21.optim.core.VariableDesc;
import at.lcm.bu21.optim.core.mapping.MappingManager;
import at.lcm.bu21.optim.core.mapping.utils.dstructures.TIndividualInfo;
import at.lcm.bu21.optim.core.mapping.utils.dstructures.TMappingConstants;
import at.lcm.bu21.optim.core.mapping.utils.dstructures.TSensitivityCell;

public class SensitivityAnalysisHolder implements Serializable {
	private static final long serialVersionUID = -0000000000000000004L; // for
																		// opencv-based
																		// mapping-holders

	private String displayName;
	private TIndividualInfo baseIndividualFEResultsInfo;
	private TIndividualInfo baseIndividualMappingResultsInfo;

	private Map<String, Double> plusVariableRanges = new HashMap<String, Double>();
	private Map<String, Double> minusVariableRanges = new HashMap<String, Double>();

	private Map<String, Map<String, TSensitivityCell>> localSensitivityResultsMappings = new HashMap<String, Map<String, TSensitivityCell>>();
	private Map<String, TSensitivityCell> globalSensitivityResultsMappings = new HashMap<String, TSensitivityCell>();

	private Map<String, Map<String, TSensitivityCell>> localSensitivityResultsFE = new HashMap<String, Map<String, TSensitivityCell>>();
	private Map<String, TSensitivityCell> globalSensitivityResultsFE = new HashMap<String, TSensitivityCell>();

	private boolean containsIndividuals;
	private boolean containsMappingResults;
	private boolean containsFEResults;
	private boolean waitingForFEAnalysis;
	private int totalIndividualsCount;

	private Map<String, List<Double[]>> persistedLocalSensitivityIndVarValues = new HashMap<String, List<Double[]>>();
	private List<Double[]> persistedGlobalSensitivityIndVarValues = new ArrayList<Double[]>();

	private transient double FESensitivityProgress;
	private transient Map<String, List<Individual>> localSensitivityIndividuals = new HashMap<String, List<Individual>>();
	private transient List<Individual> globalSensitivityIndividuals = new ArrayList<Individual>();
	private transient MappingManager parentMappingManager;

	public Map<String, List<Individual>> getLocalSensitivityIndividuals() {
		return this.localSensitivityIndividuals;
	}

	public List<Individual> getGlobalSensitivityIndividuals() {
		return this.globalSensitivityIndividuals;
	}

	public SensitivityAnalysisHolder(MappingManager parentMappingManager, Individual baseIndividual) {

		this.parentMappingManager = parentMappingManager;
		this.baseIndividualFEResultsInfo = this.extractIndividualInfo(this.duplicateIndividual(baseIndividual, true),
				true);

		for (VariableDesc vDesc : this.parentMappingManager.getProblem().getVariedVariables()) {
			this.plusVariableRanges.put(vDesc.getName(), this.baseIndividualFEResultsInfo.getVarValue(vDesc.getName())
					* TMappingConstants.DEFAULT_SENSITIVITY_RANGE);
			this.minusVariableRanges.put(vDesc.getName(), this.baseIndividualFEResultsInfo.getVarValue(vDesc.getName())
					* TMappingConstants.DEFAULT_SENSITIVITY_RANGE);
		}

		this.displayName = this.generateDisplayName();
		this.totalIndividualsCount = 0;
		this.containsIndividuals = false;
		this.containsMappingResults = false;
		this.waitingForFEAnalysis = false;
	}

	public double getPlusVariableRange(String variableName) {
		return this.plusVariableRanges.get(variableName);
	}

	public double getMinusVariableRange(String variableName) {
		return this.minusVariableRanges.get(variableName);
	}

	public void setPlusVariableRange(String variableName, double value) {
		this.plusVariableRanges.put(variableName, value);
	}

	public void setMinusVariableRange(String variableName, double value) {
		this.minusVariableRanges.put(variableName, value);
	}

	public boolean getContainsIndividuals() {
		return this.containsIndividuals;
	}

	public void setContainsIndividuals(boolean containsIndividuals) {
		this.containsIndividuals = containsIndividuals;
	}

	public void createLocalSensitivityIndividuals() {

		this.localSensitivityIndividuals = new HashMap<String, List<Individual>>();

		this.parentMappingManager
				.addLog("Local individuals to generate: " + (TMappingConstants.DEFAULT_SENSITIVITY_INTERVALS
						* this.parentMappingManager.getProblem().getVariedVariables().length * 2) + ".");

		for (VariableDesc vDesc : this.parentMappingManager.getProblem().getVariedVariables()) {
			List<Individual> variedIndividualsList = new ArrayList<Individual>();
			double rangeInterval = this.minusVariableRanges.get(vDesc.getName())
					+ this.plusVariableRanges.get(vDesc.getName());
			for (int i = 0; (i <= (2 * TMappingConstants.DEFAULT_SENSITIVITY_INTERVALS)); i++) {
				if (i != TMappingConstants.DEFAULT_SENSITIVITY_INTERVALS) {
					try {
						Individual newInd = this.duplicateIndividual(this.getBaseIndividual("FE"), false);
						newInd.setFromMapping(true);
						double newVal = (newInd.getVariable(vDesc.getName())
								- this.minusVariableRanges.get(vDesc.getName()))
								+ (i * (rangeInterval / (2 * TMappingConstants.DEFAULT_SENSITIVITY_INTERVALS)));
						newInd.setVariable(vDesc.getName(), newVal);
						variedIndividualsList.add(newInd);
					} catch (Exception ex) {
						System.out.println("Error by creation individium");
					}
				}
				this.localSensitivityIndividuals.put(vDesc.getName(), variedIndividualsList);
			}
		}
	}

	public void createGlobalSensitivityIndividuals() {

		this.globalSensitivityIndividuals = new ArrayList<Individual>();

		int nrVariedVariables = this.parentMappingManager.getProblem().getVariedVariables().length;
		boolean performGridSearch = false;
		int latinHypercubeInds = TMappingConstants.DEFAULT_SENSITIVITY_INTERVALS * nrVariedVariables
				* nrVariedVariables;
		int gridInds = 0;
		if (nrVariedVariables <= TMappingConstants.MAX_VARIABLES_FOR_SENSITIVITY_GRID_SEARCH) {
			performGridSearch = true;
			gridInds = (int) Math.pow(2, nrVariedVariables);
			latinHypercubeInds = TMappingConstants.MAX_SENSITIVITY_EVALUATIONS - gridInds;
		} else {
			latinHypercubeInds = TMappingConstants.MAX_SENSITIVITY_EVALUATIONS;
		}

		this.parentMappingManager.addLog("Global individuals to generate: " + gridInds + " grid and "
				+ latinHypercubeInds + " latin hypercube.");

		// latin hypercube sampling
		List<Individual> latinGenerated = new ArrayList<Individual>();

		// create distribution of variables
		Map<String, LinkedList<Double>> dist = new HashMap<String, LinkedList<Double>>();
		for (VariableDesc vDesc : this.parentMappingManager.getProblem().getVariedVariables()) {
			LinkedList<Double> values = new LinkedList<Double>();
			for (int i = 0; i < latinHypercubeInds; i++) {
				double val = (-this.minusVariableRanges.get(vDesc.getName())
						+ (((this.plusVariableRanges.get(vDesc.getName())
								+ this.minusVariableRanges.get(vDesc.getName())) * i) / (latinHypercubeInds - 1)))
						+ this.baseIndividualFEResultsInfo.getVarValue(vDesc.getName());
				values.add(val);
			}
			dist.put(vDesc.getName(), values);
		}

		// combine values from distribution
		while (latinGenerated.size() < latinHypercubeInds) {
			try {
				Individual newInd = this.duplicateIndividual(this.getBaseIndividual("FE"), false);
				for (VariableDesc vDesc : this.parentMappingManager.getProblem().getVariedVariables()) {
					LinkedList<Double> values = dist.get(vDesc.getName());

					newInd.setVariable(vDesc.getName(), values.remove((int) (Math.random() * values.size())));
				}
				newInd.setFromMapping(true);
				latinGenerated.add(newInd);
			} catch (Exception ex) {
				System.out.println("Error by creation individium");
			}

		}

		// edge values grid
		if (performGridSearch) {
			this.globalSensitivityIndividuals.add(this.duplicateIndividual(this.getBaseIndividual("FE"), false));
			for (VariableDesc vDesc : this.parentMappingManager.getProblem().getVariedVariables()) {

				List<Individual> newMinus = new ArrayList<Individual>();
				List<Individual> newPlus = new ArrayList<Individual>();

				for (Individual existingInd : this.globalSensitivityIndividuals) {
					try {
						Individual newInd = this.duplicateIndividual(existingInd, false);
						newInd.setFromMapping(true);

						newInd.setVariable(vDesc.getName(),
								this.baseIndividualFEResultsInfo.getVarValue(vDesc.getName())
										- this.getMinusVariableRange(vDesc.getName()));
						newMinus.add(newInd);

						newInd.setVariable(vDesc.getName(),
								this.baseIndividualFEResultsInfo.getVarValue(vDesc.getName())
										+ this.getPlusVariableRange(vDesc.getName()));
						newPlus.add(newInd);
					} catch (Exception ex) {
						System.out.println("Error by creation individium");
					}
				}

				this.globalSensitivityIndividuals.clear();

				this.globalSensitivityIndividuals.addAll(newMinus);
				this.globalSensitivityIndividuals.addAll(newPlus);
			}
		}
		this.globalSensitivityIndividuals.addAll(latinGenerated);
	}

	private Individual duplicateIndividual(Individual originalInd, boolean duplicateResults) {
		Individual newInd = null;
		try {
			newInd = new Individual(this.parentMappingManager.getProblem().getArchive());
			newInd.setFromMapping(true);

			for (VariableDesc vDesc : this.parentMappingManager.getProblem().getVariables()) {
				newInd.setVariable(vDesc.getName(), originalInd.getVariable(vDesc.getName()));
			}
			if (duplicateResults) {
				for (ResultDesc rDesc : this.parentMappingManager.getProblem().getResults()) {
					newInd.setResult(rDesc.getName(), originalInd.getResult(rDesc.getName()));
				}
			}
		} catch (Exception ex) {
			System.out.println("Error by creation individium");
		}

		return newInd;
	}

	@SuppressWarnings("unused")
	private TIndividualInfo duplicateIndividualInfo(TIndividualInfo originalIndInfo, boolean duplicateResults) {
		TIndividualInfo newIndInfo = new TIndividualInfo();

		for (VariableDesc vDesc : this.parentMappingManager.getProblem().getVariables()) {
			newIndInfo.setVarValue(vDesc.getName(), originalIndInfo.getVarValue(vDesc.getName()));
		}

		for (ResultDesc rDesc : this.parentMappingManager.getProblem().getResults()) {
			if (duplicateResults) {
				newIndInfo.setResValue(rDesc.getName(), originalIndInfo.getResValue(rDesc.getName()));
			} else {
				newIndInfo.setResValue(rDesc.getName(), Double.MAX_VALUE);
			}
		}

		return newIndInfo;
	}

	private TIndividualInfo extractIndividualInfo(Individual ind, boolean extractResultValues) {
		TIndividualInfo indInfo = new TIndividualInfo();

		for (VariableDesc vDesc : this.parentMappingManager.getProblem().getVariables()) {
			indInfo.setVarValue(vDesc.getName(), ind.getVariable(vDesc.getName()));
		}

		for (ResultDesc rDesc : this.parentMappingManager.getProblem().getResults()) {
			if (extractResultValues) {
				indInfo.setResValue(rDesc.getName(), ind.getResult(rDesc.getName()));
			} else {
				indInfo.setResValue(rDesc.getName(), Double.MAX_VALUE);
			}
		}

		return indInfo;
	}

	private Individual createIndividualFromInfo(TIndividualInfo indInfo) {
		Individual ind = null;
		try {
			ind = new Individual(this.parentMappingManager.getProblem().getArchive());
			ind.setFromMapping(true);

			for (VariableDesc vDesc : this.parentMappingManager.getProblem().getVariables()) {

				ind.setVariable(vDesc.getName(), indInfo.getVarValue(vDesc.getName()));
			}

			for (ResultDesc rDesc : this.parentMappingManager.getProblem().getResults()) {
				ind.setResult(rDesc.getName(), indInfo.getResValue(rDesc.getName()));
			}
		} catch (Exception ex) {
			System.out.println("Error by creation individium");
		}

		return ind;
	}

	public void updateIndividualsCount() {
		this.totalIndividualsCount = 0;
		for (VariableDesc vDesc : this.parentMappingManager.getProblem().getVariedVariables()) {
			List<Individual> listIndVarValues = this.localSensitivityIndividuals.get(vDesc.getName());
			if (listIndVarValues != null) {
				this.totalIndividualsCount += listIndVarValues.size();
			}
		}

		this.parentMappingManager.addLog(this.totalIndividualsCount + " individuals generated during local analysis!");

		this.totalIndividualsCount += this.globalSensitivityIndividuals.size();

		this.parentMappingManager.addLog(this.totalIndividualsCount + " total individuals generated during analysis!");

		if (this.totalIndividualsCount != 0) {
			this.containsIndividuals = true;
		}
	}

	public int getTotalIndividualsCount() {
		return this.totalIndividualsCount;
	}

	public String getLocalSensitivityValueToDisplay(String variableName, String targetName) {
		String result = "";

		if (!this.containsMappingResults) {
			result += "-";
		} else {
			result += String.format("%.4f",
					this.localSensitivityResultsMappings.get(variableName).get(targetName).getRelativeResToDisplay());
		}

		if (!this.containsFEResults) {
			result += "/-";
		} else {
			result += "/" + String.format("%.4f",
					this.localSensitivityResultsFE.get(variableName).get(targetName).getRelativeResToDisplay());
		}

		return result;
	}

	public String getGlobalSensitivityRelativeValueToDisplay(String targetName) {
		String result = "";

		if (!this.containsMappingResults) {
			result += "-";
		} else {
			result += String.format("%.4f",
					this.globalSensitivityResultsMappings.get(targetName).getRelativeResToDisplay());
		}

		if (!this.containsFEResults) {
			result += "/-";
		} else {
			result += "/"
					+ String.format("%.4f", this.globalSensitivityResultsFE.get(targetName).getRelativeResToDisplay());
		}

		return result;
	}

	public Double getGlobalSensitivityAbsoluteMinValueToDisplayDouble(String targetName) {
		Double result = Double.NEGATIVE_INFINITY;

		if (this.containsMappingResults) {
			result = this.baseIndividualFEResultsInfo.getResValue(targetName)
					- this.globalSensitivityResultsMappings.get(targetName).getAbsoluteMinResToDisplay();
		}
		return result;
	}

	public String getGlobalSensitivityAbsoluteMinValueToDisplayStr(String targetName) {
		String result = "";

		if (!this.containsMappingResults) {
			result += "-";
		} else {
			result += String.format("%.4f", this.baseIndividualFEResultsInfo.getResValue(targetName)
					- this.globalSensitivityResultsMappings.get(targetName).getAbsoluteMinResToDisplay());
		}

		if (!this.containsFEResults) {
			result += "/-";
		} else {
			result += "/" + String.format("%.4f", this.baseIndividualFEResultsInfo.getResValue(targetName)
					- this.globalSensitivityResultsFE.get(targetName).getAbsoluteMinResToDisplay());
		}

		return result;
	}

	public Double getGlobalSensitivityAbsoluteMaxValueToDisplayDouble(String targetName) {
		Double result = Double.POSITIVE_INFINITY;

		if (this.containsMappingResults) {

			result = this.baseIndividualFEResultsInfo.getResValue(targetName)
					+ this.globalSensitivityResultsMappings.get(targetName).getAbsoluteMaxResToDisplay();
		}

		return result;
	}

	public String getGlobalSensitivityAbsoluteMaxValueToDisplayStr(String targetName) {
		String result = "";

		if (!this.containsMappingResults) {
			result += "-";
		} else {
			result += String.format("%.4f", this.baseIndividualFEResultsInfo.getResValue(targetName)
					+ this.globalSensitivityResultsMappings.get(targetName).getAbsoluteMaxResToDisplay());
		}

		if (!this.containsFEResults) {
			result += "/-";
		} else {
			result += "/" + String.format("%.4f", this.baseIndividualFEResultsInfo.getResValue(targetName)
					+ this.globalSensitivityResultsFE.get(targetName).getAbsoluteMaxResToDisplay());
		}

		// System.out.println(targetName
		// + " "
		// + this.baseIndividualFEResultsInfo.getResValue(targetName)
		// + " "
		// + this.globalSensitivityResultsFE.get(targetName)
		// .getAbsoluteMaxResToDisplay());

		return result;
	}

	public Double getNormalizedSensitivity(String variableName, String targetName, boolean useMappingValues,
			boolean useFEValues) {

		if (useMappingValues) {
			if (!this.containsMappingResults) {
				return -1.0;
			}
			return this.localSensitivityResultsMappings.get(variableName).get(targetName).getResNormRelativeRange();
		}

		if (useFEValues) {
			if (!this.containsFEResults) {
				return -1.0;
			}
			return this.localSensitivityResultsFE.get(variableName).get(targetName).getResNormRelativeRange();
		}

		return -1.0;
	}

	private void postProcessSensitivityInformation(Individual baseInd, Map<String, List<Individual>> localSensInds,
			List<Individual> globalSensInds, Map<String, Map<String, TSensitivityCell>> localSensResults,
			Map<String, TSensitivityCell> globalSensResults) {

		// local sensitivity post-processing
		int pass = 0;
		for (VariableDesc vDesc : this.parentMappingManager.getProblem().getVariedVariables()) {
			HashMap<String, TSensitivityCell> newTargetMap = new HashMap<String, TSensitivityCell>();
			for (String rName : this.parentMappingManager.getSensitivityTargetNames()) {
				newTargetMap.put(rName, new TSensitivityCell());

				if (pass == 0) {
					globalSensResults.put(rName, new TSensitivityCell());
				}
			}
			pass++;
			localSensResults.put(vDesc.getName(), newTargetMap);
		}

		for (VariableDesc vDesc : this.parentMappingManager.getProblem().getVariedVariables()) {
			for (Individual ind : localSensInds.get(vDesc.getName())) {
				for (String rName : this.parentMappingManager.getSensitivityTargetNames()) {
					TSensitivityCell curCell = localSensResults.get(vDesc.getName()).get(rName);

					double baseIndividualVarValue = baseInd.getVariable(vDesc.getName());

					if (baseIndividualVarValue == 0) {
						baseIndividualVarValue = (this.parentMappingManager.getProblem().getVariable(vDesc.getName())
								.getUB() - this.parentMappingManager.getProblem().getVariable(vDesc.getName()).getLB());
					}

					curCell.updateCell(ind.getVariable(vDesc.getName()), ind.getResult(rName), baseIndividualVarValue,
							baseInd.getResult(rName));

					curCell = globalSensResults.get(rName);
					curCell.updateCell(ind.getVariable(vDesc.getName()), ind.getResult(rName), baseIndividualVarValue,
							baseInd.getResult(rName));
				}
			}

		}

		double minValue = Double.MAX_VALUE;
		double maxValue = -Double.MAX_VALUE;
		for (VariableDesc vDesc : this.parentMappingManager.getProblem().getVariedVariables()) {

			for (String rName : this.parentMappingManager.getSensitivityTargetNames()) {

				TSensitivityCell curCell = localSensResults.get(vDesc.getName()).get(rName);

				if (curCell.getResRelativeRange() < minValue) {
					minValue = curCell.getResRelativeRange();
				}

				if (curCell.getResRelativeRange() > maxValue) {
					maxValue = curCell.getResRelativeRange();
				}
			}

		}

		for (VariableDesc vDesc : this.parentMappingManager.getProblem().getVariedVariables()) {

			for (String rName : this.parentMappingManager.getSensitivityTargetNames()) {

				TSensitivityCell curCell = localSensResults.get(vDesc.getName()).get(rName);
				curCell.setResNormRelativeRange((curCell.getResRelativeRange() - minValue) / (maxValue - minValue));
			}

		}

		// global sensitivity post-processing
		Map<String, TSensitivityCell> varRanges = new HashMap<String, TSensitivityCell>();
		for (VariableDesc vDesc : this.parentMappingManager.getProblem().getVariedVariables()) {
			varRanges.put(vDesc.getName(), new TSensitivityCell());
		}
		Map<String, TSensitivityCell> resRanges = new HashMap<String, TSensitivityCell>();
		for (String targetName : this.parentMappingManager.getSensitivityTargetNames()) {
			resRanges.put(targetName, new TSensitivityCell());
		}

		double maxVarRange = -Double.MAX_VALUE;

		if (!this.waitingForFEAnalysis && !this.containsFEResults) {
			List<Individual> newGlobalList = new ArrayList<Individual>();
			for (VariableDesc vDesc : this.parentMappingManager.getProblem().getVariedVariables()) {
				newGlobalList.addAll(localSensInds.get(vDesc.getName()));
			}
			newGlobalList.addAll(globalSensInds);
			globalSensInds.clear();
			globalSensInds.addAll(newGlobalList);
		}

		for (Individual ind : globalSensInds) {
			for (String rName : this.parentMappingManager.getSensitivityTargetNames()) {
				TSensitivityCell curCell = globalSensResults.get(rName);

				Double cValue = ind.getResult(rName);
				if (!Double.isNaN(cValue)) {
					curCell.updateAbsoluteMinResToDisplay(cValue, baseInd.getResult(rName));
					curCell.updateAbsoluteMaxResToDisplay(cValue, baseInd.getResult(rName));
				}

			}
		}

		for (Individual ind : globalSensInds) {

			for (VariableDesc vDesc : this.parentMappingManager.getProblem().getVariedVariables()) {

				double baseIndividualVarValue = Math.abs(baseInd.getVariable(vDesc.getName()));

				if (baseIndividualVarValue == 0) {
					baseIndividualVarValue = (this.parentMappingManager.getProblem().getVariable(vDesc.getName())
							.getUB() - this.parentMappingManager.getProblem().getVariable(vDesc.getName()).getLB());
				}

				if (ind.getVariable(vDesc.getName()) < varRanges.get(vDesc.getName()).getVarMin()) {
					varRanges.get(vDesc.getName()).setVarMin(ind.getVariable(vDesc.getName()));
				}

				if (ind.getVariable(vDesc.getName()) > varRanges.get(vDesc.getName()).getVarMax()) {
					varRanges.get(vDesc.getName()).setVarMax(ind.getVariable(vDesc.getName()));
				}

				varRanges.get(vDesc.getName()).setVarAbsRange(
						varRanges.get(vDesc.getName()).getVarMax() - varRanges.get(vDesc.getName()).getVarMin());

				varRanges.get(vDesc.getName()).setVarPercentualRange(
						varRanges.get(vDesc.getName()).getVarAbsRange() / baseIndividualVarValue);

				if (varRanges.get(vDesc.getName()).getVarPercentualRange() > maxVarRange) {
					maxVarRange = varRanges.get(vDesc.getName()).getVarPercentualRange();
				}
			}

			for (String targetName : this.parentMappingManager.getSensitivityTargetNames()) {
				if (ind.getResult(targetName) < resRanges.get(targetName).getResMin()) {
					resRanges.get(targetName).setResMin(ind.getResult(targetName));
				}

				if (ind.getResult(targetName) > resRanges.get(targetName).getResMax()) {
					resRanges.get(targetName).setResMax(ind.getResult(targetName));
				}

				resRanges.get(targetName)
						.setResAbsRange(resRanges.get(targetName).getResMax() - resRanges.get(targetName).getResMin());

				resRanges.get(targetName).setResPercentualRange(
						resRanges.get(targetName).getResAbsRange() / Math.abs(baseInd.getResult(targetName)));
			}

		}

		for (String targetName : this.parentMappingManager.getSensitivityTargetNames()) {
			TSensitivityCell curCell = globalSensResults.get(targetName);

			curCell.setResRelativeRange(resRanges.get(targetName).getResPercentualRange() / maxVarRange);
		}

		// for (String targetName : this.parentMappingManager
		// .getSensitivityTargetNames()) {
		// System.out.println(targetName + " "
		// + resRanges.get(targetName).getResAbsRange() + " "
		// + resRanges.get(targetName).getResPercentualRange());
		// }
		//
		// System.out.println("FE: " + this.getBaseIndividual("FE"));
		// System.out.println("MAPPINTG: " + this.getBaseIndividual("MAPPING"));
	}

	public void postProcessMappingBasedSensitivityInformation() {

		this.postProcessSensitivityInformation(this.getBaseIndividual("MAPPING"), this.localSensitivityIndividuals,
				this.globalSensitivityIndividuals, this.localSensitivityResultsMappings,
				this.globalSensitivityResultsMappings);
		this.containsMappingResults = true;

	}

	public void postProcessFEBasedSensitivityInformation(boolean allFEResultsComputed) {

		this.postProcessSensitivityInformation(this.getBaseIndividual("FE"), this.localSensitivityIndividuals,
				this.globalSensitivityIndividuals, this.localSensitivityResultsFE, this.globalSensitivityResultsFE);
		if (allFEResultsComputed) {
			this.waitingForFEAnalysis = false;
			this.containsFEResults = true;
		}
	}

	public void performPostDeserializationActions(MappingManager parentManager) {
		this.parentMappingManager = parentManager;
		this.localSensitivityIndividuals = new HashMap<String, List<Individual>>();
		this.globalSensitivityIndividuals = new ArrayList<Individual>();

		this.restoreSavedIndividualsFromVarValues();
		// this.waitingForFEAnalysis = false;

		if (this.waitingForFEAnalysis) {
			this.parentMappingManager.addLog("Waiting for FE results - recompute memory size: "
					+ this.parentMappingManager.getProblem().getArchive().getRecomputeMemory().getNumIndividuals());
		}

		this.checkFEProgress();
	}

	public Individual getBaseIndividual(String dataOrigin) {
		if (dataOrigin.equals("FE")) {
			return this.createIndividualFromInfo(this.baseIndividualFEResultsInfo);
		}
		Individual ind = this.createIndividualFromInfo(this.baseIndividualFEResultsInfo);
		this.parentMappingManager.evaluateRegressionTargetsOfIndividual(ind);

		return ind;
	}

	private String generateDisplayName() {
		String result = "| ";

		for (ResultDesc rDesc : this.parentMappingManager.getProblem().getResults()) {
			if (rDesc.isObjective() || rDesc.isConstraint()) {
				result += rDesc.getName().substring(rDesc.getName().lastIndexOf(".") + 1) + "="
						+ String.format("%.4f", this.baseIndividualFEResultsInfo.getResValue(rDesc.getName())) + " | ";
			}
		}

		for (VariableDesc vDesc : this.parentMappingManager.getProblem().getVariedVariables()) {
			result += vDesc.getName().substring(vDesc.getName().lastIndexOf(".") + 1) + "="
					+ String.format("%.4f", this.baseIndividualFEResultsInfo.getVarValue(vDesc.getName())) + " | ";
		}

		return result;
	}

	public String getDisplayName() {
		return this.displayName;
	}

	public boolean getContainsFEResults() {
		return this.containsFEResults;
	}

	public boolean getContainsMappingResults() {
		return this.containsMappingResults;
	}

	public boolean getIsWaitingForFEResults() {
		return this.waitingForFEAnalysis;
	}

	public void setIsWaitingForFEResults(boolean isWaitingForFEResults) {
		this.waitingForFEAnalysis = isWaitingForFEResults;
	}

	public void addSensitivityIndividualsToSensitivityMemory() {
		// the global individuals list also contains the local individuals

		this.parentMappingManager.addLog("Initial recompute memory size: "
				+ this.parentMappingManager.getProblem().getArchive().getRecomputeMemory().getNumIndividuals());

		int indsToAdd = this.globalSensitivityIndividuals.size();

		// List<Individual> allLocalIndividuals = new ArrayList<Individual>();
		// for (VariableDesc vDesc : this.parentMappingManager.getProblem()
		// .getVariedVariables()) {
		// List<Individual> individuals = this.localSensitivityIndividuals
		// .get(vDesc.getName());
		// indsToAdd += individuals.size();
		// allLocalIndividuals.addAll(individuals);
		// }
		// this.globalSensitivityIndividuals.addAll(allLocalIndividuals);

		this.parentMappingManager.addLog("Adding: " + indsToAdd + " individuals to recompute memory!");

		int i = 0;
		for (Individual ind : this.globalSensitivityIndividuals) {
			this.parentMappingManager.getProblem().getArchive().getRecomputeMemory().addIndividual(ind);
			i++;
		}
		this.parentMappingManager.addLog("Added: " + i + " individuals to recompute memory!");

		this.parentMappingManager.addLog("New recompute memory size: "
				+ this.parentMappingManager.getProblem().getArchive().getRecomputeMemory().getNumIndividuals());
	}

	public void saveIndVarValues() {

		if (!(this.waitingForFEAnalysis) && !(this.containsFEResults)) {

			this.persistedLocalSensitivityIndVarValues = new HashMap<String, List<Double[]>>();
			this.persistedGlobalSensitivityIndVarValues = new ArrayList<Double[]>();

			for (VariableDesc vDesc : this.parentMappingManager.getProblem().getVariedVariables()) {
				List<Double[]> localValuesList = new ArrayList<Double[]>();
				List<Individual> individuals = this.localSensitivityIndividuals.get(vDesc.getName());
				for (Individual ind : individuals) {
					Double[] indVarValues = new Double[this.parentMappingManager.getProblem()
							.getVariedVariables().length];
					int i = 0;
					for (VariableDesc vd : this.parentMappingManager.getProblem().getVariedVariables()) {
						indVarValues[i] = ind.getVariable(vd.getName());
						i++;
					}
					localValuesList.add(indVarValues);
				}
				this.persistedLocalSensitivityIndVarValues.put(vDesc.getName(), localValuesList);
			}

			/** used to create a mock list of individuals for testing purposes */
			// int k = 10;
			// List<Individual> shortList = new ArrayList<Individual>();
			// for (Individual ind : this.globalSensitivityIndividuals) {
			// shortList.add(ind);
			// k--;
			// if (k == 0) {
			// break;
			// }
			// }
			// this.globalSensitivityIndividuals.clear();
			// this.globalSensitivityIndividuals.addAll(shortList);

			this.parentMappingManager.addLog(
					"Saved: " + this.globalSensitivityIndividuals.size() + "individuals in the recompute memeory");

			for (Individual ind : this.globalSensitivityIndividuals) {
				Double[] indVarValues = new Double[this.parentMappingManager.getProblem().getVariedVariables().length];
				int i = 0;
				for (VariableDesc vd : this.parentMappingManager.getProblem().getVariedVariables()) {
					indVarValues[i] = ind.getVariable(vd.getName());
					i++;
				}
				this.persistedGlobalSensitivityIndVarValues.add(indVarValues);
			}

			// System.out.println("Saved global individuals: "
			// + this.persistedGlobalSensitivityIndVarValues.size());
		} else {
			this.parentMappingManager.addLog(
					"No new individuals have been saved in the recompute memory! Current sensitivity individuals in recompute memory: "
							+ this.globalSensitivityIndividuals.size());
		}
	}

	public void restoreSavedIndividualsFromVarValues() {
		this.localSensitivityIndividuals = new HashMap<String, List<Individual>>();
		for (VariableDesc vDesc : this.parentMappingManager.getProblem().getVariedVariables()) {
			List<Individual> localIndividuals = new ArrayList<Individual>();
			List<Double[]> localValuesList = this.persistedLocalSensitivityIndVarValues.get(vDesc.getName());
			for (Double[] vals : localValuesList) {
				try {
					Individual ind = new Individual(this.parentMappingManager.getProblem().getArchive());
					int i = 0;
					for (VariableDesc vd : this.parentMappingManager.getProblem().getVariedVariables()) {
						ind.setVariable(vd.getName(), vals[i]);
						i++;
					}
					ind.setFromMapping(true);
					for (VariableDesc vd : this.parentMappingManager.getProblem().getVariables()) {
						if (!vd.isVaried()) {
							ind.setVariable(vd.getName(),
									this.parentMappingManager.getProblem().getVariable(vd.getName()).getLB());
						}
					}

					localIndividuals.add(ind);
				} catch (Exception ex) {
					System.out.println("Error by creation individium");
				}

			}
			this.localSensitivityIndividuals.put(vDesc.getName(), localIndividuals);
		}

		this.globalSensitivityIndividuals = new ArrayList<Individual>();
		for (Double[] vals : this.persistedGlobalSensitivityIndVarValues) {
			try {
				Individual ind = new Individual(this.parentMappingManager.getProblem().getArchive());
				int i = 0;
				for (VariableDesc vd : this.parentMappingManager.getProblem().getVariedVariables()) {
					ind.setVariable(vd.getName(), vals[i]);
					i++;
				}
				ind.setFromMapping(true);
				for (VariableDesc vd : this.parentMappingManager.getProblem().getVariables()) {
					if (!vd.isVaried()) {
						ind.setVariable(vd.getName(),
								this.parentMappingManager.getProblem().getVariable(vd.getName()).getLB());
					}
				}
				this.globalSensitivityIndividuals.add(ind);
			} catch (Exception ex) {
				System.out.println("Error by creation individium");
			}
		}
	}

	public String getFESensitivityProgress() {
		String result = "Check FE prog. (last = 0.00%)";

		if (this.FESensitivityProgress >= 0) {
			result = "Check FE prog. (last = " + String.format("%.2f", this.FESensitivityProgress) + "%)";
		} else {
			result = "Check FE prog. (last = 0.00%)";
		}

		return result;
	}

	public void checkFEProgress() {

		if (!this.containsFEResults || this.waitingForFEAnalysis) {

			int indToEvaluate = this.globalSensitivityIndividuals.size();
			this.parentMappingManager.addLog("Individuals to evaluate: " + indToEvaluate);

			if (indToEvaluate > 0) {

				int evaluatedIndividuals = 0;
				int notEvaluatedIndividuals = 0;
				int noMatches = 0;
				for (Individual ind : this.globalSensitivityIndividuals) {
					Individual memInd = this.parentMappingManager.getProblem().getArchive().getRecomputeMemory()
							.getIndividual(ind);

					if (memInd == null) {
						noMatches++;
					} else {
						if (!memInd.isFromMapping()) {
							evaluatedIndividuals++;
						} else {
							notEvaluatedIndividuals++;
						}
					}
				}

				this.parentMappingManager.addLog("No match individuals: " + noMatches);
				this.parentMappingManager.addLog("Evaluated individuals: " + evaluatedIndividuals);
				this.parentMappingManager.addLog("Not evaluated individuals: " + notEvaluatedIndividuals);

				this.FESensitivityProgress = ((evaluatedIndividuals * 1.0) / indToEvaluate) * 100;

				// for (Individual ind : this.globalSensitivityIndividuals) {
				// System.out.println("-----");
				// System.out.println("SH ind:");
				// this.printIndividulInfo(ind);
				// for (Individual indMem :
				// this.parentMappingManager.getProblem()
				// .getArchive().getRecomputeMemory().getIndividual()) {
				// int equalsCount = 0;
				// for (int j = 0; j < this.parentMappingManager.getProblem()
				// .getNumVariables(); j++) {
				// double val1 = ind.getVariable(j);
				// double val2 = indMem.getVariable(j);
				// if (val1 == val2) {
				// equalsCount++;
				// // System.out.println(val1 + " " + val2 + " "
				// // + equalsCount);
				// }
				// }
				// if (equalsCount == this.parentMappingManager.getProblem()
				// .getNumVariables()) {
				// // System.out.println(equalsCount);
				// // this.printIndividulInfo(indMem);
				// //
				// // System.out.println("+++++");
				// // System.out.println(ind);
				// // System.out.println(indMem);
				// // System.out.println("+++++");
				// System.out.println("Mem ind:");
				// this.printIndividulInfo(indMem);
				// }
				//
				// }
				// System.out.println("-----");
				// }

				// for (Individual ind : this.globalSensitivityIndividuals) {
				// System.out.println(ind);
				// }
				// System.out.println("-----------------------------------");
				// for (Individual ind : this.parentMappingManager.getProblem()
				// .getArchive().getRecomputeMemory().getIndividual()) {
				// System.out.println(ind);
				// }
			} else {
				this.FESensitivityProgress = -1;
			}

			if (this.FESensitivityProgress == 100.0) {
				this.persistIndividualFEResults();
				this.postProcessFEBasedSensitivityInformation(true);
				this.parentMappingManager.saveFEBasedSensitivityInformation();
			} else {
				// this.simulateFEEvaluation(75);

				if (this.FESensitivityProgress > 3.0) {
					this.persistIndividualFEResults();
					this.postProcessFEBasedSensitivityInformation(false);
					this.containsFEResults = true;
				}

			}
		}
	}

	private void persistIndividualFEResults() {

		for (VariableDesc vDesc : this.parentMappingManager.getProblem().getVariedVariables()) {
			for (Individual ind : this.localSensitivityIndividuals.get(vDesc.getName())) {
				Individual memInd = this.parentMappingManager.getProblem().getArchive().getRecomputeMemory()
						.getIndividual(ind);

				if (memInd != null) {
					for (ResultDesc rDesc : this.parentMappingManager.getProblem().getResults()) {
						ind.setResult(rDesc.getName(), memInd.getResult(rDesc.getName()));

						// if (!Double.isNaN(memInd.getResult(rDesc.getName())))
						// {
						// System.out.println("LR set! :"
						// + memInd.getResult(rDesc.getName()));
						// }
					}
				}
			}

		}

		for (Individual ind : this.globalSensitivityIndividuals) {
			Individual memInd = this.parentMappingManager.getProblem().getArchive().getRecomputeMemory()
					.getIndividual(ind);

			if (memInd != null) {
				for (ResultDesc rDesc : this.parentMappingManager.getProblem().getResults()) {
					ind.setResult(rDesc.getName(), memInd.getResult(rDesc.getName()));

					// if (!Double.isNaN(memInd.getResult(rDesc.getName()))) {
					// System.out.println("GR set! :"
					// + memInd.getResult(rDesc.getName()));
					// }
				}
			}
		}

	}

	public void evaluateSensitivityUsingMappings() {
		for (VariableDesc vDesc : this.parentMappingManager.getProblem().getVariedVariables()) {
			for (Individual ind : this.localSensitivityIndividuals.get(vDesc.getName())) {
				this.parentMappingManager.evaluateRegressionTargetsOfIndividual(ind);
			}

		}

		for (Individual ind : this.globalSensitivityIndividuals) {
			this.parentMappingManager.evaluateRegressionTargetsOfIndividual(ind);
		}
	}
}
