/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 247 $
 * $Date: 2020-03-19 10:13:52 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.optim.core.project;

import java.io.Serializable;
import java.util.LinkedList;

import at.lcm.bu21.general.dtypes.TArray;
import at.lcm.bu21.general.dtypes.TString;
import at.lcm.bu21.general.dtypes.TStruct;
import at.lcm.bu21.general.utils.Pair;
import at.lcm.bu21.optim.core.Problem;
import at.lcm.bu21.optim.core.event.OptimizerEvent;

/**
 * Plot description of a user-defined plot
 * 
 * @author Werner Koppelsttter
 */
public class PlotDescription implements Serializable {

	/** Name of plot */
	private String name;

	/** Problem of plot */
	private Problem problem;

	/** List of axis-parameter for all plots in sub-plot */
	private LinkedList<LinkedList<Pair<String, String>>> axes;

	/**
	 * Create a user-plot description
	 * 
	 * @param problem
	 *            Plot problem
	 * @param name
	 *            Name of plot
	 */
	public PlotDescription(Problem problem, String name) {
		this.name = name;
		this.problem = problem;
		this.axes = new LinkedList<LinkedList<Pair<String, String>>>();
		this.axes.add(new LinkedList<Pair<String, String>>());
		this.axes.get(0).add(new Pair<String, String>("", ""));
	}

	/**
	 * Resume plot-description
	 * 
	 * @param problem
	 *            Problem of plot
	 * @param resumeData
	 *            Resume data
	 * @throws Exception
	 */
	public PlotDescription(Problem problem, TStruct resumeData) throws Exception {
		this.problem = problem;
		this.setResumeData(resumeData);
	}

	/**
	 * Get name of plot
	 * 
	 * @return Name
	 */
	public String getName() {
		return this.name;
	}

	/**
	 * Get number of rows
	 * 
	 * @return Number of rows
	 */
	public int getNumRows() {
		return this.axes.size();
	}

	/**
	 * Get number of columns
	 * 
	 * @return Number of columns
	 */
	public int getNumColumns() {
		return this.axes.get(0).size();
	}

	/**
	 * Get name of x-axis
	 * 
	 * @param row
	 *            Row of plot
	 * @param column
	 *            Column of plot
	 * @return Name of x-axis
	 */
	public String getXAxis(int row, int column) {
		return this.axes.get(row).get(column).getFirst();
	}

	/**
	 * Set name of x-axis
	 * 
	 * @param row
	 *            Row of plot
	 * @param column
	 *            Column of plot
	 * @param identifier
	 *            Name of x-axis
	 */
	public void setXAxis(int row, int column, String identifier) {
		this.axes.get(row).get(column).setFirst(identifier);
		if (this.problem.getOptimization() != null) {
			this.problem.getOptimization()
					.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.PLOT, "Init: " + this.name));
		}
	}

	/**
	 * Get name of y-axis
	 * 
	 * @param row
	 *            Row of plot
	 * @param column
	 *            Column of plot
	 * @return Name of y-axis
	 */
	public String getYAxis(int row, int column) {
		return this.axes.get(row).get(column).getSecond();
	}

	/**
	 * Set name of y-axis
	 * 
	 * @param row
	 *            Row of plot
	 * @param column
	 *            Column of plot
	 * @param identifier
	 *            Name of y-axis
	 */
	public void setYAxis(int row, int column, String identifier) {
		this.axes.get(row).get(column).setSecond(identifier);
		if (this.problem.getOptimization() != null) {
			this.problem.getOptimization()
					.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.PLOT, "Init: " + this.name));
		}
	}

	/**
	 * Add row of plots
	 * 
	 * @param index
	 *            Index to insert
	 */
	public void addRow(int index) {
		LinkedList<Pair<String, String>> row;

		row = new LinkedList<Pair<String, String>>();
		for (int k = 0; k < this.getNumColumns(); k++) {
			row.add(new Pair<String, String>("", ""));
		}
		this.axes.add(index, row);
		if (this.problem.getOptimization() != null) {
			this.problem.getOptimization()
					.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.PLOT, "Init: " + this.name));
		}
	}

	/**
	 * Remove row of plots
	 * 
	 * @param index
	 *            Index to remove
	 */
	public void rmRow(int index) {
		if (this.axes.size() > 1) {
			this.axes.remove(index);
			if (this.problem.getOptimization() != null) {
				this.problem.getOptimization()
						.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.PLOT, "Init: " + this.name));
			}
		}
	}

	/**
	 * Add column of plots
	 * 
	 * @param index
	 *            Index to insert
	 */
	public void addColumn(int index) {
		for (LinkedList<Pair<String, String>> row : this.axes) {
			row.add(index, new Pair<String, String>("", ""));
		}
		if (this.problem.getOptimization() != null) {
			this.problem.getOptimization()
					.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.PLOT, "Init: " + this.name));
		}
	}

	/**
	 * Remove column of plots
	 * 
	 * @param index
	 *            Index to remove
	 */
	public void rmColumn(int index) {
		if (this.axes.get(0).size() > 1) {
			for (LinkedList<Pair<String, String>> row : this.axes) {
				row.remove(index);
			}
			if (this.problem.getOptimization() != null) {
				this.problem.getOptimization()
						.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.PLOT, "Init: " + this.name));
			}
		}
	}

	/**
	 * Get resume data
	 * 
	 * @return Resume data
	 * @throws Exception
	 */
	public TStruct getResumeData() throws Exception {
		TStruct resumeData;
		TArray array;
		TStruct plot;
		resumeData = new TStruct();
		Pair<String, String> pair;

		// name
		resumeData.put("name", new TString(this.name));

		// axis
		array = new TArray(this.getNumRows(), this.getNumColumns());
		resumeData.put("axes", array);
		for (int r = 0; r < this.getNumRows(); r++) {
			for (int c = 0; c < this.getNumColumns(); c++) {
				plot = new TStruct();
				array.set(r, c, plot);
				pair = this.axes.get(r).get(c);
				plot.put("x", new TString(pair.getFirst()));
				plot.put("y", new TString(pair.getSecond()));
			}
		}

		return resumeData;
	}

	/**
	 * Set resume data
	 * 
	 * @param resumeData
	 *            Resume Data
	 * @throws Exception
	 */
	private void setResumeData(TStruct resumeData) throws Exception {
		TArray array;
		TStruct plot;

		// name
		this.name = ((TString) resumeData.get("name")).getString();

		// variable description
		array = (TArray) resumeData.get("axes");
		this.axes = new LinkedList<LinkedList<Pair<String, String>>>();
		for (int r = 0; r < array.size()[0]; r++) {
			this.axes.add(new LinkedList<Pair<String, String>>());
			for (int c = 0; c < array.size()[1]; c++) {
				plot = (TStruct) array.get(r, c);
				this.axes.get(r).add(new Pair<String, String>(((TString) plot.get("x")).getString(),
						((TString) plot.get("y")).getString()));
			}
		}
	}

}
