#!/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 9880 $
$LastChangedDate:: 2020-06-17 19:08:35 +0200#$

This file is part of SyMSpace.

Copyright (c) 2020, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""

import os
import base64
import pickle
import webbrowser
import tempfile
import shutil
from io import BytesIO
from matplotlib.figure import Figure as _Figure
import tkinter as tk
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2Tk
import symspace.gui.qt
from qtpy import QtWidgets
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg, NavigationToolbar2QT

symspace.gui.qt.set_QT_PLUGIN_PATH()


class FigureMPL:
    """Helper class to preview Matplot graphs in SyMSpace"""

    @staticmethod
    def __getTempDir() -> str:
        tmp_dir = os.path.join(tempfile.gettempdir(), 'tmp_SyMSpace_Preview')
        return tmp_dir

    @classmethod
    def __makeTempDir(cls) -> str:
        tmp_dir = cls.__getTempDir()
        if not os.path.isdir(tmp_dir):
            os.makedirs(tmp_dir)
        return tmp_dir

    @classmethod
    def removeTempDir(cls):
        # Todo: probably this method should be called at some point to free space on the drive,
        #  not so important if data handover is not file based -> pickled base64 obj
        tmp_dir = cls.__getTempDir()
        if os.path.isdir(tmp_dir):
            shutil.rmtree(tmp_dir)

    @classmethod
    def fromPickle(cls, file: str) -> 'FigureMPL':
        """
        Instantiate the class from a pickle file

        Parameters
        ----------
        file : str
            Filepath of a pickled class object

        Returns
        -------
        obj : FigureMPL
        """
        with open(file, "rb") as fh:
            figure = pickle.load(fh)
        return cls(figure)

    @classmethod
    def fromBase64Pickle(cls, encoded_pickle_obj: str) -> 'FigureMPL':
        """
        Instantiate the class from a base64 pickle string

        Deserialize a base64 encoded
        string representation of a pickle instance
        The serialized instance gets base64 decoded and converted to a figure


        Parameters
        ----------
        encoded_pickle_obj : str
            Serialized object

        Returns
        -------
        obj : FigureMPL
        """

        ''' decode string to byte stream '''
        pickle_obj_string = base64.b64decode(encoded_pickle_obj)
        figure = pickle.loads(pickle_obj_string)
        return cls(figure)

    @classmethod
    def openHTML(cls, html: str, suffix: str = '.html'):
        """
        Open a html string in a webbrowser

        Parameters
        ----------
        html : str
            html string
        suffix : str
            html file suffix
        """
        with tempfile.NamedTemporaryFile(dir=cls.__makeTempDir(), mode='w+t', suffix=suffix, delete=False) as temp:
            temp.write(html)
            webbrowser.open('file://' + temp.name, new=2)

    def __init__(self, figure: _Figure):
        self.__figure = figure

    def toPickle(self) -> str:
        """
        Save a class instance as pickle file

        Returns
        -------
        filepath : str
            Filepath of the pickled object
        """
        with tempfile.NamedTemporaryFile(dir=self.__makeTempDir(), mode='w+t', suffix='.p', delete=False) as temp:
            with open(temp.name, 'wb') as fh:
                pickle.dump(self.__figure, fh)
        return temp.name

    def toBase64Pickle(self) -> str:
        """
        Serialize this object to a string

        Returns
        -------
        str
        """
        pickle_obj_string = pickle.dumps(self.__figure)
        ''' encode to base64 '''
        encoded = base64.b64encode(pickle_obj_string)
        ''' decode to string'''
        encoded_pickle_obj = encoded.decode('utf-8')
        return encoded_pickle_obj

    def toBase64PNG(self, dpi: int = None) -> str:
        """
        Convert Figure to a base64 PNG string

        Parameters
        ----------
        dpi : int
            image resolution in dots per inch

        Returns
        -------
        data : str
            base64 encoded PNG
        """
        # Save it to a temporary buffer.
        buf = BytesIO()
        self.__figure.savefig(buf, format="png", dpi=dpi)

        # Embed the result in the html output.
        data = base64.b64encode(buf.getbuffer()).decode("utf-8")
        return data

    def toHTML(self, figure_format='png', dpi: int = None) -> str:
        """
        Convert Figure to a HTML string

        Parameters
        ----------
        figure_format : str
            Format of the embedded image, 'png' or 'svg'
        dpi : int
            image resolution in dots per inch

        Returns
        -------
        html : str
            HTML string
        """
        if figure_format.lower() == 'png':
            html = self.__toPNGHTML(dpi=dpi)
        elif figure_format.lower() == 'svg':
            html = self.__toSVGHTML(dpi=dpi)
        else:
            raise NotImplementedError()
        return html

    def showHTML(self, figure_format='png', dpi: int = None):
        """
        Show Figure in the webbrowser

        Parameters
        ----------
        figure_format : str
            Format of the embedded image, 'png' or 'svg'
        dpi : int
            image resolution in dots per inch

        Returns
        -------
        html : str
            HTML string
        """
        html = self.toHTML(figure_format=figure_format, dpi=dpi)
        if isinstance(dpi, int):
            suffix = '_{}_dpi{:d}.html'.format(figure_format, dpi)
        else:
            suffix = '_{}.html'.format(figure_format)
        self.openHTML(html, suffix=suffix)

    def showQT(self):
        """Show Figure in a Qt Window"""
        app = QtWidgets.QApplication([])
        w = QtWindowMPL(self.__figure)
        w.show()
        app.exec_()

    def showTK(self):
        """Show Figure in a Tk Window"""
        w = TkWindowMPL(self.__figure)
        w.mainloop()

    def toSyMPreview(self, title: str, framework: str = 'qt', file_based: bool = False) -> tuple:
        """
        Return a PNG preview for SyMSpace, calls the SyMSpace function method "doubleClickMPL" on doubleclick

        Parameters
        ----------
        title : str
            preview tab title
        framework : str
            plot framework, 'qt', 'tk or 'html'
        file_based : bool
            Data handover as file (True) or pickled string (False, Default)

        Returns
        -------
        tuple
        """
        if file_based:
            fig_pickle = self.toPickle()
        else:
            fig_pickle = self.toBase64Pickle()

        ret = (title, 'image/png', self.toBase64PNG(), 'doubleClickMPL', fig_pickle, framework.lower(), file_based)
        return ret

    @staticmethod
    def doubleClick(fig_pickle: str, framework: str, file_based: bool):
        """Specific double click function for a preview tab
        """
        if file_based:
            fig = FigureMPL.fromPickle(fig_pickle)
        else:
            fig = FigureMPL.fromBase64Pickle(fig_pickle)

        if framework == 'qt':
            fig.showQT()
        elif framework == 'tk':
            fig.showTK()
        else:
            fig.showHTML()

    def __toPNGHTML(self, dpi) -> str:
        data = self.toBase64PNG(dpi=dpi)
        html = f"<img style='max-width:100%;max-height:100%;height:100%' src='data:image/png;base64,{data}'/>"
        return html

    def __toSVGHTML(self, dpi) -> str:
        # Save it to a temporary buffer.
        buf = BytesIO()
        self.__figure.savefig(buf, format="svg", dpi=dpi)

        # Embed the result in the html output.
        html = buf.getvalue().decode("utf-8")
        return html


class QtWindowMPL(QtWidgets.QMainWindow):

    def __init__(self, figure: _Figure, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Create the maptlotlib FigureCanvas object
        sc = FigureCanvasQTAgg(figure)
        # self.setCentralWidget(sc)

        # Create toolbar, passing canvas as first parameter, parent (self, the MainWindow) as second.
        toolbar = NavigationToolbar2QT(sc, self)
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(toolbar)
        layout.addWidget(sc)

        # Create a placeholder widget to hold our toolbar and canvas.
        widget = QtWidgets.QWidget()
        widget.setLayout(layout)
        self.setCentralWidget(widget)
        self.setWindowTitle('Matplot in QT')


class TkWindowMPL(tk.Tk):

    def __init__(self, figure: _Figure):
        super().__init__()

        self.wm_title("Matplot in TK")
        canvas = FigureCanvasTkAgg(figure, master=self)
        canvas.get_tk_widget().pack(side=tk.TOP, fill=tk.BOTH, expand=1)

        toolbar = NavigationToolbar2Tk(canvas, self)
        toolbar.update()
        # canvas._tkcanvas.pack(side=tk.TOP, fill=tk.BOTH, expand=1)


if __name__ == '__main__':
    import matplotlib.pyplot as plt

    fig = plt.figure()
    plt.plot([0, 1, 2, 3, 4], [10, 1, 20, 3, 40])

    fig_mpl = FigureMPL(fig)
    fig_mpl.showHTML()
    fig_mpl.showQT()
    fig_mpl.showTK()

    p_file = fig_mpl.toPickle()
    fig_mpl2 = FigureMPL.fromPickle(p_file)

    fig_mpl.showQT()
