/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9767 $
 * $Date: 2020-04-20 20:04:00 +0200 (Mon, 20 Apr 2020) $
 */
package jobmanager;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.net.Socket;
import java.net.UnknownHostException;
import java.util.Arrays;
import java.util.Hashtable;
import java.util.LinkedList;

import at.lcm.bu21.general.utils.FileUtils;
import at.lcm.bu21.magopt.utils.MagOptConfig;

/**
 * Description and interface of a job for JobManager.
 * 
 * @author Werner Koppelsttter
 */
public class JobDesc {

	/**
	 * Get port of JobManager
	 * 
	 * @return Port of JobManager
	 */
	public static int getPort() {
		return Integer.parseInt(MagOptConfig.getConfig("JobManagerPort"));
	}

	/**
	 * Get debug-port of JobManager
	 * 
	 * @return Debug-port of JobManager
	 */
	public static int getDebugPort() {
		int port = -1;
		String str;

		str = MagOptConfig.getConfig("JobManagerDebugPort");
		if (str != null) {
			port = Integer.parseInt(str);
		}

		return port;
	}

	/**
	 * Get JobMangaer working-directory
	 * 
	 * @return Working directory
	 */
	public static File getWorkingDir() {
		String str;

		str = MagOptConfig.getConfig("JobManagerTempDir");
		if (str == null) {
			str = System.getProperty("java.io.tmpdir");
		}

		return new File(str, "JobManager");
	}

	/**
	 * Connect to JobManager or start new JobManager
	 * 
	 * @return Connection to JobManager
	 */
	private static Socket getSocket() {
		Socket s = null;
		File path, jep;
		StringBuffer cp;
		ProcessBuilder pb;
		int repeat, debug;
		File tmpFolder;
		LinkedList<String> cmd;
		String home;

		while (s == null) {
			repeat = 5;
			// try to connect to JobManager
			while ((s == null) && (repeat-- > 0)) {
				try {
					s = new Socket("", JobDesc.getPort());
				} catch (Exception e) {
					try {
						Thread.sleep(100);
					} catch (Exception ex) {
					}
				}
			}

			// start JobManager
			if (s == null) {

				// clear temp-folder (javasysmon, temp)
				tmpFolder = new File(System.getProperty("java.io.tmpdir"));
				for (File f : tmpFolder.listFiles()) {
					if (f.isFile()) {
						if (f.getName().startsWith("javasysmo") && f.getName().endsWith(".dll")) {
							f.delete();
						} else if (f.getName().startsWith("temp") && f.getName().endsWith(".jar")) {
							f.delete();
						}
					}
				}

				home = MagOptConfig.getHome();

				// start jobmanager
				cp = new StringBuffer();
				for (File f : FileUtils.getFile(home, "Tools", "JobManager", "Java").listFiles()) {
					if (f.getName().toLowerCase().endsWith(".jar")) {
						cp.append(f.getPath());
						cp.append(File.pathSeparator);
					}
				}
				cp.append(FileUtils.getFile(home, "System", "Java", "magopt.jar").getPath());
				cp.append(File.pathSeparator);
				cp.append(FileUtils.getFile(home, "External", "javasysmon", "javasysmon-0.3.3.jar").getPath());
				cp.append(File.pathSeparator);
				cp.append(FileUtils.getFile(home, "System", "General", "Java", "general.jar").getPath());
				cp.append(File.pathSeparator);
				cp.append(FileUtils.getFile(home, "System", "General", "External", "commons", "commons-math3-3.0.jar")
						.getPath());

				JobDesc.addClasspath(FileUtils.getFile(home, "System", "General", "External", "jfreechart"), cp);
				JobDesc.addClasspath(FileUtils.getFile(home, "System", "General", "External", "freehep"), cp);
				JobDesc.addClasspath(FileUtils.getFile(home, "Tools", "JobManager", "External"), cp);

				// TODO
				jep = MagOptConfig.getJepPath();
				if (jep != null) {
					for (File f : jep.listFiles()) {
						if (f.getName().toLowerCase().endsWith(".jar")) {
							cp.append(File.pathSeparator);
							cp.append(f.getPath());
						}
					}
				}

				path = FileUtils.getFile(home, "Tools", "JobManager");

				cmd = new LinkedList<String>();
				debug = JobDesc.getDebugPort();
				cmd.add(new File(new File(System.getProperty("java.home"), "bin"), "java").getPath());
				cmd.add("-cp");
				cmd.add(cp.toString());
				// TODO
				cmd.add("-Djava.library.path=" + jep.getPath());

				if (debug != -1) {
					cmd.add("-Xdebug");
					cmd.add("-Xrunjdwp:transport=dt_socket,server=y,suspend=n,address=" + debug);
				}
				cmd.add("jobmanager.server.JobManager");
				try {
					pb = new ProcessBuilder(cmd);
					pb.directory(path);
					pb.start();
					Thread.sleep(1000);
				} catch (Exception e) {
				}
			}
		}

		return s;
	}

	/**
	 * Add recursive jar-files to classpath
	 * 
	 * @param path
	 * @param cp
	 */
	private static void addClasspath(File path, StringBuffer cp) {
		if (path.exists()) {
			for (File f : path.listFiles()) {
				if (f.isDirectory()) {
					JobDesc.addClasspath(f, cp);
				} else if (f.isFile() && f.getName().toLowerCase().endsWith(".jar")) {
					cp.append(File.pathSeparator + f.getPath());
				}
			}
		}
	}

	/**
	 * Set visible of the JobManager-Gui
	 * 
	 * @param visible
	 *            Visible
	 */
	public static void setVisible(boolean visible) {
		Socket s = null;
		ObjectOutputStream out = null;
		ObjectInputStream in = null;

		try {
			s = JobDesc.getSocket();
			out = new ObjectOutputStream(s.getOutputStream());
			in = new ObjectInputStream(s.getInputStream());

			out.writeObject(Commands.VIEW);
			out.writeBoolean(visible);

		} catch (IOException e) {
		} finally {
			if (out != null) {
				try {
					out.close();
				} catch (IOException e) {
				}
			}
			if (in != null) {
				try {
					in.close();
				} catch (IOException e) {
				}
			}
			if (s != null) {
				try {
					s.close();
				} catch (IOException e) {
				}
			}
		}
	}

	/**
	 * Create a job on the JobManager
	 * 
	 * @param project
	 *            Name of project
	 * @param dataset
	 *            Dataset-name
	 * @param parameter
	 *            Parameter-name
	 * @param name
	 *            Name of job
	 * @param args
	 *            Arguments of job
	 * @return Identifier of job
	 */
	public static String createJob(String project, String dataset, String parameter, String name, String... args) {
		LinkedList<String> arguments = new LinkedList<String>(Arrays.asList(args));
		arguments.addFirst(jobmanager.server.JobStarter.class.getName());
		return JobDesc.createJavaJob(project, dataset, parameter, name, arguments.toArray(new String[0]));
	}

	/**
	 * Create a job on the JobManager
	 * 
	 * @param project
	 *            Name of project
	 * @param dataset
	 *            Dataset-name
	 * @param parameter
	 *            Parameter-name
	 * @param name
	 *            Name of job
	 * @param args
	 *            Arguments of job
	 * @return Identifier of job
	 */
	public static String createJavaJob(String project, String dataset, String parameter, String name, String... args) {
		Socket s = null;
		ObjectOutputStream out = null;
		ObjectInputStream in = null;
		String identifier = null;

		try {
			s = JobDesc.getSocket();
			out = new ObjectOutputStream(s.getOutputStream());
			in = new ObjectInputStream(s.getInputStream());

			out.writeObject(Commands.CREATE_JAVA_JOB);
			out.writeObject(project);
			out.writeObject(dataset);
			out.writeObject(parameter);
			out.writeObject(name);
			out.writeObject(args);

			identifier = (String) in.readObject();
		} catch (IOException e) {
		} catch (ClassNotFoundException e) {
		} finally {
			if (out != null) {
				try {
					out.close();
				} catch (IOException e) {
				}
			}
			if (in != null) {
				try {
					in.close();
				} catch (IOException e) {
				}
			}
			if (s != null) {
				try {
					s.close();
				} catch (IOException e) {
				}
			}
		}
		return identifier;
	}

	/**
	 * Create a job on the JobManager
	 * 
	 * @param project
	 *            Name of project
	 * @param dataset
	 *            Dataset-name
	 * @param parameter
	 *            Parameter-name
	 * @param name
	 *            Name of job
	 * @param dockerImage
	 *            Name of docker-image
	 * @param command
	 *            Command of job
	 * @param args
	 *            Arguments of job
	 * @return Identifier of job
	 */
	public static String createDockerJob(String project, String dataset, String parameter, String name,
			String dockerImage, String command, String... args) {
		Socket s = null;
		ObjectOutputStream out = null;
		ObjectInputStream in = null;
		String identifier = null;

		try {
			s = JobDesc.getSocket();
			out = new ObjectOutputStream(s.getOutputStream());
			in = new ObjectInputStream(s.getInputStream());

			out.writeObject(Commands.CREATE_DOCKER_JOB);
			out.writeObject(project);
			out.writeObject(dataset);
			out.writeObject(parameter);
			out.writeObject(name);
			out.writeObject(dockerImage);
			out.writeObject(command);
			out.writeObject(args);

			identifier = (String) in.readObject();
		} catch (IOException e) {
		} catch (ClassNotFoundException e) {
		} finally {
			if (out != null) {
				try {
					out.close();
				} catch (IOException e) {
				}
			}
			if (in != null) {
				try {
					in.close();
				} catch (IOException e) {
				}
			}
			if (s != null) {
				try {
					s.close();
				} catch (IOException e) {
				}
			}
		}
		return identifier;
	}

	/**
	 * Get local path of job
	 * 
	 * @param identifier
	 *            Identifier of job
	 * @return Local path
	 */
	public static File getPath(String identifier) {
		Socket s = null;
		ObjectOutputStream out = null;
		ObjectInputStream in = null;
		File path = null;

		try {
			s = JobDesc.getSocket();
			out = new ObjectOutputStream(s.getOutputStream());
			in = new ObjectInputStream(s.getInputStream());

			out.writeObject(Commands.GET_PATH);
			out.writeObject(identifier);
			path = (File) in.readObject();

		} catch (IOException e) {
		} catch (ClassNotFoundException e) {
		} finally {
			if (out != null) {
				try {
					out.close();
				} catch (IOException e) {
				}
			}
			if (in != null) {
				try {
					in.close();
				} catch (IOException e) {
				}
			}
			if (s != null) {
				try {
					s.close();
				} catch (IOException e) {
				}
			}
		}
		return path;
	}

	/**
	 * Get status of job
	 * 
	 * @param identifier
	 *            Identifier of job
	 * @return Status of job
	 */
	public static JobStatus getStatus(String identifier) {
		Socket s = null;
		ObjectOutputStream out = null;
		ObjectInputStream in = null;
		JobStatus status = JobStatus.UNDEFINED;

		try {
			s = JobDesc.getSocket();
			out = new ObjectOutputStream(s.getOutputStream());
			in = new ObjectInputStream(s.getInputStream());

			out.writeObject(Commands.REQUEST_JOBSTATUS);
			out.writeObject(identifier);
			status = (JobStatus) in.readObject();

		} catch (IOException e) {
		} catch (ClassNotFoundException e) {
		} finally {
			if (out != null) {
				try {
					out.close();
				} catch (IOException e) {
				}
			}
			if (in != null) {
				try {
					in.close();
				} catch (IOException e) {
				}
			}
			if (s != null) {
				try {
					s.close();
				} catch (IOException e) {
				}
			}
		}
		return status;
	}

	/**
	 * Link a file to job
	 * 
	 * @param identifier
	 *            Identifier of job
	 * @param file
	 *            File to link
	 */
	public static void linkFile(String identifier, String file) {
		JobDesc.linkFile(identifier, new File(file));
	}

	/**
	 * Link a file to job
	 * 
	 * @param identifier
	 *            Identifier of job
	 * @param file
	 *            File to link
	 */
	public static void linkFile(String identifier, File file) {
		Socket s = null;
		ObjectOutputStream out = null;
		ObjectInputStream in = null;

		try {
			s = JobDesc.getSocket();
			out = new ObjectOutputStream(s.getOutputStream());
			in = new ObjectInputStream(s.getInputStream());

			out.writeObject(Commands.LINK_FILE);
			out.writeObject(identifier);
			out.writeObject(file);

			in.readObject();
		} catch (IOException e) {
		} catch (ClassNotFoundException e) {
		} finally {
			if (out != null) {
				try {
					out.close();
				} catch (IOException e) {
				}
			}
			if (in != null) {
				try {
					in.close();
				} catch (IOException e) {
				}
			}
			if (s != null) {
				try {
					s.close();
				} catch (IOException e) {
				}
			}
		}
	}

	/**
	 * Copy the given file to the local path
	 * 
	 * @param identifier
	 *            Identifier of job
	 * @param file
	 * @throws IOException
	 */
	public static void addFile(String identifier, File file) throws IOException {
		JobDesc.addFile(identifier, file, file.getName());
	}

	/**
	 * Copy the given file to the local path with given filename
	 * 
	 * @param identifier
	 *            Identifier of job
	 * @param file
	 * @param name
	 * @throws IOException
	 */
	public static void addFile(String identifier, File file, String name) throws IOException {
		File path, dest;

		path = JobDesc.getPath(identifier);
		if (path != null) {
			dest = new File(path, name);
			FileUtils.copy(file, dest);
			JobDesc.linkFile(identifier, dest);
		} else {
			throw new FileNotFoundException("Undefined path");
		}
	}

	/**
	 * Get a OutputStream to create local file with the given name
	 * 
	 * @param identifier
	 *            Identifier of job
	 * @param name
	 *            Name of the local file
	 * @return OutputStream from the given filename
	 * @throws IOException
	 */
	public static OutputStream addFile(String identifier, String name) throws IOException {
		File path, dest;
		OutputStream out = null;

		path = JobDesc.getPath(identifier);
		if (path != null) {
			dest = new File(path, name);
			out = new FileOutputStream(dest);
			JobDesc.linkFile(identifier, dest);
		} else {
			throw new FileNotFoundException("Undefined path");
		}
		return out;
	}

	/**
	 * Set Requirement for job execution
	 * 
	 * @param identifier
	 *            Identifier of job
	 * @param requirement
	 *            Requirements for job
	 */
	public static void setRequirement(String identifier, String requirement) {
		Socket s = null;
		ObjectOutputStream out = null;
		ObjectInputStream in = null;

		try {
			s = JobDesc.getSocket();
			out = new ObjectOutputStream(s.getOutputStream());
			in = new ObjectInputStream(s.getInputStream());

			out.writeObject(Commands.SET_REQUIREMENT);
			out.writeObject(identifier);
			out.writeObject(requirement);

			in.readObject();
		} catch (IOException e) {
		} catch (ClassNotFoundException e) {
		} finally {
			if (out != null) {
				try {
					out.close();
				} catch (IOException e) {
				}
			}
			if (in != null) {
				try {
					in.close();
				} catch (IOException e) {
				}
			}
			if (s != null) {
				try {
					s.close();
				} catch (IOException e) {
				}
			}
		}
	}

	/**
	 * Set timeout for job execution
	 * 
	 * @param identifier
	 *            Identifier of job
	 * @param timeout
	 *            Timeout (unit seconds)
	 */
	public static void setTimeout(String identifier, long timeout) {
		Socket s = null;
		ObjectOutputStream out = null;
		ObjectInputStream in = null;

		try {
			s = JobDesc.getSocket();
			out = new ObjectOutputStream(s.getOutputStream());
			in = new ObjectInputStream(s.getInputStream());

			out.writeObject(Commands.SET_TIMEOUT);
			out.writeObject(identifier);
			out.writeObject(timeout);

			in.readObject();
		} catch (IOException e) {
		} catch (ClassNotFoundException e) {
		} finally {
			if (out != null) {
				try {
					out.close();
				} catch (IOException e) {
				}
			}
			if (in != null) {
				try {
					in.close();
				} catch (IOException e) {
				}
			}
			if (s != null) {
				try {
					s.close();
				} catch (IOException e) {
				}
			}
		}
	}

	/**
	 * Submit job to an execution queue with the given type
	 * 
	 * @param identifier
	 *            Identifier of job
	 * @param type
	 *            Type of execution
	 */
	public static void submit(String identifier, JobType type) {
		Socket s = null;
		ObjectOutputStream out = null;
		ObjectInputStream in = null;

		// TODO JobType
		if (type == JobType.CONDOR) {
			type = JobType.CLUSTER;
		}

		try {
			s = JobDesc.getSocket();
			out = new ObjectOutputStream(s.getOutputStream());
			in = new ObjectInputStream(s.getInputStream());

			out.writeObject(Commands.SUBMIT);
			out.writeObject(identifier);
			out.writeObject(type);

			in.readObject();
		} catch (IOException e) {
		} catch (ClassNotFoundException e) {
		} finally {
			if (out != null) {
				try {
					out.close();
				} catch (IOException e) {
				}
			}
			if (in != null) {
				try {
					in.close();
				} catch (IOException e) {
				}
			}
			if (s != null) {
				try {
					s.close();
				} catch (IOException e) {
				}
			}
		}
	}

	/**
	 * Get the list of file in the local path
	 * 
	 * @param identifier
	 *            Identifier of job
	 * @return List of files
	 * @throws FileNotFoundException
	 */
	public static File[] getFilelist(String identifier) throws FileNotFoundException {
		File path;
		File[] list = null;

		path = JobDesc.getPath(identifier);
		if (path != null) {
			list = path.listFiles();
		} else {
			throw new FileNotFoundException("Undefined path");
		}
		return list;
	}

	/**
	 * Check if file exists
	 * 
	 * @param identifier
	 *            Identifier of job
	 * @param name
	 *            Filename
	 * @return Exists file
	 */
	public static boolean existFile(String identifier, String name) {
		return new File(JobDesc.getPath(identifier), name).exists();
	}

	/**
	 * Get the file from local path an copy to the given file
	 * 
	 * @param identifier
	 *            Identifier of job
	 * @param file
	 *            File to copy
	 * @throws IOException
	 */
	public static void getFile(String identifier, File file) throws IOException {
		JobDesc.getFile(identifier, file.getName(), file);
	}

	/**
	 * Get the file with filename from local path and copy to the given file
	 * 
	 * @param identifier
	 *            Identifier of job
	 * @param name
	 *            Name of the file to get
	 * @param file
	 *            File where the file should copied
	 * @throws IOException
	 */
	public static void getFile(String identifier, String name, File file) throws IOException {
		File path, source;

		path = JobDesc.getPath(identifier);
		if (path != null) {
			source = new File(path, name);
			FileUtils.copy(source, file, true);
		} else {
			throw new FileNotFoundException("Undefined path");
		}
	}

	/**
	 * Get a InputStream for given filename
	 * 
	 * @param identifier
	 *            Identifier of job
	 * @param name
	 *            name of the file which should returned
	 * @return InputStream for the given filename
	 * @throws IOException
	 */
	public static InputStream getFile(String identifier, String name) throws IOException {
		File path;
		InputStream in = null;

		path = JobDesc.getPath(identifier);
		if (path != null) {
			in = new FileInputStream(new File(path, name));
		} else {
			throw new FileNotFoundException("Undefined path");
		}
		return in;
	}

	/**
	 * Remove the job from JobManager
	 * 
	 * @param identifier
	 *            Identifier of job
	 */
	public static void stop(String identifier) {
		Socket s = null;
		ObjectOutputStream out = null;
		ObjectInputStream in = null;

		try {
			s = JobDesc.getSocket();
			out = new ObjectOutputStream(s.getOutputStream());
			in = new ObjectInputStream(s.getInputStream());

			out.writeObject(Commands.STOP);
			out.writeObject(identifier);

			in.readObject();
		} catch (IOException e) {
		} catch (ClassNotFoundException e) {
		} finally {
			if (out != null) {
				try {
					out.close();
				} catch (IOException e) {
				}
			}
			if (in != null) {
				try {
					in.close();
				} catch (IOException e) {
				}
			}
			if (s != null) {
				try {
					s.close();
				} catch (IOException e) {
				}
			}
		}
	}

	/**
	 * Restart job
	 * 
	 * @param identifier
	 *            Identifier of job
	 */
	public static void restart(String identifier) {
		Socket s = null;
		ObjectOutputStream out = null;
		ObjectInputStream in = null;

		try {
			s = JobDesc.getSocket();
			out = new ObjectOutputStream(s.getOutputStream());
			in = new ObjectInputStream(s.getInputStream());

			out.writeObject(Commands.RESTART);
			out.writeObject(identifier);

			in.readObject();
		} catch (IOException e) {
		} catch (ClassNotFoundException e) {
		} finally {
			if (out != null) {
				try {
					out.close();
				} catch (IOException e) {
				}
			}
			if (in != null) {
				try {
					in.close();
				} catch (IOException e) {
				}
			}
			if (s != null) {
				try {
					s.close();
				} catch (IOException e) {
				}
			}
		}
	}

	/**
	 * Restart job
	 * 
	 * @param identifier
	 *            Identifier of job
	 * @param type
	 *            New job type
	 */
	public static void restart(String identifier, JobType type) {
		Socket s = null;
		ObjectOutputStream out = null;
		ObjectInputStream in = null;

		// TODO JobType
		if (type == JobType.CONDOR) {
			type = JobType.CLUSTER;
		}

		try {
			s = JobDesc.getSocket();
			out = new ObjectOutputStream(s.getOutputStream());
			in = new ObjectInputStream(s.getInputStream());

			out.writeObject(Commands.RESTART_TYPE);
			out.writeObject(identifier);
			out.writeObject(type);

			in.readObject();
		} catch (IOException e) {
		} catch (ClassNotFoundException e) {
		} finally {
			if (out != null) {
				try {
					out.close();
				} catch (IOException e) {
				}
			}
			if (in != null) {
				try {
					in.close();
				} catch (IOException e) {
				}
			}
			if (s != null) {
				try {
					s.close();
				} catch (IOException e) {
				}
			}
		}
	}

	/**
	 * Remove the job from JobManager
	 * 
	 * @param identifier
	 *            Identifier of job
	 */
	public static void remove(String identifier) {
		Socket s = null;
		ObjectOutputStream out = null;
		ObjectInputStream in = null;

		try {
			s = JobDesc.getSocket();
			out = new ObjectOutputStream(s.getOutputStream());
			in = new ObjectInputStream(s.getInputStream());

			out.writeObject(Commands.REMOVE_JOB);
			out.writeObject(identifier);

			in.readObject();
		} catch (IOException e) {
		} catch (ClassNotFoundException e) {
		} finally {
			if (out != null) {
				try {
					out.close();
				} catch (IOException e) {
				}
			}
			if (in != null) {
				try {
					in.close();
				} catch (IOException e) {
				}
			}
			if (s != null) {
				try {
					s.close();
				} catch (IOException e) {
				}
			}
		}
	}

	/**
	 * Check if job is finished. The job is finished if it is completed, removed or
	 * stopped because of an error or timeout
	 * 
	 * @param identifier
	 *            Identifier of job
	 * @return Is job finished
	 */
	public static boolean isFinished(String identifier) {
		return JobDesc.isFinished(JobDesc.getStatus(identifier));
	}

	/**
	 * Check if job-status is finished
	 * 
	 * @param status
	 *            Status to check
	 * @return Job is finished
	 */
	private static boolean isFinished(JobStatus status) {
		boolean finish = false;

		switch (status) {
		case COMPLETED:
		case ERROR:
		case REMOVED:
		case TIMEOUT:
		case HELD:
		case UNDEFINED:
			finish = true;
			break;
		default:
			break;
		}

		return finish;

	}

	/**
	 * List parameter categories
	 * 
	 * @param project
	 *            Filter for project
	 * @param dataset
	 *            Filter for dataset
	 * @return List of parameter categories
	 */
	public static String[] getParameterList(String project, String dataset) {
		Socket s = null;
		ObjectOutputStream out = null;
		ObjectInputStream in = null;
		String[] list = null;

		try {
			s = JobDesc.getSocket();
			out = new ObjectOutputStream(s.getOutputStream());
			in = new ObjectInputStream(s.getInputStream());

			out.writeObject(Commands.REQUEST_PARAMETER_LIST);
			out.writeObject(project);
			out.writeObject(dataset);
			list = (String[]) in.readObject();
		} catch (IOException e) {
		} catch (ClassNotFoundException e) {
		} finally {
			if (out != null) {
				try {
					out.close();
				} catch (IOException e) {
				}
			}
			if (in != null) {
				try {
					in.close();
				} catch (IOException e) {
				}
			}
			if (s != null) {
				try {
					s.close();
				} catch (IOException e) {
				}
			}
		}

		return list;
	}

	/**
	 * Check if all job with filter are finished
	 * 
	 * @param project
	 *            Filter for project
	 * @param dataset
	 *            Filter for dataset
	 * @param parameter
	 *            Filter for parameter
	 * @return Are all job of filter finished
	 */
	public static boolean allFinished(String project, String dataset, String parameter) {
		Hashtable<String, JobStatus> list;
		int working = 0;

		list = JobDesc.getStatusList(project, dataset, parameter);
		if (list != null) {
			for (JobStatus status : list.values()) {
				if (!JobDesc.isFinished(status)) {
					working++;
				}
			}
		}

		return working == 0;

	}

	/**
	 * Get list of jobStatus of filtered jobs
	 * 
	 * @param project
	 *            Filter for project
	 * @param dataset
	 *            Filter for dataset
	 * @param parameter
	 *            Filter for parameter
	 * @return Status list of jobs
	 */
	@SuppressWarnings("unchecked")
	public static Hashtable<String, JobStatus> getStatusList(String project, String dataset, String parameter) {
		Socket s = null;
		ObjectOutputStream out = null;
		ObjectInputStream in = null;
		Hashtable<String, JobStatus> list = null;

		try {
			s = JobDesc.getSocket();
			out = new ObjectOutputStream(s.getOutputStream());
			in = new ObjectInputStream(s.getInputStream());

			out.writeObject(Commands.REQUEST_JOBSTATUS_LIST);
			out.writeObject(project);
			out.writeObject(dataset);
			out.writeObject(parameter);
			list = (Hashtable<String, JobStatus>) in.readObject();
		} catch (IOException e) {
		} catch (ClassNotFoundException e) {
		} finally {
			if (out != null) {
				try {
					out.close();
				} catch (IOException e) {
				}
			}
			if (in != null) {
				try {
					in.close();
				} catch (IOException e) {
				}
			}
			if (s != null) {
				try {
					s.close();
				} catch (IOException e) {
				}
			}
		}
		return list;
	}

	/**
	 * Is job-manager running (do not start process)
	 * 
	 * @return Is job-manager is running
	 */
	public static boolean isRunning() {
		Socket s = null;

		try {
			s = new Socket("", JobDesc.getPort());
		} catch (UnknownHostException e1) {
		} catch (IOException e1) {
		} finally {
			if (s != null) {
				try {
					s.close();
				} catch (IOException e) {
				}
			}
		}

		return (s != null);
	}
}
