/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9720 $
 * $Date: 2020-03-19 15:59:49 +0100 (Thu, 19 Mar 2020) $
 */
package jobmanager.server.gui;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.util.Arrays;
import java.util.Hashtable;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import javax.swing.JFrame;
import javax.swing.JScrollPane;
import javax.swing.JTabbedPane;
import javax.swing.JTable;
import javax.swing.ListSelectionModel;
import javax.swing.WindowConstants;
import javax.swing.border.EmptyBorder;
import javax.swing.table.TableColumn;

import at.lcm.bu21.magopt.utils.MagOptConfig;
import jobmanager.JobType;
import jobmanager.server.Job;
import jobmanager.server.JobManager;
import jobmanager.server.gui.icons.JobIcons;
import magopt.gui.console.Console;

/**
 * View for JobManager object
 * 
 * @author Werner Koppelsttter
 */
public class JobManagerView extends JFrame {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** Tabs */
	JTabbedPane tab;

	/** Table for job-list */
	JTable table = null;

	/** Table model for job-list */
	JobListModel model = null;

	/** Status panel */
	StatusPanel statusPanel = null;

	/** Console */
	Console console = null;

	/**
	 * Create a JobManager-view
	 */
	public JobManagerView() {
		super("JobManager:" + JobManager.getPort());

		this.setIconImages(JobIcons.getIcons());

		// Frame
		this.setSize(1100, 600);
		this.setLayout(new BorderLayout());
		this.setDefaultCloseOperation(WindowConstants.DO_NOTHING_ON_CLOSE);
		this.addWindowListener(new WindowAdapter() {
			@Override
			public void windowClosing(WindowEvent e) {
				JobManager.setVisible(false);
			}
		});

		this.tab = new JTabbedPane();
		this.add(this.tab, BorderLayout.CENTER);

		this.initJobTable();
		this.tab.addTab("Jobs", new JScrollPane(this.table));

		this.console = new Console();
//		this.tab.addTab("Console", this.console);
		System.setErr(this.console.getErrorPrintStream());
		System.setOut(this.console.getOutputPrintStream());

		try {
			Class.forName("jobmanager.server.htcondor.HTCondorInterface")
					.getMethod("refreshPanels", JTabbedPane.class, Boolean.class).invoke(null, this.tab, false);
		} catch (Throwable ex) {
		}

		this.initStatusPanel();
		this.add(this.statusPanel, BorderLayout.SOUTH);

		this.setVisible(true);

		this.refresh();

		this.table.requestFocus();
	}

	/**
	 * Initialize job-table
	 */
	private void initJobTable() {
		TableColumn col;

		// Table
		this.model = new JobListModel();
		this.table = new JTable(this.model);
		this.table.setBorder(new EmptyBorder(5, 5, 5, 5));

		/* set max and prefered width of columns */
		col = this.table.getColumnModel().getColumn(JobListColumn.PROJECT.getValue());
		col.setPreferredWidth(160);
		col = this.table.getColumnModel().getColumn(JobListColumn.DATASET.getValue());
		col.setMaxWidth(100);
		col.setPreferredWidth(60);
		col = this.table.getColumnModel().getColumn(JobListColumn.PARAMETER.getValue());
		col.setMaxWidth(100);
		col.setPreferredWidth(60);
		col = this.table.getColumnModel().getColumn(JobListColumn.NAME.getValue());
		col.setPreferredWidth(160);
		col = this.table.getColumnModel().getColumn(JobListColumn.STATUS.getValue());
		col.setMaxWidth(150);
		col.setPreferredWidth(120);
		col = this.table.getColumnModel().getColumn(JobListColumn.ID.getValue());
		col.setMaxWidth(160);
		col.setPreferredWidth(120);
		col = this.table.getColumnModel().getColumn(JobListColumn.RUNTIME.getValue());
		col.setMaxWidth(120);
		col.setPreferredWidth(100);
		col = this.table.getColumnModel().getColumn(JobListColumn.HOST.getValue());
		col.setPreferredWidth(200);

		this.table.setDefaultRenderer(String.class, new JobListRenderer());
		this.table.getTableHeader().setReorderingAllowed(false);

		this.table.addKeyListener(new KeyAdapter() {
			@Override
			public void keyPressed(KeyEvent e) {
				Job.refreshList();
				if (!e.isControlDown() && !e.isAltDown() && !e.isShiftDown()) {
					try {
						switch (e.getKeyCode()) {
						case KeyEvent.VK_DELETE:
							JobManagerView.this.removeSelected();
							break;
						case KeyEvent.VK_R:
							JobManagerView.this.restartSelected();
							break;
						case KeyEvent.VK_S:
							JobManagerView.this.stopSelected();
							break;
						case KeyEvent.VK_C:
							JobManagerView.this.restartClusterSelected();
							break;
						case KeyEvent.VK_L:
							JobManagerView.this.restartLocalSelected();
							break;
						case KeyEvent.VK_E:
							JobManagerView.this.openDirectory();
							break;
						case KeyEvent.VK_F5:
							JobManager.refresh();
							JobManagerView.this.refresh();
							break;
						}
					} catch (Exception e1) {
						// e1.printStackTrace();
					}
				}
			}
		});
		this.table.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseReleased(MouseEvent me) {
				if (me.isPopupTrigger()) {
					new JobListPopupMenu(JobManagerView.this).show(me.getComponent(), me.getX(), me.getY());
				}
			}
		});
	}

	/**
	 * Initialize status-panel
	 */
	private void initStatusPanel() {
		// summary panel
		this.statusPanel = new StatusPanel(new Color[] { Color.GREEN, Color.ORANGE, Color.RED, Color.GRAY });
		this.statusPanel.addMouseListener(0, new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent event) {
				if (event.getClickCount() > 1) {
					JobManagerView.this.refresh();
					JobManagerView.this.select(JobManagerView.this.groupJobs().get("Completed"));
				}
			}
		});
		this.statusPanel.addMouseListener(1, new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent event) {
				if (event.getClickCount() > 1) {
					JobManagerView.this.refresh();
					JobManagerView.this.select(JobManagerView.this.groupJobs().get("InProgress"));
				}
			}
		});
		this.statusPanel.addMouseListener(2, new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent event) {
				if (event.getClickCount() > 1) {
					JobManagerView.this.refresh();
					JobManagerView.this.select(JobManagerView.this.groupJobs().get("Error"));
				}
			}
		});
		this.statusPanel.addMouseListener(3, new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent event) {
				if (event.getClickCount() > 1) {
					JobManagerView.this.refresh();
					JobManagerView.this.select(JobManagerView.this.groupJobs().get("Undefined"));
				}
			}
		});
	}

	/**
	 * Close view
	 */
	public void close() {
		this.setVisible(false);
		this.dispose();
	}

	/**
	 * Refresh job-list
	 */
	public void refresh() {
		Job[] jobs;
		Map<String, List<Job>> groups;
		int[] selected;
		LinkedList<Job> selectedJobs;
		String val;
		boolean jobDebug = false;

		val = MagOptConfig.getConfig("JobDebug");
		if (val != null) {
			jobDebug = Boolean.parseBoolean(val);
		}

		if (jobDebug) {
			if (this.tab.indexOfComponent(this.console) == -1) {
				this.tab.addTab("Console", this.console);
			}
		} else {
			if (this.tab.indexOfComponent(this.console) != -1) {
				this.tab.remove(this.console);
			}
		}

		try {
			Class.forName("jobmanager.server.htcondor.HTCondorInterface")
					.getMethod("refreshPanels", JTabbedPane.class, Boolean.class).invoke(null, this.tab, jobDebug);
		} catch (Throwable ex) {
			System.err.println(ex.getMessage());
			ex.printStackTrace();
		}

		Job.refreshList();

		// get selection
		selected = this.table.getSelectedRows();
		selectedJobs = new LinkedList<Job>();
		for (int k = 0; k < selected.length; k++) {
			selectedJobs.add((Job) this.model.getValueAt(selected[k], -1));
		}

		// fill jobs
		jobs = Job.joblist.values().toArray(new Job[0]);
		Arrays.sort(jobs);
		this.model.setJobs(jobs);

		// set selection
		this.select(selectedJobs);

		// refresh status-panel
		groups = this.groupJobs();
		this.statusPanel.setWeight(new int[] { groups.get("Completed").size(), groups.get("InProgress").size(),
				groups.get("Error").size(), groups.get("Undefined").size() });
	}

	/**
	 * Group jobs to status lists
	 * 
	 * @return Group of status
	 */
	private Map<String, List<Job>> groupJobs() {
		Hashtable<String, List<Job>> groups;
		LinkedList<Job> complete, progress, error, undefined;

		// initialize lists
		groups = new Hashtable<String, List<Job>>();
		complete = new LinkedList<Job>();
		groups.put("Completed", complete);
		progress = new LinkedList<Job>();
		groups.put("InProgress", progress);
		error = new LinkedList<Job>();
		groups.put("Error", error);
		undefined = new LinkedList<Job>();
		groups.put("Undefined", undefined);

		for (Job job : Job.joblist.values()) {
			switch (job.getStatus()) {
			case COMPLETED:
				complete.add(job);
				break;
			case CREATED:
			case SUBMIT:
			case SUBMITTED:
			case RUNNING:
			case COMPLETE:
			case RESTART:
			case STOP:
			case SUSPENDED:
				progress.add(job);
				break;
			case ERROR:
			case TIMEOUT:
			case HELD:
				error.add(job);
				break;
			case UNDEFINED:
			case REMOVED:
				undefined.add(job);
				break;
			default:
				break;
			}
		}

		return groups;
	}

	/**
	 * Select jobs of list
	 * 
	 * @param select
	 *            List of job to select
	 */
	private void select(List<Job> select) {
		ListSelectionModel selectModel;

		selectModel = this.table.getSelectionModel();
		selectModel.clearSelection();
		for (int k = 0; k < this.model.getRowCount(); k++) {
			if (select.remove(this.model.getValueAt(k, -1))) {
				selectModel.addSelectionInterval(k, k);
			}
		}
	}

	/**
	 * Stop selected Jobs
	 */
	void stopSelected() {
		int[] selected;
		Job job;

		selected = this.table.getSelectedRows();
		for (int sel : selected) {
			job = ((Job) this.model.getValueAt(sel, -1));
			job.stop();
			this.model.fireTableRowsUpdated(sel, sel);
		}
	}

	/**
	 * Restart selected jobs
	 */
	void restartSelected() {
		int[] selected;
		Job job;

		selected = this.table.getSelectedRows();
		for (int sel : selected) {
			job = ((Job) this.model.getValueAt(sel, -1));
			job.restart();
			this.model.fireTableRowsUpdated(sel, sel);
		}
	}

	/**
	 * Restart selected jobs local
	 */
	void restartLocalSelected() {
		int[] selected;
		Job job;

		selected = this.table.getSelectedRows();
		for (int sel : selected) {
			job = ((Job) this.model.getValueAt(sel, -1));
			job.setType(JobType.LOCAL);
			job.restart();
			this.model.fireTableRowsUpdated(sel, sel);
		}
	}

	/**
	 * Restart selected jobs on cluster
	 */
	void restartClusterSelected() {
		int[] selected;
		Job job;

		selected = this.table.getSelectedRows();
		for (int sel : selected) {
			job = ((Job) this.model.getValueAt(sel, -1));
			job.setType(JobType.CLUSTER);
			job.restart();
			this.model.fireTableRowsUpdated(sel, sel);
		}
	}

	/**
	 * Remove selected jobs
	 */
	void removeSelected() {
		int[] selected;
		Job job;

		selected = this.table.getSelectedRows();
		for (int sel : selected) {
			job = ((Job) this.model.getValueAt(sel, -1));
			job.remove();
			this.model.fireTableRowsUpdated(sel, sel);
		}
	}

	/**
	 * Open directory of selected job
	 */
	void openDirectory() {
		int sel;
		Job job;

		sel = this.table.getSelectedRow();
		job = ((Job) this.model.getValueAt(sel, -1));
		job.openDirectory();
		this.model.fireTableRowsUpdated(sel, sel);
	}
}
