/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9720 $
 * $Date: 2020-03-19 15:59:49 +0100 (Thu, 19 Mar 2020) $
 */
package viewer3d;

import java.awt.BorderLayout;
import java.awt.GraphicsConfiguration;

import javax.swing.JFrame;
import javax.swing.JPanel;

import com.sun.j3d.utils.behaviors.vp.OrbitBehavior;
import com.sun.j3d.utils.universe.PlatformGeometry;
import com.sun.j3d.utils.universe.SimpleUniverse;
import com.sun.j3d.utils.universe.ViewingPlatform;

import javax.media.j3d.AmbientLight;
import javax.media.j3d.Background;
import javax.media.j3d.BoundingSphere;
import javax.media.j3d.BranchGroup;
import javax.media.j3d.Canvas3D;
import javax.media.j3d.DirectionalLight;
import javax.media.j3d.Transform3D;

import javax.vecmath.Color3f;
import javax.vecmath.Matrix4d;
import javax.vecmath.Point3d;
import javax.vecmath.Vector3d;
import javax.vecmath.Vector3f;

/**
 * Java 3D viewer program to display an .obj object.
 * 
 * @author Werner Koppelsttter
 */
public class Viewer extends JPanel {

	/**
	 * Standard-view direction front
	 */
	public final static int FRONT = 0;

	/**
	 * Standard-view direction back
	 */
	public final static int BACK = 1;

	/**
	 * Standard-view direction left
	 */
	public final static int LEFT = 2;

	/**
	 * Standard-view direction right
	 */
	public final static int RIGHT = 3;

	/**
	 * Standard-view direction bottom
	 */
	public final static int BOTTOM = 4;

	/**
	 * Standard-view direction top
	 */
	public final static int TOP = 5;

	/**
	 * Serialization version number
	 */
	private static final long serialVersionUID = 1L;

	/**
	 * Universe
	 */
	public SimpleUniverse univ = null;

	/**
	 * Scene-graph branch-group
	 */
	private BranchGroup scene = null;

	/**
	 * Base transformation group
	 */
	private Transform baseTransform = null;

	/**
	 * Scale base transformation group
	 * 
	 * @param scale
	 *            scale-value
	 */
	public void scale(double scale) {
		this.baseTransform.scale(scale);
	}

	/**
	 * Get base transform group
	 * 
	 * @return transform group
	 */
	public Transform getBaseTransform() {
		return this.baseTransform;
	}

	/**
	 * Add model to scene-graph
	 * 
	 * @param filename
	 *            filename of model
	 * @return transform group object
	 */
	public Transform addModel(String filename) {
		Transform trans = new Transform();

		trans.addModel(filename);

		this.baseTransform.addTransform(trans);

		// Transform3D axis = new Transform3D();
		// float[] knots = new float[361];
		// Quat4f[] quats = new Quat4f[361];
		// Point3f[] pos = new Point3f[361];
		//
		// for (int k=0; k < 361; k++) {
		// knots[k] = k/360.0f;
		// quats[k] = new Quat4f((float) Math.sin(k*2*Math.PI/360/2), 0.0f,
		// 0.0f, (float) Math.cos(k*2*Math.PI/360/2));
		// pos[k] = new Point3f(0.0f, 0.0f, 0.0f);
		// }
		//
		// Alpha rotationAlpha = new Alpha(
		// -1, // loopCount
		// Alpha.INCREASING_ENABLE, // mode
		// 0, // triggerTime
		// 0, // phaseDelayDuration
		// 40000, // increasingAlphaDuration
		// 0, // increasingAlphaRampDuration
		// 0, // alphaAtOneDuration
		// 0, // decreasingAlphaDuration
		// 0, // decreasingAlphaRampDuration
		// 0 // alphaAtZeroDuration
		// );
		// RotPosPathInterpolator rotator = new
		// RotPosPathInterpolator(rotationAlpha, objTrans, axis, knots, quats,
		// pos);
		// rotator.setSchedulingBounds(bounds);
		// objTrans.addChild(rotator);

		return trans;
	}

	/**
	 * Add transform-object to scene-graph
	 * 
	 * @param trans
	 *            Transform object
	 */
	public void addTransform(Transform trans) {
		this.baseTransform.addTransform(trans);
	}

	/**
	 * Create scene-graph basis
	 * 
	 * @return scene-graph
	 */
	private BranchGroup createSceneGraph() {
		BranchGroup objRoot = new BranchGroup();
		objRoot.setCapability(BranchGroup.ALLOW_CHILDREN_EXTEND);

		BoundingSphere bounds = new BoundingSphere(new Point3d(0.0, 0.0, 0.0),
				100.0);

		this.baseTransform = new Transform();
		objRoot.addChild(this.baseTransform.getBranchGroup());

		// Set up the background
		Color3f bgColor = new Color3f(0.05f, 0.05f, 0.5f);
		Background bgNode = new Background(bgColor);
		bgNode.setApplicationBounds(bounds);
		objRoot.addChild(bgNode);

		return objRoot;
	}

	/**
	 * Create Universe for viewing models
	 * 
	 * @return universe
	 */
	private Canvas3D createUniverse() {
		// Get the preferred graphics configuration for the default screen
		GraphicsConfiguration config = SimpleUniverse
				.getPreferredConfiguration();

		// Create a Canvas3D using the preferred configuration
		Canvas3D canvas3d = new Canvas3D(config);

		// Create simple universe with view branch
		univ = new SimpleUniverse(canvas3d);
		BoundingSphere bounds = new BoundingSphere(new Point3d(0.0, 0.0, 0.0),
				100.0);

		// add mouse behaviors and lights to the ViewingPlatform
		ViewingPlatform viewingPlatform = univ.getViewingPlatform();

		// Set OrbitBehavior for mouse control
		viewingPlatform.setNominalViewingTransform();
		OrbitBehavior orbit = new MyOrbitBehavior(canvas3d,
				OrbitBehavior.REVERSE_ALL);
		orbit.setSchedulingBounds(bounds);
		viewingPlatform.setViewPlatformBehavior(orbit);

		// add lights to ViewingPlatform
		PlatformGeometry pg = new PlatformGeometry();

		// Set up the ambient light
		Color3f ambientColor = new Color3f(0.1f, 0.1f, 0.1f);
		AmbientLight ambientLightNode = new AmbientLight(ambientColor);
		ambientLightNode.setInfluencingBounds(bounds);
		pg.addChild(ambientLightNode);

		// Set up the directional lights
		Color3f light1Color = new Color3f(1.0f, 1.0f, 0.9f);
		Vector3f light1Direction = new Vector3f(1.0f, 1.0f, 1.0f);
		Color3f light2Color = new Color3f(1.0f, 1.0f, 1.0f);
		Vector3f light2Direction = new Vector3f(-1.0f, -1.0f, -1.0f);

		DirectionalLight light1 = new DirectionalLight(light1Color,
				light1Direction);
		light1.setInfluencingBounds(bounds);
		pg.addChild(light1);

		DirectionalLight light2 = new DirectionalLight(light2Color,
				light2Direction);
		light2.setInfluencingBounds(bounds);
		pg.addChild(light2);

		viewingPlatform.setPlatformGeometry(pg);

		// // Ensure at least 5 msec per frame (i.e., < 200Hz)
		// univ.getViewer().getView().setMinimumFrameCycleTime(5);

		return canvas3d;
	}

	/**
	 * Creates new Viewer
	 */
	public Viewer() {

		// Initialize the GUI components
		this.setLayout(new BorderLayout());

		// Create Canvas3D and SimpleUniverse; add canvas to drawing panel
		Canvas3D c = createUniverse();
		this.add(c, java.awt.BorderLayout.CENTER);

		// Create the content branch and add it to the universe
		scene = createSceneGraph();
		univ.addBranchGraph(scene);

	}

	/**
	 * Create frame for the panel
	 * 
	 * @return frame contains panel
	 */
	public JFrame createFrame() {
		JFrame frame = new JFrame("Viewer3D");
		frame.setPreferredSize(new java.awt.Dimension(700, 700));
		frame.setDefaultCloseOperation(javax.swing.WindowConstants.DISPOSE_ON_CLOSE);

		frame.getContentPane().add(this, BorderLayout.CENTER);
		frame.pack();
		frame.setVisible(true);

		return frame;
	}

	/**
	 * Entrypoint for application
	 * 
	 * @param args
	 *            command line arguments
	 */
	public static void main(final String args[]) {
		Viewer view;
		JFrame frame;

		if ((args != null) && (args.length != 0)) {
			view = new Viewer();
			frame = view.createFrame();
			frame.setDefaultCloseOperation(javax.swing.WindowConstants.EXIT_ON_CLOSE);

			for (int i = 0; i < args.length; i++) {
				view.addModel(args[i]);
			}
		} else {
			System.out.println("Viewer {<.obj file> ...}");
		}

	}

	/**
	 * Set direction of view
	 * 
	 * @param phiX
	 *            rotation-angle on x-axis (radiant)
	 * @param phiY
	 *            rotation-angle on y-axis (radiant)
	 * @param phiZ
	 *            rotation-angle on z-axis (radiant)
	 */
	public void setView(double phiX, double phiY, double phiZ) {
		double disdance = 1 / Math.tan(Math.PI / 8);
		Transform3D trans = new Transform3D();

		trans.set(new Matrix4d(Math.cos(phiY) * Math.cos(phiZ), -Math.cos(phiY)
				* Math.sin(phiZ), Math.sin(phiY), 0.0, Math.sin(phiX)
				* Math.sin(phiY) * Math.cos(phiZ) + Math.cos(phiX)
				* Math.sin(phiZ), -Math.sin(phiX) * Math.sin(phiY)
				* Math.sin(phiZ) + Math.cos(phiX) * Math.cos(phiZ), -Math
				.sin(phiX) * Math.cos(phiY), 0.0, -Math.cos(phiX)
				* Math.sin(phiY) * Math.cos(phiZ) + Math.sin(phiX)
				* Math.sin(phiZ), Math.cos(phiX) * Math.sin(phiY)
				* Math.sin(phiZ) + Math.sin(phiX) * Math.cos(phiZ), Math
				.cos(phiX) * Math.cos(phiY), 0.0, 0.0, 0.0, 0.0, 1.0));

		trans.setTranslation(new Vector3d(Math.sin(phiY) * disdance, -Math
				.sin(phiX) * Math.cos(phiY) * disdance, Math.cos(phiX)
				* Math.cos(phiY) * disdance));
		this.univ.getViewingPlatform().getViewPlatformTransform()
				.setTransform(trans);
	}

	/**
	 * Set standard direction of view
	 * 
	 * @param view
	 *            Constant for standard-view
	 */
	public void setView(int view) {
		switch (view) {
		case FRONT:
			this.setView(0.0, 0.0, 0.0);
			break;
		case BACK:
			this.setView(0.0, Math.PI, 0.0);
			break;
		case RIGHT:
			this.setView(0.0, Math.PI / 2, 0.0);
			break;
		case LEFT:
			this.setView(0.0, -Math.PI / 2, 0.0);
			break;
		case BOTTOM:
			this.setView(Math.PI / 2, 0.0, 0.0);
			break;
		case TOP:
			this.setView(-Math.PI / 2, 0.0, 0.0);
			break;
		}
	}

}
